#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <stdint.h>
#include "aoedisk.h"
#include "etherproto.h"
#include "enet.h"

#include <sys/types.h>
#include <sys/ioctl.h>
#include <netinet/in.h>		/* for htons etc */
#include <net/if.h>
#include <sys/socket.h>
#include <netpacket/packet.h>

struct enet {
	int fd;
	int proto;
	int mtu;
	char mac[LEN_EADDR];
	char ifname[IFNAMSIZ];
};

static int selectif(int, char *, void (*)(char *, ...));
static int getifname(Enet, void (*)(char *, ...), char *);
static int setmtu(Enet, void (*)(char *, ...), char *);
static int getmtu(Enet, void (*)(char *, ...), char *);

Enet
enetinit(dev, mtu, proto, eout)
char *dev;
int mtu, proto;
void (*eout)(char *, ...);
{
	Enet ep;
	char *dn = dev ? dev : "(default)";

	if ((ep = malloc(sizeof(struct enet))) == NULL) {
		if (eout)
			(*eout)("no memory for new Enet");
		return (NULL);
	}
	ep->proto = proto;
	ep->mtu = mtu;
	if ((ep->fd = socket(PF_PACKET, SOCK_RAW, htons(ep->proto))) < 0) {
		if (eout)
			(*eout)("init ether %s: cannot create raw socket: %s",
				dn, strerror(errno));
		free(ep);
		return (NULL);
	}
	if (dev && selectif(ep->fd, dev, eout) < 0) {
		close(ep->fd);
		free(ep);
		return (NULL);
	}
	if (getifname(ep, eout, dn) < 0) {
		close(ep->fd);
		free(ep);
		return (NULL);
	}
	if (ep->mtu > 0 && setmtu(ep, eout, dn) < 0) {
		close(ep->fd);
		free(ep);
		return (NULL);
	}
	if (getmtu(ep, eout, dn) < 0) {
		close(ep->fd);
		free(ep);
		return (NULL);
	}
	return (ep);
}

static int
selectif(fd, dev, eout)
int fd;
char *dev;
void (*eout)(char *, ...);
{
	struct ifreq ifr;
	struct sockaddr_ll sa;
	socklen_t salen;

	strncpy(ifr.ifr_name, dev, IFNAMSIZ);
	if (ioctl(fd, SIOCGIFINDEX, &ifr) < 0) {
		if (eout)
			(*eout)("init ether %s: cannot find interface index: %s",
				dev, strerror(errno));
		return (-1);
	}
	salen = sizeof(sa);
	if (getsockname(fd, (struct sockaddr *)&sa, &salen) < 0) {
		if (eout)
			(*eout)("init ether %s: cannot get socket name (selectif): %s",
				dev, strerror(errno));
		return (-1);
	}
	salen = sizeof(sa);	/* getsockname returns a short value! */
	sa.sll_ifindex = ifr.ifr_ifindex;
	if (bind(fd, (struct sockaddr *)&sa, salen) < 0) {
		if (eout)
			(*eout)("init ether %s: cannot bind to interface: %s",
				dev, strerror(errno));
		return (-1);
	}
	return (1);
}

static int
getifname(ep, eout, dev)
Enet ep;
void (*eout)(char *, ...);
char *dev;
{
	struct ifreq ifr;
	struct sockaddr_ll sa;
	socklen_t salen;

	salen = sizeof(sa);
	if (getsockname(ep->fd, (struct sockaddr *)&sa, &salen) < 0) {
		if (eout)
			(*eout)("init ether %s: cannot get socket name (getifname): %s",
				dev, strerror(errno));
		return (-1);
	}
	memcpy(ep->mac, sa.sll_addr, LEN_EADDR);
	ifr.ifr_ifindex = sa.sll_ifindex;
	if (ioctl(ep->fd, SIOCGIFNAME, &ifr) < 0) {
		if (eout)
			(*eout)("init ether %s: cannot get interface name: %s",
				dev, strerror(errno));
		return (-1);
	}
	strncpy(ep->ifname, ifr.ifr_name, sizeof(ep->ifname));
	return (1);
}

static int
getmtu(ep, eout, dev)
Enet ep;
void (*eout)(char *, ...);
char *dev;
{
	struct ifreq ifr;

	strncpy(ifr.ifr_name, ep->ifname, sizeof(ifr.ifr_name));
	if (ioctl(ep->fd, SIOCGIFMTU, &ifr) < 0) {
		if (eout)
			(*eout)("init ether %s: cannot get interface MTU: %s",
				dev, strerror(errno));
		return (-1);
	}
	ep->mtu = ifr.ifr_mtu;
	return (1);
}

static int
setmtu(ep, eout, dev)
Enet ep;
void (*eout)(char *, ...);
char *dev;
{
	struct ifreq ifr;

	strncpy(ifr.ifr_name, ep->ifname, sizeof(ifr.ifr_name));
	ifr.ifr_mtu = ep->mtu;
	if (ioctl(ep->fd, SIOCSIFMTU, &ifr) < 0) {
		if (eout)
			(*eout)("init ether %s: cannot get interface MTU: %s",
				dev, strerror(errno));
		return (-1);
	}
	return (1);
}

int
enetmtu(ep)
Enet ep;
{
	return (ep->mtu);
}

void
enetclose(ep)
Enet ep;
{
	close(ep->fd);
	free(ep);
}

int
enetrecv(ep, buf, len)
Enet ep;
void *buf;
int len;
{
	return (recv(ep->fd, buf, len, 0));
}

int
enetsend(ep, buf, len)
Enet ep;
void *buf;
int len;
{
	char *p;

	p = buf;
	memcpy(p + ETHER_SRC, ep->mac, LEN_EADDR);
	return (send(ep->fd, buf, len, 0));
}
