#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <errno.h>
#include <stdint.h>
#include "aoedisk.h"
#include "etherproto.h"
#include "enet.h"

#include <unistd.h>
#include <signal.h>
#include <setjmp.h>
#include <fcntl.h>
#include <sys/time.h>
#include <stropts.h>
#include <sys/dlpi.h>
#include <inet/nd.h>

#define	DLREPLY	1024
#define	INITTMO	(3*1000)	/* msec */

#define	RDYSECS	30

#define	NOUNIT	(-1)

#define	ARB	100

struct enet {
	int fd;
	int proto;
	int mtu;
	char mac[LEN_EADDR];
	void (*eout)(char *, ...);
};

static int etheropen(Enet, char *);
static int dlpisetup(Enet, int, char *);
static int msgout(Enet, void *, int, void *, int);
static int msgin(Enet, void *, int, int);
static void msgtmo(void);
static int etherready(Enet, char *, int);
static char *ndop(Enet, char *, char *);
static int rawether(Enet, char *);
static void nullerr(char *, ...);

Enet
enetinit(dev, mtu, proto, eout)
char *dev;
int mtu, proto;
void (*eout)(char *, ...);
{
	Enet ep;
	char *dn = dev ? dev : "(default)";

	if ((ep = malloc(sizeof(struct enet))) == NULL) {
		if (eout)
			(*eout)("no memory for new Enet");
		return (NULL);
	}
	ep->proto = proto;
	ep->mtu = mtu;	/* assume ok */
	if ((ep->eout = eout) == NULL)
		ep->eout = nullerr;
	if (etheropen(ep, dev) < 0) {
		free(ep);
		return (NULL);
	}
	return (ep);
}

/*
 * Solaris/SVr4 DLPI gore
 * to open an Ethernet interface
 * and configure it to send and receive
 * a particular Ethernet protocol type
 */
static int
etheropen(ep, name)
Enet ep;
char *name;
{
	char d[ARB];
	int fd;
	char *up;
	int unit;

	strncpy(d, name, sizeof(d) - 1);
	d[sizeof(d)-1] = 0;
	/*
	 * split d into the part before and after numeric suffix
	 */
	for (up = d + strlen(d); up > d && isdigit(up[-1]); --up)
		;
	if (*up == 0)
		unit = 0;	/* no digits */
	else {
		unit = atoi(up);
		*up = 0;
	}
	if ((ep->fd = open(name, O_RDWR)) >= 0)
		unit = NODEV;
	else if ((ep->fd = open(d, O_RDWR)) < 0) {
		(*ep->eout)("%s: cannot open: %s",
			name, strerror(errno));
		return (-1);
	}
	if (dlpisetup(ep, unit, name) < 0) {
		close(ep->fd);
		return (-1);
	}
	if (etherready(ep, d, unit) < 0) {
		close(ep->fd);
		return (-1);
	}
	if (rawether(ep, name) < 0) {
		close(ep->fd);
		return (-1);
	}
	return (1);
}

/*
 * DLPI initialization commands:
 * attach to the PPA, bind the SAP
 */
static int
dlpisetup(ep, unit, name)
Enet ep;
int unit;
char *name;
{
	dl_attach_req_t dlatt;
	union {
		dl_ok_ack_t o;
		dl_error_ack_t e;
	} dlstat;
	dl_bind_req_t dlbind;
	union {
		char a[DLREPLY];	/* room for addr */
		dl_bind_ack_t b;
		dl_phys_addr_ack_t p;
	} dlback;
	int n;

	/*
	 * DL_ATTACH_REQ: attach to desired PPA (device unit)
	 * required only if we opened a unitless device
	 */
	if (unit != NOUNIT) {
		dlatt.dl_primitive = DL_ATTACH_REQ;
		dlatt.dl_ppa = unit;
		if (msgout(ep, &dlatt, sizeof(dlatt), NULL, 0) < 0
		|| (n = msgin(ep, &dlstat, sizeof(dlstat), 0)) < sizeof(dlstat.o)) {
			(*ep->eout)("%s: cannot DL_ATTACH: %d/%d bytes",
				name, n, sizeof(dlstat.o));
			return (-1);
		}
		if (dlstat.e.dl_primitive == DL_ERROR_ACK) {
			(*ep->eout)("%s: DL_ATTACH error: dlpi %d%s%s\n",
				name, dlstat.e.dl_errno,
				dlstat.e.dl_errno == DL_SYSERR ? ": " : "",
				dlstat.e.dl_errno == DL_SYSERR ?
					strerror(dlstat.e.dl_unix_errno) : "");
			return (-1);
		}
		if (dlstat.o.dl_primitive != DL_OK_ACK) {
			(*ep->eout)("%s: DL_ATTACH returned 0x%x",
				name, dlstat.o.dl_primitive);
			return (-1);
		}
		if (dlstat.o.dl_correct_primitive != DL_ATTACH_REQ) {
			(*ep->eout)("%s: unexpected ack for cmd 0x%x",
				name, dlstat.o.dl_correct_primitive);
			return (-1);
		}
	}
	/*
	 * DL_BIND: set Ethernet protocol number
	 */
	memset(&dlbind, 0, sizeof(dlbind));
	dlbind.dl_primitive = DL_BIND_REQ;
	dlbind.dl_sap = ep->proto;
	dlbind.dl_service_mode = DL_CLDLS;
	if (msgout(ep, &dlbind, sizeof(dlbind), NULL, 0) < 0
	|| (n = msgin(ep, &dlback, sizeof(dlback), 0)) < sizeof(dlback.b)) {
		(*ep->eout)("%s: cannot DL_BIND: %d/%d",
			name, n, sizeof(dlback));
		return (-1);
	}
	if (dlback.b.dl_primitive != DL_BIND_ACK) {
		(*ep->eout)("%s: DL_BIND_ACK returned 0x%x",
			name, dlback.b.dl_primitive);
		return (-1);
	}
	if (dlback.b.dl_addr_offset + dlback.b.dl_addr_length > n) {
		(*ep->eout)("%s: DL_BIND_ACK garbled: addr_offset %d, length %d, message len %d",
			name, dlback.b.dl_addr_offset,
			dlback.b.dl_addr_length, n);
		return (-1);
	}
	/*
	 * assume address returned by bind is [MAC addr||ether proto]
	 */
	if (dlback.b.dl_addr_length != LEN_EADDR + 2) {
		(*ep->eout)("%s: DL_BIND_ACK address length unexpected: %d/%d",
			name, dlback.b.dl_addr_length, LEN_EADDR + 2);
		return (-1);
	}
	memcpy(ep->mac, dlback.a + dlback.b.dl_addr_offset, LEN_EADDR);
	return (1);
}

static int
msgout(ep, obuf, olen, dbuf, dlen)
Enet ep;
void *obuf;
int olen;
void *dbuf;
int dlen;
{
	struct strbuf ctl, dat;

	ctl.len = olen;
	ctl.buf = obuf;
	dat.len = dlen;
	dat.buf = dbuf;
	if (putmsg(ep->fd, &ctl, dbuf ? &dat : NULL, 0) < 0) {
		(*ep->eout)("msgout: putmsg: %s", strerror(errno));
		return (-1);
	}
	return (1);
}

static jmp_buf msgjmp;

static int
msgin(ep, ibuf, ilen, tmo)
Enet ep;
void *ibuf;
int ilen;
int tmo;
{
	struct strbuf ctl;
	int gflg;
	void (*osig)();

	ctl.maxlen = ilen;
	ctl.len = 0;
	ctl.buf = ibuf;
	gflg = 0;
	if (tmo) {
		if (setjmp(msgjmp)) {
			signal(SIGALRM, osig);
			(*ep->eout)("msgin: timeout");
			return (-1);
		}
		osig = signal(SIGALRM, msgtmo);
		alarm(tmo);
	}
	if (getmsg(ep->fd, &ctl, NULL, &gflg) < 0) {
		(*ep->eout)("msgin: getmsg: %d", strerror(errno));
		return (-1);
	}
	if (tmo) {
		alarm(0);
		signal(SIGALRM, osig);
	}
	return (ctl.len);
}

static void
msgtmo()
{
	longjmp(msgjmp, 1);
}

/*
 * wait for Ethernet device to become ready
 * may take a few seconds for certain devices under certain conditions
 *
 * --nuisance complication:
 * older drivers apparently (this is not well tested)
 * have a persistent idea of which is the current instance
 * for ND_GET/ND_SET calls
 * hence we must remember what it was, change it, and restore it
 */
static int
etherready(ep, d, unit)
Enet ep;
char *d;
int unit;
{
	int i, ans;
	char *p;
	char inst[20];
	char oinst[20];

	snprintf(inst, sizeof(inst), "%d", unit);
	for (i = 0; i < RDYSECS; i++) {
		if (i > 0)
			sleep(1);
		if (unit != NOUNIT) {
			if ((p = ndop(ep->fd, "instance", NULL)) == NULL) {
				(*ep->eout)("%s: warning: cannot get instance", d);
				return (1);
			}
			strncpy(oinst, p, sizeof(oinst));
			if (ndop(ep->fd, "instance", inst) == NULL) {
				(*ep->eout)("%s: warning: cannot set instance %s",
					d, inst);
				return (1);
			}
		}
		if ((p = ndop(ep->fd, "link_status", NULL)) == NULL) {
			(*ep->eout)("%s: warning: cannot get link_status", d);
			return (1);
		}
		ans = atoi(p);
		if (unit != NOUNIT) {
			if (ndop(ep->fd, "instance", oinst) == NULL) {
				(*ep->eout)("%s: warning: cannot reset instance to %s",
					d, oinst);
				return (1);
			}
		}
		if (ans > 0)
			return (1);
	}
	return (-1);
}

static char *
ndop(ep, pname, arg)
Enet ep;
char *pname;
char *arg;
{
	struct strioctl sioc;
	static char buf[ARB];
	char *p, *q, *ebuf;

	memset(buf, 0, sizeof(buf));
	p = buf;
	ebuf = buf + sizeof(buf);
	q = pname;
	while (*q && p < ebuf)
		*p++ = *q++;
	if (p >= ebuf) {
		(*ep->eout)("internal error: ndop overflow");
		return (NULL);
	}
	*p++ = 0;
	if (arg == NULL)
		sioc.ic_cmd = ND_GET;
	else {
		sioc.ic_cmd = ND_SET;
		q = arg;
		while (*q && p < ebuf)
			*p++ = *q++;
		if (p >= ebuf) {
			(*ep->eout)("internal error: ndop overflow");
			return (NULL);
		}
		*p++ = 0;
	}
	sioc.ic_timout = -1;
	sioc.ic_len = sizeof(buf);	/* for return value too */
	sioc.ic_dp = buf;
	if (ioctl(ep->fd, I_STR, &sioc) < 0) {
		(*ep->eout)("I_STR: %s", strerror(errno));
		return (NULL);
	}
	/*
	 * ensure NUL-terminated string (sanity check)
	 */
	for (p = buf; p < ebuf; p++)
		if (*p == 0)
			return (buf);	/* NUL-term, so OK */
	return (NULL);
}

/*
 * set Ethernet device to raw mode
 */
static int
rawether(ep, name)
Enet ep;
char *name;
{
	struct strioctl sioc;

	sioc.ic_cmd = DLIOCRAW;
	sioc.ic_timout = -1;
	sioc.ic_len = 0;
	sioc.ic_dp = NULL;
	if (ioctl(ep->fd, I_STR, &sioc) < 0) {
		(*ep->eout)("%s: DLIOCRAW: %s",
			name, strerror(errno));
		return (-1);
	}
	return (1);
}

int
enetmtu(ep)
Enet ep;
{
	return (ep->mtu);
}

void
enetclose(ep)
Enet ep;
{
	close(ep->fd);
	free(ep);
}

int
enetrecv(ep, buf, len)
Enet ep;
void *buf;
int len;
{
	return (read(ep->fd, buf, len));
}

int
enetsend(ep, buf, len)
Enet ep;
void *buf;
int len;
{
	char *p = buf;

	memcpy(p + ETHER_SRC, ep->mac, LEN_EADDR);
	return (write(ep->fd, buf, len));
}

/*
 * avoid constant eout == NULL checks
 */
static void
nullerr(char *, ...)
{
}
