#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <errno.h>
#include <stdint.h>
#include <stdarg.h>
#include "aoeproto.h"
#include "etherproto.h"
#include "aoedisk.h"
#include "disk.h"
#include "enet.h"
#include "aoedev.h"

static char *progname = "aoedisk";
int verbose;
static char *dtype;
static char *dfile;
static disksize_t dsize;
static char *efile;
static int emtu;
static int eproto = EPROTO_AOE;
static int maj, min;
static int nbufs = 3;
static struct aoeconf aoeconf;

static void doargs(char **);
static int errarg(char *);
static argnum_t scalednum(char *, int);
static unsigned long num(char *);
static void aoeaddr(char *);
static void usage(void);

int
main(argc, argv)
int argc;
char **argv;
{
	char *p;
	Enet ep;
	Disk dp;
	Aoe ap;
	char *buf;
	int alen, rlen;

	if (argc > 0) {
		progname = *argv++;
		if ((p = strrchr(progname, '/')) != NULL)
			progname = p + 1;
	}
	aoeconf.nbuf = 3;
	doargs(argv);
	if (verbose)
		aoeconf.flags |= AF_VERBOSE;
	if (verbose > 1)
		aoeconf.flags |= AF_LOGALL;
	aoeconf.eproto = eproto;
	if ((dp = diskinit(dtype, dfile, dsize, error)) == NULL)
		return (1);
	if ((ep = enetinit(efile, emtu, eproto, error)) == NULL) {
		diskclose(dp);
		return (1);
	}
	if ((alen = enetmtu(ep)) <= 0) {
		error("unreasonably short MTU %d", alen);
		enetclose(ep);
		diskclose(dp);
		return (1);
	}
	alen += LEN_ETHER;
	if ((buf = malloc(alen)) == NULL) {
		error("cannot alloc %ld-byte network buffer", alen);
		enetclose(ep);
		diskclose(dp);
		return (1);
	}
	if ((ap = aoeinit(ep, dp, &aoeconf, error)) == NULL) {
		enetclose(ep);
		diskclose(dp);
		return (1);
	}
	for (;;) {
		if ((rlen = enetrecv(ep, buf, alen)) <= 0) {
			if (rlen == 0)
				error("network read EOF");
			else
				error("network read error: %s",
					strerror(errno));
			return (1);
		}
		aoecmd(ap, buf, rlen);
	}
	/* NOTREACHED */
}

static void
doargs(av)
char **av;
{
	char *p;

	for (; *av && **av == '-'; av++)
	for (p = *av + 1; *p; p++)
	switch (*p) {
	case 'B':
		aoeconf.flags |= AF_NOBCAST;
		continue;

	case 'b':
		aoeconf.nbuf = num(*++av);
		continue;

	case 'c':
		if ((aoeconf.cstr = *++av) == NULL)
			usage();
		continue;

	case 'd':
		if ((dtype = *++av) == NULL)
			usage();
		continue;

	case 'e':
		if (errarg(*++av) <= 0)
			usage();
		continue;

	case 'f':
		if ((dfile = *++av) == NULL)
			usage();
		continue;

	case 'm':
		emtu = num(*++av);
		continue;

	case 'n':
		if ((efile = *++av) == NULL)
			usage();
		continue;

	case 'P':
		aoeconf.flags |= AF_PEDANTIC;
		continue;

	case 'p':
		eproto = num(*++av);
		continue;

	case 's':
		dsize = scalednum(*++av, AOEASEC);
		continue;

	case 'v':
		verbose++;
		continue;

	default:
		usage();
	}
	if (*av == 0)
		usage();
	aoeaddr(*av++);
	if (*av)
		usage();
}

static int
errarg(s)
char *s;
{
	if (s == NULL)
		return (-1);
	if (strcmp(s, "hard") == 0) {
		aoeconf.eflags &=~ (AFE_HARD|AFE_SOFT);
		aoeconf.eflags |= AFE_HARD;
		return (1);
	}
	if (strcmp(s, "soft") == 0) {
		aoeconf.eflags &=~ (AFE_HARD|AFE_SOFT);
		aoeconf.eflags |= AFE_SOFT;
		return (1);
	}
	if (strcmp(s, "ident") == 0) {
		aoeconf.eflags |= AFE_IDENT;
		return (1);
	}
	if (strcmp(s, "drop") == 0) {
		aoeconf.eflags |= AFE_DROP;
		return (1);
	}
	if (strncmp(s, "mod", 3) == 0) {
		s += 3;
		if (*s == '=')
			s++;
		else if (!isdigit(*s))
			return (-1);
		aoeconf.emod = scalednum(s, 0);
		return (1);
	}
	return (-1);
}

static argnum_t
scalednum(s, defscale)
char *s;
int defscale;
{
	char *ep;
	argnum_t n;

	if (s == NULL)
		usage();
	n = strtoul(s, &ep, 0);
	if (ep == s)
		usage();
	for (s = ep; *s; s++)
	switch (*s) {
	case 'b': case 'B':
		n *= AOEASEC;
		defscale = 1;
		continue;

	case 'k': case 'K':
		n *= 1024;
		defscale = 1;
		continue;

	case 'm': case 'M':
		n *= 1024*1024UL;
		defscale = 1;
		continue;

	case 'g': case 'G':
		n *= 1024*1024*1024UL;
		defscale = 1;
		continue;

	case 't': case 'T':
		n *= 1024*1024*1024ULL;
		defscale = 1;
		continue;

	default:
		usage();
	}
	return (n * defscale);
}

static unsigned long
num(s)
char *s;
{
	char *ep;
	unsigned long n;

	if (s == NULL)
		usage();
	n = strtoul(s, &ep, 0);
	if (ep == s || *ep)
		usage();
	return (n);
}

static void
aoeaddr(s)
char *s;
{
	char *ep;
	int n;

	if (!isdigit(*s))
		usage();
	n = strtoul(s, &ep, 0);
	if (ep == s)
		usage();
	switch (*ep) {
	case 0:
		aoeconf.maj = 0;
		aoeconf.min = n;
		break;

	case '.': case '/':
		ep++;
		if (!isdigit(*ep))
			usage();
		aoeconf.maj = n;
		aoeconf.min = strtoul(ep, &ep, 0);
		if (*ep)
			usage();
		break;

	default:
		usage();
	}
	if (aoeconf.maj < 0 || aoeconf.maj >= AOEMAJMAX
	||  aoeconf.min < 0 || aoeconf.min >= AOEMINMAX) {
		error("address out of range: %d.%d\n",
			aoeconf.maj, aoeconf.min);
		usage();
	}
}

static void
usage()
{
	fprintf(stderr, "usage: %s [-vB] [-b nbufs] [-e errtype] [-c cstr] [-p proto] [-n netdev] [-d type] [-f file] [-s disksize] maj.min\n", progname);
	exit(1);
}

void
error(char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	fseek(stderr, 0L, SEEK_END);
	fprintf(stderr, "%s: ", progname);
	vfprintf(stderr, fmt, ap);
	putc('\n', stderr);
	fflush(stderr);
	va_end(ap);
}
