/*
 * Copyright 2005-2007 Coraid, Inc.
 * See accompanying file doc/LICENSE for conditions
 */

/*
 * AoE disk driver
 * maintain various lists of Acmds
 */

/*
 * these includes taken from one of the sample drivers
 * Sun's Writing Device Drivers doesn't mention them,
 * but without at least some of them the driver won't compile
 */
#include <sys/types.h>
#include <sys/file.h>
#include <sys/errno.h>
#include <sys/uio.h>
#include <sys/open.h>
#include <sys/cred.h>
#include <sys/kmem.h>
#include <sys/conf.h>
#include <sys/cmn_err.h>
#include <sys/ksynch.h>
#include <sys/modctl.h>
#include <sys/stat.h>
#include <sys/buf.h>

/*
 * these are mentioned in the book
 */
#include <sys/ddi.h>
#include <sys/sunddi.h>

/*
 * these are needed because it's a disk driver
 * or because we may keep statistics
 */
#include <sys/dkio.h>
#include <sys/vtoc.h>
#include <sys/kstat.h>

/*
 * these are specific to AoE
 */
#include "aoeproto.h"
#include "etherproto.h"	/* for Adrive */
#include "ata.h"	/* for Adrive */
#include "aoed.h"

/*
 * find the pending command corresponding to a reply
 */
Acmd *
adlookup(chan, p)
int chan;
char *p;
{
	Achan *cp;
	Acmd *dp;
	Tag tag;

	if ((cp = ddi_get_soft_state(aoechannels, chan)) == NULL)
		cmn_err(CE_PANIC, "adlookup: aoechan %d no state\n", chan);
	tag = fraoelong(p, AOE_TAG);
	mutex_enter(&cp->lock);
	for (dp = cp->head; dp; dp = dp->next) {
		if (dp->state != CSCHAN)
			cmn_err(CE_PANIC, "aoed: adlookup: cp %p dp %p not CSCHAN\n",
				cp, dp);
		if (dp->tag != tag
		||  dp->ep->chan != chan
		||  dp->ep->minor != fraoechar(p, AOE_MINOR)
		||  dp->ep->major != fraoeshort(p, AOE_MAJOR)
		||  fraoechar(dp->aoehd, AOE_COMMAND) != fraoechar(p, AOE_COMMAND))
			continue;
		/*
		 * found it; unlink
		 */
		dp->state = CSLOOSE;
		if (dp->prev)
			dp->prev->next = dp->next;
		else
			cp->head = dp->next;
		if (dp->next)
			dp->next->prev = dp->prev;
		else
			cp->tail = dp->prev;
		if (dp->ep)
			dp->ep->nact--;
		break;
	}
	mutex_exit(&cp->lock);
	return (dp);
}

/*
 * add command to the pending list for chan
 */
void
adstore(chan, dp)
int chan;
Acmd *dp;
{
	Achan *cp;

	if ((cp = ddi_get_soft_state(aoechannels, chan)) == NULL)
		cmn_err(CE_PANIC, "adstore: aoechan %d no state\n", chan);
	mutex_enter(&cp->lock);
	if (dp->state != CSLOOSE)
		cmn_err(CE_PANIC, "aoed: adstore: cp %p dp %p not CSLOOSE\n",
			cp, dp);
	dp->state = CSCHAN;
	if (cp->tail)
		cp->tail->next = dp;
	else
		cp->head = dp;
	dp->prev = cp->tail;
	cp->tail = dp;
	dp->next = NULL;
	if (dp->ep)
		dp->ep->nact++;
	mutex_exit(&cp->lock);
}

/*
 * remove command from the pending list
 * --used only for error recovery,
 * so who cares if it's slow for safety
 *
 * adremove is useful when all we have in hand is dp and perhaps ep;
 * adcremove when we have cp in hand and locked (e.g. in chantimer)
 */
void
adremove(chan, dp)
int chan;
Acmd *dp;
{
	Achan *cp;
	Acmd *ndp;

	if ((cp = ddi_get_soft_state(aoechannels, chan)) == NULL)
		cmn_err(CE_PANIC, "adremove: aoechan %d no state\n", chan);
	mutex_enter(&cp->lock);
	adcremove(cp, dp);
	mutex_exit(&cp->lock);
}

/*
 * -- caller has cp locked,
 * and promises to restart any walk through
 * the cp->head list
 */
void
adcremove(cp, dp)
Achan *cp;
Acmd *dp;
{
	Acmd *ndp;

	if (dp->state != CSCHAN)
		cmn_err(CE_PANIC, "aoed: adcremove: cp %p dp %p not CSCHAN\n",
			cp, dp);
	dp->state = CSLOOSE;
	for (ndp = cp->head; ndp; ndp = ndp->next)
		if (ndp == dp)
			break;
	if (ndp == NULL)
		return;
	if (ndp->prev)
		ndp->prev->next = ndp->next;
	else
		cp->head = ndp->next;
	if (ndp->next)
		ndp->next->prev = ndp->prev;
	else
		cp->tail = ndp->prev;
	if (ndp->ep)
		ndp->ep->nact--;
}

/*
 * kill any pending commands for drive ep
 */
void
adkilldrive(chan, ep)
int chan;
Adrive *ep;
{
	Achan *cp;
	Acmd *dp, *ndp;

	if ((cp = ddi_get_soft_state(aoechannels, chan)) == NULL)
		cmn_err(CE_PANIC, "adkilldrive: aoechan %d no state\n", chan);
	if (ep->chan != chan)
		cmn_err(CE_PANIC, "adkilldrive: ep %p ->chan %d != arg chan %d\n",
			ep, ep->chan, chan);
	mutex_enter(&cp->lock);
	for (dp = cp->head; dp; dp = ndp) {
		if (dp->state != CSCHAN)
			cmn_err(CE_PANIC, "aoed: adkilldrive: cp %p dp %p not CSCHAN\n",
				cp, dp);
		ndp = dp->next;
		if (dp->ep != ep)
			continue;
		dp->state = CSLOOSE;
		if (dp->prev)
			dp->prev->next = dp->next;
		else
			cp->head = dp->next;
		if (dp->next)
			dp->next->prev = dp->prev;
		else
			cp->tail = dp->prev;
		ep->nact--;
		acfree(dp->ep, dp);
	}
	mutex_exit(&cp->lock);
}

/*
 * kill any pending commands for buffer bp
 */
void
adkillbuf(chan, bp)
int chan;
struct buf *bp;
{
	Achan *cp;
	Acmd *dp, *ndp;

	if ((cp = ddi_get_soft_state(aoechannels, chan)) == NULL)
		cmn_err(CE_PANIC, "adkillbuf: aoechan %d no state\n", chan);
	mutex_enter(&cp->lock);
	for (dp = cp->head; dp; dp = ndp) {
		if (dp->state != CSCHAN)
			cmn_err(CE_PANIC, "aoed: adkillbuf: cp %p dp %p not CSCHAN\n",
				cp, dp);
		ndp = dp->next;
		if (dp->bp != bp)
			continue;
		dp->state = CSLOOSE;
		if (dp->prev)
			dp->prev->next = dp->next;
		else
			cp->head = dp->next;
		if (dp->next)
			dp->next->prev = dp->prev;
		else
			cp->tail = dp->prev;
		if (dp->ep)
			dp->ep->nact--;
		acfree(dp->ep, dp);
	}
	mutex_exit(&cp->lock);
}

/*
 * allocate a new command
 * fill in some constants:
 *	-- allocate a tag
 *	-- set up the common AoE header
 * caller must set up aoelen, command-specific AoE argument,
 * bp if any, remtries, etc
 */
Acmd *
acalloc(ep, aoecmd, hdlen)
Adrive *ep;
int aoecmd, hdlen;
{
	Acmd *dp;

	if (hdlen > sizeof(dp->aoehd)) {
		admsg(NODEV, ep, "acalloc: header len %d refused, max %d\n",
			hdlen, sizeof(dp->aoehd));
		return (NULL);
	}	
	if ((dp = ep->cfree) == NULL)
		return (NULL);
	if (dp->state != CSFREE)
		cmn_err(CE_PANIC, "aoed: acalloc: ep %p dp %p not CSFREE\n",
			ep, dp);
	dp->state = CSLOOSE;
	ep->cfree = dp->next;
	bzero(dp, sizeof(Acmd));
	dp->ep = ep;
	if ((dp->tag = ep->nexttag++ &~ TAGOFFSET) == 0)
		dp->tag = ep->nexttag++ &~ TAGOFFSET;
	dp->aoelen = hdlen;
	toaoechar(dp->aoehd, AOE_VF, aoemkvf(0));
	toaoechar(dp->aoehd, AOE_COMMAND, aoecmd);
	toaoeshort(dp->aoehd, AOE_MAJOR, ep->major);
	toaoechar(dp->aoehd, AOE_MINOR, ep->minor);
	toaoelong(dp->aoehd, AOE_TAG, dp->tag);
	return (dp);
}

/*
 * free a command
 * -- beware the sanity check on dp->ep
 */
void
acfree(ep, dp)
Adrive *ep;
Acmd *dp;
{

	if (dp->ep != ep)
		cmn_err(CE_PANIC, "acfree dp %p dp->ep %p != ep %p",
			dp, dp->ep, ep);
	if (dp->state != CSLOOSE)
		cmn_err(CE_PANIC, "aoed: acfree: ep %p dp %p not CSLOOSE\n",
			ep, dp);
	dp->state = CSFREE;
	dp->next = ep->cfree;
	ep->cfree = dp;
	if (ep->closing && ep->nact <= 0)
		cv_broadcast(&ep->cv);
}

/*
 * allocate maxbuf commands
 */
void
acinit(ep, maxbuf)
Adrive *ep;
int maxbuf;
{
	Acmd *dp;
	int needed;

	if (maxbuf > ep->maxmaxbuf)
		maxbuf = ep->maxmaxbuf;
	needed = maxbuf - ep->maxbuf;
	if (needed < 0)
		cmn_err(CE_PANIC, "%s%d: shrinking buffer count: %d -> %d\n",
			adrvname, ep->instance, ep->maxbuf, maxbuf);
	while (--needed >= 0) {
		if ((dp = kmem_zalloc(sizeof(Acmd), KM_NOSLEEP)) == NULL)
			break;
		dp->next = ep->cfree;
		dp->state = CSFREE;
		ep->cfree = dp;
		ep->maxbuf++;
	}
}

/*
 * discard any commands in ep
 * called on close, or after an error during setup
 */
void
acclear(ep)
Adrive *ep;
{
	Acmd *dp;

	while ((dp = ep->cfree) != NULL) {
		ep->cfree = dp->next;
		kmem_free(dp, sizeof(Acmd));
	}
	ep->maxbuf = 0;
}
