/*
 * Copyright 2007 Coraid, Inc.
 * See accompanying file doc/LICENSE for conditions
 */

/*
 * AoE disk driver
 * code to handle DOS (fdisk) labels
 * segregated because of complicated OS dependencies
 */

/*
 * these includes taken from one of the sample drivers
 * Sun's Writing Device Drivers doesn't mention them,
 * but without at least some of them the driver won't compile
 */
#include <sys/types.h>
#include <sys/file.h>
#include <sys/errno.h>
#include <sys/uio.h>
#include <sys/open.h>
#include <sys/cred.h>
#include <sys/kmem.h>
#include <sys/conf.h>
#include <sys/ksynch.h>
#include <sys/modctl.h>
#include <sys/stat.h>
#include <sys/buf.h>

/*
 * these are mentioned in the book
 */
#include <sys/ddi.h>
#include <sys/sunddi.h>

/*
 * these are needed because it's a disk driver
 * or because we may keep statistics
 */
#include <sys/dkio.h>
#include <sys/vtoc.h>
#include <sys/kstat.h>
#include <sys/efi_partition.h>

/*
 * these are specific to AoE
 */
#include "etherproto.h"		/* for Adrive */
#include "ata.h"		/* for Adrive */
#include "aoed.h"
#include "dospart.h"

static int readdos(Adrive *, diskaddr_t, char *);

/*
 * read DOS partition table
 * only primary partitions
 * if there is at least one partition
 * that might contain a VTOC label,
 * record the highest-numbered one's sizes
 * for later inspection
 */
int
fetchdos(ep)
Adrive *ep;
{
	char bsec[LEN_BSEC];
	char *p;
	int i;
	diskaddr_t base, len, cs;

	if (readdos(ep, BSEC_LOC, bsec) <= 0)
		return (0);
	mutex_enter(&ep->lock);
	for (i = 0, p = bsec + BSEC_PARTS; i < NDOSPART; i++, p += LEN_DOSP) {
		base = frdosplong(p, DOSP_STLBA);
		len = frdosplong(p, DOSP_LBASZ);
		ep->ptab[FIRSTDOSPART + i].base = base;
		ep->ptab[FIRSTDOSPART + i].len = len;
		switch (frdospchar(p, DOSP_SYSID)) {
		case SYSID_SUNOS:
		case SYSID_SUNOS2:
			ep->slaboff = base;
			ep->slabsize = len;
			break;
		}
	}
	if (ep->slabsize > 0
	&&  (cs = ep->geom.dkg_nhead * ep->geom.dkg_nsect) > 0) {
		/*
		 * fake geometry to fit VTOC partition
		 */
		ep->geom.dkg_pcyl = ep->slabsize / cs;
		ep->geom.dkg_acyl = DK_ACYL;
		if (ep->geom.dkg_pcyl >= ep->geom.dkg_acyl)
			ep->geom.dkg_acyl = 0;
		ep->geom.dkg_ncyl = ep->geom.dkg_pcyl - ep->geom.dkg_acyl;
	}
	ep->vflags |= AVLABDOS;
	mutex_exit(&ep->lock);
	return (1);
}

/*
 * expunge any DOS labels on this drive
 * called when some other sort is written
 */
void
zapdos(ep)
Adrive *ep;
{
	char bsec[LEN_BSEC];

	mutex_enter(&ep->lock);
	if (ep->slabsize > 0) {
		/*
		 * might be rewriting encapsulated VTOC;
		 * don't damage the capsule!
		 */
		mutex_exit(&ep->lock);
		return;
	}
	ep->vflags &=~ AVLABDOS;
	mutex_exit(&ep->lock);
	if (readdos(ep, BSEC_LOC, bsec) > 0)
		zeroblock(ep, BSEC_LOC);
}

/*
 * read DOS boot sector into buf;
 * make a few simple validity checks;
 * return > 0 if OK, 0 if invalid, < 0 if can't read
 */
static int
readdos(ep, bno, buf)
Adrive *ep;
diskaddr_t bno;
char *buf;
{
	int e;
	long n;

	e = adrwkern(ep, B_READ, bno, buf, LEN_BSEC);
	if (e != 0) {
		admsg(NODEV, ep, "sector %lld: error (%d) reading DOS disk label\n",
			bno, e);
		return (-1);
	}
	if (frdospword(buf, BSEC_MAGIC) != BSECMAGIC)
		return (0);
	return (1);
}

int
dosioc(ep, cmd, arg, mode)
Adrive *ep;
int cmd;
intptr_t arg;
int mode;
{
	int err;
	char sec[LEN_BSEC];

	switch (cmd) {
	case DKIOCGMBOOT:
		mutex_exit(&ep->lock);
		err = adrwkern(ep, B_READ, BSEC_LOC, sec, LEN_BSEC);
		mutex_enter(&ep->lock);
		if (err != 0)
			return (err);
		if (ddi_copyout(sec, (caddr_t)arg, LEN_BSEC, mode) < 0)
			return (EFAULT);
		return (0);

	case DKIOCSMBOOT:
		if (ddi_copyin((caddr_t)arg, sec, LEN_BSEC, mode) < 0)
			return (EFAULT);
		if (frdospword(sec, BSEC_MAGIC) != BSECMAGIC)
			return (EINVAL);
		mutex_exit(&ep->lock);
		zaplabels(ep, AVLABDOS);
		err = adrwkern(ep, B_WRITE, BSEC_LOC, sec, LEN_BSEC);
		mutex_enter(&ep->lock);
		if (err != 0)
			return (err);
		return (0);

	default:
		return (ENOTTY);
	}
}
