/*
 * Copyright 2006 Coraid, Inc.
 * See accompanying file doc/LICENSE for conditions
 */

/*
 * AoE disk driver
 * ioctls that read EFI labels;
 * segregated because of complicated OS dependencies
 */

/*
 * these includes taken from one of the sample drivers
 * Sun's Writing Device Drivers doesn't mention them,
 * but without at least some of them the driver won't compile
 */
#include <sys/types.h>
#include <sys/file.h>
#include <sys/errno.h>
#include <sys/uio.h>
#include <sys/open.h>
#include <sys/cred.h>
#include <sys/kmem.h>
#include <sys/conf.h>
#include <sys/ksynch.h>
#include <sys/modctl.h>
#include <sys/stat.h>
#include <sys/buf.h>

/*
 * these are mentioned in the book
 */
#include <sys/ddi.h>
#include <sys/sunddi.h>

/*
 * these are needed because it's a disk driver
 * or because we may keep statistics
 */
#include <sys/dkio.h>
#include <sys/vtoc.h>
#include <sys/kstat.h>
#include <sys/efi_partition.h>

/*
 * these are specific to AoE
 */
#include "etherproto.h"		/* for Adrive */
#include "ata.h"		/* for Adrive */
#include "aoed.h"
#include "efilabel.h"

static int readefi(Adrive *, diskaddr_t, char *);
static int efigetset(Adrive *, int, intptr_t, int);
static int efipart(Adrive *, int, intptr_t, int);

/*
 * read EFI partition table
 */
int
fetchefi(ep)
Adrive *ep;
{
	char gpt[LEN_GPT];
	int npart, gplen;
	char *gpbuf, *gpe;
	diskaddr_t bno;
	int i;

	if (readefi(ep, EFI_GPT_LOC, gpt) <= 0) {
		if (readefi(ep, ep->nblocks - 1, gpt) <= 0)
			return (0);
		admsg(NODEV, ep, "primary EFI label invalid, using backup\n");
	}
	npart = frefilong(gpt, GPT_NUMBEROFPARTITIONENTRIES);
	if (npart > NSOLPART)
		npart = NSOLPART;
	gplen = npart * LEN_GPE;
	gplen += DEV_BSIZE - 1;		/* round up to even sector count */
	gplen /= DEV_BSIZE;
	gplen *= DEV_BSIZE;
	if ((gpbuf = kmem_alloc(gplen, KM_NOSLEEP)) == NULL)
		return (0);
	bno = frefill(gpt, GPT_PARTITIONENTRYLBA);
	if ((i = adrwkern(ep, B_READ, bno, gpbuf, gplen)) != 0) {
		admsg(NODEV, ep, "sector %lld: error (%d) reading EFI partition table\n",
			bno, i);
		kmem_free(gpbuf, gplen);
		return (-1);
	}
	mutex_enter(&ep->lock);
	for (i = 0, gpe = gpbuf; i < npart; i++, gpe += LEN_GPE) {
		ep->ptab[FIRSTSOLPART + i].base = frefill(gpe, GPE_STARTINGLBA);
		ep->ptab[FIRSTSOLPART + i].len = frefill(gpe, GPE_ENDINGLBA) + 1
			- ep->ptab[FIRSTSOLPART + i].base;
	}
	ep->vflags |= AVLABEFI;
	mutex_exit(&ep->lock);
	kmem_free(gpbuf, gplen);
	return (1);
}

/*
 * expunge any EFI labels on this drive
 * called when some other sort is written
 */
void
zapefi(ep)
Adrive *ep;
{
	char gpt[LEN_GPT];

	mutex_enter(&ep->lock);
	ep->vflags &=~ AVLABEFI;
	mutex_exit(&ep->lock);
	if (readefi(ep, EFI_GPT_LOC, gpt) > 0)
		zeroblock(ep, EFI_GPT_LOC);
	if (readefi(ep, ep->nblocks - 1, gpt) > 0)
		zeroblock(ep, ep->nblocks - 1);
	mutex_enter(&ep->lock);
	ep->vflags &=~ AVLABEFI;
	mutex_exit(&ep->lock);
}

/*
 * read EFI GPT into buf;
 * make a few simple validity checks;
 * return > 0 if OK, 0 if not valid EFI, < 0 if can't read
 */
static int
readefi(ep, bno, buf)
Adrive *ep;
diskaddr_t bno;
char *buf;
{
	int e;
	long n;

	e = adrwkern(ep, B_READ, bno, buf, LEN_GPT);
	if (e != 0) {
		admsg(NODEV, ep, "sector %lld: error (%d) reading %s EFI disk label\n",
			bno, e, bno == EFI_GPT_LOC ? "primary" : "backup");
		return (-1);
	}
	if (frefill(buf, GPT_SIGNATURE) != EFI_SIGNATURE)
		return (0);
	n = frefilong(buf, GPT_HEADERSIZE);
	if (n < MINLEN_GPT || n > LEN_GPT)
		return (0);
	if (frefilong(buf, GPT_NUMBEROFPARTITIONENTRIES) <= 0)
		return (0);
	if (frefilong(buf, GPT_SIZEOFPARTITIONENTRY) != LEN_GPE)
		return (0);
	if (frefill(buf, GPT_FIRSTUSABLELBA) > frefill(buf, GPT_LASTUSABLELBA))
		return (0);
	return (1);
}

int
efiioc(ep, cmd, arg, mode)
Adrive *ep;
int cmd;
intptr_t arg;
int mode;
{
	switch (cmd) {
	case DKIOCGETEFI:
	case DKIOCSETEFI:
		return (efigetset(ep, cmd, arg, mode));

	case DKIOCPARTITION:
		return (efipart(ep, cmd, arg, mode));

	default:
		return (ENOTTY);
	}
}

/*
 * old_dk_efi belongs in a header file
 */
struct old_dk_efi {
	diskaddr_t	lba;	/* starting block */
	len_t		length;	/* length in bytes */
	caddr32_t	data;	/* 32-bit address */
};

static int
efigetset(ep, cmd, arg, mode)
Adrive *ep;
int cmd;
intptr_t arg;
int mode;
{
	int rw, err;
	dk_efi_t ea;
	struct old_dk_efi oldea;

	switch (cmd) {
	case DKIOCGETEFI:
		rw = B_READ;
		break;

	case DKIOCSETEFI:
		rw = B_WRITE;
		break;

	default:
		return (EINVAL);
	}
	/*
	 * old systems use a slightly different structure,
	 * but only in 32-bit mode
	 */
	if ((ep->cflags & ACOLDDKEFI) == 0
	||  (mode & FMODELS) == DDI_MODEL_LP64) {
		if (ddi_copyin((caddr_t)arg, &ea, sizeof(ea), mode) < 0)
			return (EFAULT);
		ea.dki_data = (void *)ea.dki_data_64;
	} else {
		if (ddi_copyin((caddr_t)arg, &oldea, sizeof(oldea), mode) < 0)
			return (EFAULT);
		ea.dki_lba = oldea.lba;
		ea.dki_length = oldea.length;
		ea.dki_data = (void *)oldea.data;
	}
	if ((ea.dki_length % DEV_BSIZE) != 0)
		return (EINVAL);
	mutex_exit(&ep->lock);
	err = adrwuser(ep, rw, ea.dki_lba,
		(caddr_t)ea.dki_data, ea.dki_length, NULL);
	mutex_enter(&ep->lock);
	if (err == 0 && cmd == DKIOCSETEFI) {
		ep->vflags &=~ AVLABEL;
		ep->vflags |= AVLABEFI;
		mutex_exit(&ep->lock);
		zaplabels(ep, AVLABEFI);
		mutex_enter(&ep->lock);
	}
	return (err);
}

/*
 * DKIOCPARTITION:
 * fetch partition info
 * must go to the disk to get uuid
 */

#define	GPEPERBLOCK	(DEV_BSIZE/LEN_GPE)

static int
efipart(ep, cmd, arg, mode)
Adrive *ep;
int cmd;
intptr_t arg;
int mode;
{
	struct partition64 p64;
	char gpt[LEN_GPT];
	char ebuf[DEV_BSIZE];
	char *gpe;
	diskaddr_t bno;
	int pno;
	int e;

	if (ddi_copyin((caddr_t)arg, &p64, sizeof(p64), mode) < 0)
		return (EFAULT);
	if ((pno = p64.p_partno) < 0)
		return (EINVAL);
	mutex_exit(&ep->lock);		/* readefi does I/O */
	if (readefi(ep, EFI_GPT_LOC, gpt) < 0
	&&  readefi(ep, ep->nblocks - 1, gpt) <= 0) {
		mutex_enter(&ep->lock);
		return (ENXIO);
	}
	if (pno >= frefilong(gpt, GPT_NUMBEROFPARTITIONENTRIES)) {
		mutex_enter(&ep->lock);
		return (ESRCH);
	}
	bno = frefill(gpt, GPT_PARTITIONENTRYLBA);
	bno += pno / GPEPERBLOCK;
	pno %= GPEPERBLOCK;
	if ((e = adrwkern(ep, B_READ, bno, ebuf, sizeof(ebuf))) != 0) {
		mutex_enter(&ep->lock);
		return (e);
	}
	mutex_enter(&ep->lock);
	gpe = ebuf + (pno * LEN_GPE);
	bcopy(gpe + GPE_PARTITIONTYPEGUID, &p64.p_type, LEN_UUID);
	p64.p_start = frefill(gpe, GPE_STARTINGLBA);
	p64.p_size = frefill(gpe, GPE_ENDINGLBA) + 1 - p64.p_start;
	if (ddi_copyout(&p64, (caddr_t)arg, sizeof(p64), mode) < 0)
		return (EFAULT);
	return (0);
}
