/*
 * Copyright 2005-2007 Coraid, Inc.
 * See accompanying file doc/LICENSE for conditions
 */

/*
 * AoE disk driver
 * init device, start I/O, receive replies
 */

/*
 * these includes taken from one of the sample drivers
 * Sun's Writing Device Drivers doesn't mention them,
 * but without at least some of them the driver won't compile
 */
#include <sys/types.h>
#include <sys/file.h>
#include <sys/errno.h>
#include <sys/uio.h>
#include <sys/open.h>
#include <sys/cred.h>
#include <sys/kmem.h>
#include <sys/conf.h>
#include <sys/cmn_err.h>
#include <sys/ksynch.h>
#include <sys/modctl.h>
#include <sys/stat.h>
#include <sys/buf.h>

/*
 * these are mentioned in the book
 */
#include <sys/stream.h>
#include <sys/ddi.h>
#include <sys/sunddi.h>

/*
 * these are needed because it's a disk driver
 * or because we may keep statistics
 */
#include <sys/dkio.h>
#include <sys/vtoc.h>
#include <sys/kstat.h>

/*
 * these are specific to AoE
 */
#include "aoeproto.h"
#include "aoecomm.h"
#include "aoecommproto.h"	/* for aoecomm_devenab commands */
#include "etherproto.h"
#include "ata.h"
#include "aoed.h"

#define	NOSECTOR	((diskaddr_t)~0)

#define	ARB	100	/* arbitrary-length messages */

static void chantimer(void *);
static int expirecmd(Acmd *);
static void adreceive(int, mblk_t *);
static int adqirsp(int, mblk_t *, Acmd *);
static int adatarsp(int, mblk_t *, Acmd *);
static void copyidstr(char *, char *, int);
static int adresperr(char *, Acmd *, char *);
static void endstats(Adrive *, struct buf *, int);
static void reporterr(dev_t, Adrive *, int, diskaddr_t, char *, char *);
static int adstbuf(Adrive *, struct buf *);
static int adstcmd(Acmd *);
static void adlog(Acmd *, int);
static void adbufdone(struct buf *, Adrive *, int);
static mblk_t *enetallocb(int);
static mblk_t *pullup(mblk_t *, int);
static int msgtobuf(mblk_t *, int, caddr_t, int);

/*
 * start the initialization process
 * for a drive in ADCLOSED state:
 * connect to aoecomm; start the channel timer if necessary;
 * send Query-Config broadcast
 * ep locked by caller
 */
int
driveinit(ep, ndelay)
Adrive *ep;
int ndelay;
{
	int i;
	Achan *cp;
	Acmd *dp;

	/*
	 * force MAC address to broadcast (all ones)
 	 * will be set to the real one when
	 * Query-Config response arrives
	 */
	for (i = 0; i < LEN_EADDR; i++)
		ep->mac[i] = 0xff;
	/*
	 * if no-delay open, that's it until openwait called
	 */
	if (ndelay)
		return (1);
	/*
	 * check that drive is enabled
	 * init channel
	 * send Query-Config to locate drive
	 */
	if (aoecomm_devenab(ep->chan, ep->major, ep->minor, ADQUERY) == 0)
		return (-1);
	if (aoecomm_initdriver(ep->chan, adreceive) <= 0)
		return (-1);
	if ((cp = ddi_get_soft_state(aoechannels, ep->chan)) == NULL)
		return (-1);
	mutex_enter(&cp->lock);
	if (cp->tid == 0)
		cp->tid = timeout(chantimer, cp, TMO_CLOCK);
	mutex_exit(&cp->lock);
	if (ep->maxbuf <= 0)
		acinit(ep, 1);
	if ((dp = acalloc(ep, AOECID, MINLEN_AOEI)) == NULL)
		return (-1);
	dp->maxtries = RETRY_QC;
	dp->exptime = TMO_QC;
	dp->cvp = &ep->cv;
	toaoechar(dp->aoehd, AOEI_VCC, aoeimkvcc(0, AOEICGET));
	(void) adstcmd(dp);
	ep->state = ADWAOE;
	return (1);
}

/*
 * clock tick, per channel,
 * to clear up stale requests
 * some subtleties:
 * --must unlock before expiry because
 * doing so must lock ep, and to avoid deadlock
 * that may not be done while cp is locked
 * --once we have expired a command, command list may have
 * changed underfoot whether dp remains valid or not,
 * so must start over from beginning of list
 * --if command was retried rather than killed,
 * stop for TMO_CLOCKHOLD to avoid re-sending back-to-back
 * commands to the same channel, since timeouts are most likely
 * to be caused by network congestion--don't make it worse
 */
static void
chantimer(acp)
void *acp;
{
	Achan *cp = acp;
	Acmd *dp;
	Adrive *ep;
	clock_t now;
	clock_t interval;

	interval = TMO_CLOCK;
	mutex_enter(&cp->lock);
	for (;;) {
		if (cp->tid == 0)	/* shutting down */
			break;
		now = ddi_get_lbolt();
		for (dp = cp->head; dp; dp = dp->next)
			if (dp->expires != 0 && dp->expires <= now)
				break;
		if (dp == NULL)		/* none expiring */
			break;
		adcremove(cp, dp);
		mutex_exit(&cp->lock);
		if (expirecmd(dp) == 0)
			interval = TMO_CLOCKHOLD; /* rexmt: pause */
		mutex_enter(&cp->lock);
	}
	if (cp->tid)	/* in case cleared while expiring a command */
		cp->tid = timeout(chantimer, cp, interval);
	mutex_exit(&cp->lock);
}

/*
 * dp has expired:
 * retry (return 0) or timeout (return 1)
 * assumption:
 *	if bp nonzero, this is an IO operation
 *	if not, it's part of init, so wake up ep->cv
 *
 * dp is no longer on cp->head;
 * if we retry it must be put back (e.g. by adstcmd)
 *
 * don't worry if the command cannot be sent;
 * let the next timeout catch it
 */
static int
expirecmd(dp)
Acmd *dp;
{
	Adrive *ep;
	struct buf *bp;
	clock_t now;
	kcondvar_t *cvp;

	if ((ep = dp->ep) == NULL)
		cmn_err(CE_PANIC, "%s: expirecmd dp %p ep NULL\n",
			adrvname, dp);
	mutex_enter(&ep->lock);
	if (dp->ep != ep)
		cmn_err(CE_PANIC, "%s: expirecmd dp %p ep %p != ep->ep %p\n",
					adrvname, dp, ep, dp->ep);
	if (++dp->ntries < dp->maxtries) {
		(void) adstcmd(dp);
		if (ep->knp)
			ADKS(ep, ADKS_RETRIES)++;
		mutex_exit(&ep->lock);
		return (0);
	}
	adlog(dp, ACLTIMEOUT);
	cvp = dp->cvp;
	if ((bp = dp->bp) == NULL) {
		admsg(NODEV, ep, "AoE control operation timeout\n");
		if (ep->knp)
			ADKS(ep, ADKS_ERRTIMEOUT)++;
		acfree(ep, dp);		/* already off cp->head */
		acclear(ep);
		ep->state = ADCLOSED;
	} else {
		reporterr(NODEV, ep, bp->b_flags, dp->bno,
			"timeout", NULL);
		bioerror(bp, ETIME);
		bp->b_resid = bptoabuf(bp)->bcount;
		if (bp->av_forw)
			bp->av_forw->av_back = bp->av_back;
		else
			ep->lact = bp->av_back;
		if (bp->av_back)
			bp->av_back->av_forw = bp->av_forw;
		else
			ep->fact = bp->av_forw;
		acfree(ep, dp);		/* already off cp->head */
		adkillbuf(ep->chan, bp);
		adbufdone(bp, ep, 1);
		if (ep->knp)
			ADKS(ep, ADKS_ERRTIMEOUT)++;
		adstart(ep);
	}
	if (cvp)
		cv_broadcast(cvp);
	mutex_exit(&ep->lock);
	return (1);
}

/*
 * receive a message on the given channel
 * called by aoecomm, from a service procedure;
 * it is believed this cannot happen concurrently
 * for the same channel
 *
 * per-AoE-command routines should return 
 * >=0 if the message was accepted, and dp is all done
 * < 0 if the message was garbled, meaning dp should be made pending again
 */
static void
adreceive(chan, mp)
int chan;
mblk_t *mp;
{
	Achan *cp;
	char *p;
	int n;
	Acmd *dp;

	if (mp == NULL) {
		if ((cp = ddi_get_soft_state(aoechannels, chan)) == NULL)
			return;
		mutex_enter(&cp->lock);
		cp->flag &=~ ACLOPEN;
		mutex_exit(&cp->lock);
		return;
	}
	if (msgdsize(mp) < LEN_ETHER + MINLEN_AOE) {
		aoecomm_log(chan, ACLILLFORM, mp);
		return;
	}
	if ((mp = pullup(mp, LEN_ETHER + MINLEN_AOE)) == NULL)
		return;		/* oops */
	p = (char *)mp->b_rptr + LEN_ETHER;
	n = fraoechar(p, AOE_VF);
	if (aoeversion(n) != AOEV1) {
		aoecomm_log(chan, ACLILLFORM, mp);
		return;		/* invalid version */
	}
	if ((n & AOEF_R) == 0		/* not response */
	||  (dp = adlookup(chan, p)) == NULL) {	/* unknown tag */
		aoecomm_log(chan, ACLUNSOL, mp);
		return;
	}
	switch (fraoechar(p, AOE_COMMAND)) {
	case AOECATA:
		n = adatarsp(chan, mp, dp);
		break;

	case AOECID:
		n = adqirsp(chan, mp, dp);
		break;

	default:
		break;
	}
	if (n < 0)
		adstore(chan, dp);	/* garbled answer is no answer */
}

static int
adqirsp(chan, mp, dp)
int chan;
mblk_t *mp;
Acmd *dp;
{
	char *p;
	Adrive *ep;
	Tag tag;
	int n, m;
	char xbuf[ARB];

	if ((mp = pullup(mp, LEN_ETHER + MINLEN_AOEI)) == NULL)
		return (-1);
	ep = dp->ep;
	p = (char *)mp->b_rptr + LEN_ETHER;
	if (fraoechar(p, AOE_VF) & AOEF_E) {
		reporterr(NODEV, ep, 0, NOSECTOR, "AoE error: Query-Config",
			aoeerrstr(fraoechar(p, AOE_ERROR), xbuf, sizeof(xbuf)));
		aoecomm_log(chan, ACLAOEERR, mp);
		return (-1);
	}
	if ((n = fraoeshort(p, AOEI_MAXBUF)) <= 0) {
		aoecomm_log(chan, ACLILLFORM, mp);
		return (-1);
	}
	mutex_enter(&ep->lock);
	if (ep->state != ADWAOE) {
		mutex_exit(&ep->lock);
		aoecomm_log(chan, ACLUNSOL, mp);
		return (-1);
	}
	/*
	 * solicited response:
	 * store the data we need, then free the message
	 */
	bcopy(mp->b_rptr + ETHER_SRC, ep->mac, LEN_EADDR);
	freemsg(mp);
	acfree(ep, dp);
	if (ep->closing) {	/* aborted somehow */
		mutex_exit(&ep->lock);
		return (1);
	}
	acinit(ep, n);
	if ((ep->cflags & ACHARDMAX) == 0) {
		ep->maxdata = aoecomm_maxdata(ep->chan);
		if (ep->maxdata == 0)
			ep->maxdata = AOEAMAXSEC * AOEASEC;
		n = fraoechar(p, AOEI_MAXSEC) * AOEASEC;
		if (n == 0)
			n = AOEAMAXSEC * AOEASEC;
		if (ep->maxdata > n)
			ep->maxdata = n;
	}
	/*
	 * send ATA IDENTIFY
	 */
	if ((dp = acalloc(ep, AOECATA, MINLEN_AOEA)) == NULL) {
		/*
		 * this ought to be impossible
		 * unless the system is out of memory
		 */
		acclear(ep);
		ep->state = ADCLOSED;
		cv_broadcast(&ep->cv);
		mutex_exit(&ep->lock);
		return (1);
	}
	dp->maxtries = RETRY_IDENTIFY;
	dp->exptime = TMO_IDENTIFY;
	dp->cvp = &ep->cv;
	dp->bno = NOSECTOR;	/* just for error logging */
	toaoechar(dp->aoehd, AOEA_NSEC, ATAIDSECS);
	toaoechar(dp->aoehd, AOEA_CS, ATAIDENTIFY);
	toaoechar(dp->aoehd, AOEA_LBA3, ATAHD_MB1);
	ep->state = ADWATA;
	if (ep->knp)
		ADKS(ep, ADKS_WATA)++;
	(void) adstcmd(dp);
	mutex_exit(&ep->lock);
	return (1);
}

static int
adatarsp(chan, mp, dp)
int chan;
mblk_t *mp;
Acmd *dp;
{
	int ret;
	Adrive *ep;
	char *p;
	struct buf *bp;
	Abuf *xp;
	int n, error;
	clock_t ntm;
	char xbuf[ARB];		/* for arbitrary messages */

	ret = 1;
	if ((mp = pullup(mp, LEN_ETHER + MINLEN_AOEA)) == NULL)
		return (-1);
	ep = dp->ep;
	p = (char *)mp->b_rptr + LEN_ETHER;
	mutex_enter(&ep->lock);
	if (ep->knp)
		ADKS(ep, ADKS_RATA)++;
	switch (ep->state) {
	/*
	 * ADWATA: answer to ATA IDENTIFY command; store answers
	 * -- here and elsewhere, assume new enough to support
	 * LBA addressing, but adapt for LBA48
	 * during initialization, `report' errors
	 * just by letting the messages pass to aoed
	 */
	case ADWATA:
		if (ep->closing) {	/* aborted */
			acfree(ep, dp);
			break;
		}
		if (fraoechar(dp->aoehd, AOEA_CS) != ATAIDENTIFY) {	/* safety */
			ret = -1;
			break;
		}
		if ((fraoechar(p, AOE_VF) & AOEF_E) != 0
		||  (fraoechar(p, AOEA_CS) & ATAST_ERR) != 0) {
			adresperr(p, dp, "IDENTIFY");
			if (fraoechar(p, AOE_VF) & AOEF_E)
				aoecomm_log(chan, ACLAOEERR, mp);
			else
				aoecomm_log(chan, ACLATAERR, mp);
			acfree(ep, dp);
			acclear(ep);
			ep->state = ADCLOSED;
			cv_broadcast(&ep->cv);
			/* NOT ret = -1; we've freed dp */
			break;
		}
		if ((mp = pullup(mp, LEN_ETHER+AOEA_DATA+(ATASEC*ATAIDSECS))) == NULL) {
			reporterr(NODEV, ep, 0, NOSECTOR,
				"ATA IDENTIFY", "pullup failed");
			acfree(ep, dp);
			acclear(ep);
			ep->state = ADCLOSED;
			cv_broadcast(&ep->cv);
			ret = -1;
			break;
		}
		acfree(ep, dp);
		p = (char *)mp->b_rptr + LEN_ETHER + AOEA_DATA;
		if ((frataword(p, ATAID_CAPAB) & ATAC_LBA) == 0) {
			admsg(NODEV, ep, "disk doesn't support LBA mode\n");
			freemsg(mp);
			acclear(ep);
			ep->state = ADCLOSED;
			cv_broadcast(&ep->cv);
			break;
		}
		ep->geom.dkg_acyl = DK_ACYL;	/* defined by Solaris */
		ep->geom.dkg_pcyl = frataword(p, ATAID_LCYL);
		ep->geom.dkg_nhead = frataword(p, ATAID_LHD);
		ep->geom.dkg_nsect = frataword(p, ATAID_LSEC);
		ep->geom.dkg_rpm = 5400;	/* bogus */
		ep->geom.dkg_intrlv = 1;	/* inscrutable */
		if (fratalong(p, ATAID_CMDENAB) & ATAC_LBA48)
			ep->nblocks = fratall(p, ATAID_MAX48LBN);
		else
			ep->nblocks = fratalong(p, ATAID_MAX28LBN);
		if (ep->nblocks > DK_MAX_BLOCKS) {
			/*
			 * fake up just cylsize, to reduce absolute block number
			 * to fit in b_resid, of type size_t
			 * which is assumed to be at least ulong_t
			 */
			ep->cylsize = (ULONG_MAX / ep->nblocks) + 1;
		} else {
			fudgegeom(ep);
			ep->cylsize = ep->geom.dkg_nhead * ep->geom.dkg_nsect;
			ep->geom.dkg_ncyl = ep->geom.dkg_pcyl - ep->geom.dkg_acyl;
		}
		if (ep->kcp) {
			ADKC(ep, ADKC_MBSIZE) = ep->nblocks/1024/(1024/ATASEC);
			ADKC(ep, ADKC_CYLSIZE) = ep->cylsize;
			ADKC(ep, ADKC_MAXDATA) = ep->maxdata;
			ADKC(ep, ADKC_MAXBUF) = ep->maxbuf;
			ADKC(ep, ADKC_TIMEOUT) = drv_hztousec(ep->mintime)/1000;
		}
		ep->ptab[WHOLEDISKPART].base = 0;
		ep->ptab[WHOLEDISKPART].len = ep->nblocks;
		copyidstr(p + ATAID_SERIAL, ep->ataserial, LEN_ATASERIAL);
		copyidstr(p + ATAID_FWVER, ep->atafwver, LEN_ATAFWVER);
		copyidstr(p + ATAID_MODEL, ep->atamodel, LEN_ATAMODEL);
		freemsg(mp);
		ep->state = ADGATA;
		cv_broadcast(&ep->cv);
		break;

	/*
	 * read or write completed
	 * states prior to ADREADY are those
	 * in which we might try to read the partition table
	 */
	case ADGATA:
	case ADWPTAB:
	case ADREADY:
		if ((bp = dp->bp) == NULL
		||  (xp = bptoabuf(bp)) == NULL) {
			/*
			 * some screwup, ours or the sender's
			 * free buffer, log as unsolicited message
			 */
			if (bp) {
				adkillbuf(ep->chan, bp);
				bioerror(bp, ENXIO);
				adbufdone(bp, ep, 1);
			}
			aoecomm_log(chan, ACLUNSOL, mp);
			acfree(ep, dp);
			adstart(ep);
			break;
		}
		error = 0;
		if ((fraoechar(p, AOE_VF) & AOEF_E) != 0) {
			adresperr(p, dp, NULL);
			aoecomm_log(chan, ACLAOEERR, mp);
			error = EIO;
		} else if ((fraoechar(p, AOEA_CS) & (ATAST_ERR|ATAST_CORR)) != 0) {
			adresperr(p, dp, NULL);
			aoecomm_log(chan, ACLATAERR, mp);
			error = EIO;
		} else if ((bp->b_flags & B_READ) == 0)
			freemsg(mp);	/* write: all done */
		else {		/* read: copy data, all done */
			if (msgtobuf(mp, LEN_ETHER + AOEA_DATA,
			    bp->b_un.b_addr + dp->off, dp->len) != dp->len) {
				if (ep->knp)
					ADKS(ep, ADKS_ERRTRUNC)++;
				error = EIO;
			}
			freemsg(mp);
		}
		acfree(ep, dp);
		if (--xp->ncmds < 0)
			cmn_err(CE_PANIC, "adatarsp: too many responses\n");
		if (error != 0
		||  (xp->ncmds <= 0 && xp->nextoff >= xp->bcount)) {
			if (error) {
				bioerror(bp, error);
				adkillbuf(ep->chan, bp);
			}
			if (bp->av_forw)
				bp->av_forw->av_back = bp->av_back;
			else
				ep->lact = bp->av_back;
			if (bp->av_back)
				bp->av_back->av_forw = bp->av_forw;
			else
				ep->fact = bp->av_forw;
			adbufdone(bp, ep, error);
		}
		adstart(ep);	
		break;

	default:
		aoecomm_log(chan, ACLUNSOL, mp);
		ret = -1;
		break;
	}
	mutex_exit(&ep->lock);
	return (ret);
}

/*
 * swab calls are architecture-dependent:
 * will need to be generalized for, e.g.,
 * Solaris x86
 * assumed that q has one extra byte, for the NUL
 */
static void
copyidstr(p, q, qlen)
char *p, *q;
int qlen;
{
	swab(p, q, qlen);
	p = q + qlen;
	*p-- = 0;
	for (; p > q && *p == ' '; p--)
		*p = 0;
	for (p = q; *p && *p == ' '; p++)
		;
	/* trim leading blanks too?  messier */
}

static int
adresperr(p, dp, s)
char *p;
Acmd *dp;
char *s;
{
	dev_t d;
	int f;
	Adrive *ep;
	char xbuf[ARB], ybuf[ARB];

	if (dp->bp) {
		d = dp->bp->b_edev;
		f = dp->bp->b_flags;
	} else {
		d = NODEV;
		f = 0;
	}
	ep = dp->ep;
	if (fraoechar(p, AOE_VF) & AOEF_E) {
		if (s)
			snprintf(ybuf, sizeof(ybuf), "AoE error: %s", s);
		else
			strcpy(ybuf, "AoE error");
		reporterr(d, ep, f, dp->bno, ybuf,
			aoeerrstr(fraoechar(p, AOE_ERROR), xbuf, sizeof(xbuf)));
		if (ep->knp)
			ADKS(ep, ADKS_ERRAOE)++;
		return (1);
	}
	if (fraoechar(p, AOEA_CS) & ATAST_ERR) {
		if (s)
			snprintf(ybuf, sizeof(ybuf), "ATA error: %s", s);
		else
			strcpy(ybuf, "ATA error");
		reporterr(d, ep, f, dp->bno, ybuf,
			ataerrstr(fraoechar(p, AOEA_EF), xbuf, sizeof(xbuf)));
		if (ep->knp)
			ADKS(ep, ADKS_ERRATA)++;
		return (1);
	}
	if (fraoechar(p, AOEA_CS) & ATAST_CORR) {
		if (s)
			snprintf(ybuf, sizeof(ybuf), "ATA warning: %s", s);
		else
			strcpy(ybuf, "ATA warning");
		reporterr(d, ep, f, dp->bno, ybuf, "soft error");
		if (ep->knp)
			ADKS(ep, ADKS_SOFTERR)++;
	}
	return (0);
}

/*
 * transfer finished, with or without error
 */
static void
adbufdone(bp, ep, iserr)
struct buf *bp;
Adrive *ep;
int iserr;
{
	Abuf *xp;

	endstats(ep, bp, iserr);
	xp = bptoabuf(bp);
	biodone(bp);
	if (xp)
		kmem_free(xp, sizeof(Abuf));
	if (ep->closing && ep->fact == NULL && ep->q.av_forw == NULL)
		cv_broadcast(&ep->cv);
}

/*
 * record statistics when a transfer ends
 */
static void
endstats(ep, bp, iserr)
Adrive *ep;
struct buf *bp;
int iserr;
{
	kstat_t *kp;
	int bcount;

	bcount = bptoabuf(bp)->bcount;
	if ((kp = ep->kp) != NULL) {
		kstat_runq_exit(KSTAT_IO_PTR(kp));
		if (iserr == 0) {
			if (bp->b_flags & B_READ) {
				KSTAT_IO_PTR(kp)->nread += bcount;
				KSTAT_IO_PTR(kp)->reads++;
			} else {
				KSTAT_IO_PTR(kp)->nwritten += bcount;
				KSTAT_IO_PTR(kp)->writes++;
			}
		}
	}
	if ((kp = ep->kpp[DEVTOPART(getminor(bp->b_edev))]) != NULL) {
		kstat_runq_exit(KSTAT_IO_PTR(kp));
		if (iserr == 0) {
			if (bp->b_flags & B_READ) {
				KSTAT_IO_PTR(kp)->nread += bcount;
				KSTAT_IO_PTR(kp)->reads++;
			} else {
				KSTAT_IO_PTR(kp)->nwritten += bcount;
				KSTAT_IO_PTR(kp)->writes++;
			}
		}
	}
}

static void
reporterr(dev, ep, rw, bno, s1, s2)
dev_t dev;
Adrive *ep;
int rw;
diskaddr_t bno;
char *s1, *s2;
{
	if (bno == NOSECTOR)
		admsg(dev, ep, "%s%s%s\n",
			s1 ? s1 : "(reporterr null message)",
			s2 ? ":" : "",
			s2 ? s2 : "");
	else
		admsg(dev, ep, "%s LBA %lld %s%s%s\n",
			(rw & B_READ) ? "read" : "write", bno,
			s1 ? s1 : "(reporterr null message)",
			s2 ? ": " : "",
			s2 ? s2 : "");
}

/*
 * start as many more I/O buffers as this drive will allow
 * --ep had better be locked
 *
 * a reminder of how queues are used:
 *	ep->q lists buffers accepted by adstrategy
 *	but not yet started
 *	-- ep->q.av_back is used by our no-disksort simple list,
 *	but not by disksort(9)
 *
 *	ep->fact/lact lists buffers removed from ep->q,
 *	started or about to be started for actual I/O
 *	ep->fact is the most-recently started buffer,
 *	and may have been only partly started
 *	ep->fact->av_forw ... are other buffers
 *	that have been entirely started
 */
void
adstart(ep)
Adrive *ep;
{
	struct buf *bp;
	kstat_t *kp;

	if ((bp = ep->fact) != NULL
	&&  bptoabuf(bp)->nextoff < bptoabuf(bp)->bcount)
		if (adstbuf(ep, bp) == 0)
			return;
	while ((bp = ep->q.av_forw) != NULL) {
		if ((ep->q.av_forw = bp->av_forw) == NULL)
			ep->q.av_back = NULL;
		if ((bp->av_forw = ep->fact) == NULL)
			ep->lact = bp;
		else
			bp->av_forw->av_back = bp;
		bp->av_back = NULL;
		ep->fact = bp;
		bp->b_resid = bp->b_bcount - bptoabuf(bp)->bcount;
		if ((kp = ep->kp) != NULL) {
			kstat_waitq_exit(KSTAT_IO_PTR(kp));
			kstat_runq_enter(KSTAT_IO_PTR(kp));
		}
		if ((kp = ep->kpp[DEVTOPART(getminor(bp->b_edev))]) != NULL) {
			kstat_waitq_exit(KSTAT_IO_PTR(kp));
			kstat_runq_enter(KSTAT_IO_PTR(kp));
		}
		if (adstbuf(ep, bp) == 0)
			return;
	}
}

/*
 * start as many pieces as possible
 * for transfer bp on drive ep
 * return 0 if something was left undone,
 * which means we ran out of resources
 * and there's no point starting any more
 */

static int
adstbuf(ep, bp)
Adrive *ep;
struct buf *bp;
{
	Abuf *xp = bptoabuf(bp);
	int nc, nd, ns;
	Acmd *dp;
	char *p;
	diskaddr_t bno;

	while (xp->nextoff < xp->bcount) {
		if (xp->bno > ATAMAX28LBN) {
			if (bp->b_flags & B_READ)
				nc = ATARDXSEC;
			else
				nc = ATAWRXSEC;
		} else {
			if (bp->b_flags & B_READ)
				nc = ATARDSEC;
			else
				nc = ATAWRSEC;
		}
		if ((dp = acalloc(ep, AOECATA, MINLEN_AOEA)) == NULL)
			return (0);
		dp->maxtries = RETRY_RW;
		dp->exptime = ep->mintime;
		dp->bp = bp;
		dp->off = xp->nextoff;
		dp->bno = xp->bno;
		if ((nd = xp->bcount - dp->off) > ep->maxdata)
			nd = ep->maxdata;
		ns = nd / AOEASEC;
		nd = ns * AOEASEC;	/* truncate to full sector */
		dp->len = nd;
		xp->nextoff += nd;
		ns = nd / AOEASEC;
		xp->bno += ns;
		p = dp->aoehd;
		toaoechar(p, AOEA_NSEC, ns);
		toaoechar(p, AOEA_CS, nc);
		if (bp->b_flags & B_READ)
			nc = 0;
		else
			nc = AOEAFW;
		bno = dp->bno;
		if (bno > ATAMAX28LBN) {
			nc |= AOEAFE;
			toaoechar(p, AOEA_LBA0, ((int)bno & 0xff));
			bno >>= 8;
			toaoechar(p, AOEA_LBA1, ((int)bno & 0xff));
			bno >>= 8;
			toaoechar(p, AOEA_LBA2, ((int)bno & 0xff));
			bno >>= 8;
			toaoechar(p, AOEA_LBA3, ((int)bno & 0xff));
			bno >>= 8;
			toaoechar(p, AOEA_LBA4, ((int)bno & 0xff));
			bno >>= 8;
			toaoechar(p, AOEA_LBA5, ((int)bno & 0xff));
		} else {
			toaoechar(p, AOEA_LBA0, ((int)bno & 0xff));
			bno >>= 8;
			toaoechar(p, AOEA_LBA1, ((int)bno & 0xff));
			bno >>= 8;
			toaoechar(p, AOEA_LBA2, ((int)bno & 0xff));
			bno >>= 8;
			toaoechar(p, AOEA_LBA3, ((int)bno & 0x0f)|ATAHD_MB1|ATAHD_LBA);
		}
		toaoechar(p, AOEA_AF, nc);
		if (ep->knp)
			ADKS(ep, ADKS_WATA)++;
		(void) adstcmd(dp);
		xp->ncmds++;
	}
	return (1);
}

static int
adstcmd(dp)
Acmd *dp;
{
	Adrive *ep;
	struct buf *bp;
	mblk_t *mp;
	char *p;
	int nhdr, ndata;
	clock_t now;
	Achan *cp;

	if ((ep = dp->ep) == NULL) {
		cmn_err(CE_PANIC, "%s adstcmd: dp %p no ep\n",
			adrvname, dp);
		return (-1);	/* pedantic: never reached */
	}
	bp = dp->bp;
	if ((nhdr = dp->aoelen) <= 0) {
		admsg(bp ? bp->b_edev : NODEV, ep,
			"adstcmd: %s: dp %p header len %d\n", dp, nhdr);
		return (-1);
	}
	if (bp == NULL || (bp->b_flags & B_READ) != 0)
		ndata = 0;
	else
		ndata = dp->len;
	if ((mp = enetallocb(nhdr + ndata + LEN_ETHER)) == NULL)
		return (-1);
	p = (char *)mp->b_rptr;
	bcopy(ep->mac, p + ETHER_DST, LEN_EADDR);
	p += LEN_ETHER;
	bcopy(dp->aoehd, p, nhdr);
	if (ndata) {
		p += nhdr;
		bcopy(bp->b_un.b_addr + dp->off, p, ndata);
	}
	now = ddi_get_lbolt();
	if (dp->ntries == 0)
		dp->started = now;
	if (dp->exptime <= 0)
		dp->expires = 0;
	else
		dp->expires = now + dp->exptime * (1 + dp->ntries);
	if ((cp = ddi_get_soft_state(aoechannels, ep->chan)) == NULL)
		return (-1);
	mutex_enter(&cp->lock);
	if ((cp->flag & ACLOPEN) == 0
	&&  aoecomm_initdriver(ep->chan, adreceive) > 0)
		cp->flag |= ACLOPEN;
	mutex_exit(&cp->lock);
	adstore(ep->chan, dp);
	if (aoecomm_send(ep->chan, mp) <= 0) {
		adlog(dp, ACLSENDFAIL);
		if (ep->knp)
			ADKS(ep, ADKS_ERRWMSG)++;
		return (-1);
	}
	return (1);
}

/*
 * generate log message with the given code
 * for the command in dp
 */
static void
adlog(dp, code)
Acmd *dp;
int code;
{
	Adrive *ep;
	mblk_t *mp;
	char *p;

	if ((ep = dp->ep) == NULL || dp->aoelen <= 0)
		return;
	if ((mp = enetallocb(dp->aoelen + LEN_ETHER)) == NULL)
		return;
	p = (char *)mp->b_rptr;
	bzero(p, LEN_ETHER);
	bcopy(ep->mac, p + ETHER_DST, LEN_EADDR);
	bcopy(dp->aoehd, p + LEN_ETHER, dp->aoelen);
	aoecomm_log(ep->chan, code, mp);
}

/*
 * allocate a message that will be used
 * as an Ethernet packet:
 * round the size up to minimum Ethernet size
 * zero the excess
 * --go to extra trouble to allocate a full
 * minimum-sized Ethernet frame in one go,
 * so downstream code won't have to add a
 * data block for padding
 * this is about a 5% speedup in practice
 */
static mblk_t *
enetallocb(len)
int len;
{
	int mlen;
	mblk_t *mp;

	if ((mlen = len) < MINLEN_ETHER)
		mlen = MINLEN_ETHER;
	if ((mp = allocb(mlen, 0)) == NULL)
		return (NULL);
	if (mlen > len)		/* if we added any padding */
		bzero(mp->b_wptr, mlen);
	mp->b_wptr = mp->b_rptr + mlen;
	return (mp);
}

/*
 * utility routine:
 * pull message up only if necessary;
 * if a pullup was done, free original message
 */

static mblk_t *
pullup(mp, len)
mblk_t *mp;
int len;
{
	mblk_t *nmp;

	if (len < 0) {		/* caller wants the whole message */
		if (mp->b_cont == NULL)
			return (mp);
	} else {
		if (mp->b_wptr - mp->b_rptr >= len)
			return (mp);
	}
	nmp = msgpullup(mp, len);
	freemsg(mp);
	return (nmp);
}

/*
 * copy len bytes from message (list) mp, starting at off,
 * to buf
 * return the number of bytes copied,
 * or -1 if none
 */
static int
msgtobuf(mp, off, buf, len)
mblk_t *mp;
int off;
caddr_t buf;
int len;
{
	int n, nc;

	while ((n = mp->b_wptr - mp->b_rptr) < off) {
		off -= n;
		if ((mp = mp->b_cont) == NULL)
			return (-1);	/* can't reach start */
	}
	for (nc = 0; nc < len && mp; buf += n, nc += n) {
		n = mp->b_wptr - (mp->b_rptr + off);
		if (n > len - nc)
			n = len - nc;
		bcopy(mp->b_rptr + off, buf, n);
		off = 0;
		mp = mp->b_cont;
	}
	return (nc);
}
