/*
 * Copyright 2005-2007 Coraid, Inc.
 * See accompanying file doc/LICENSE for conditions
 */

/*
 * AoE disk driver
 * little subroutines
 */

/*
 * these includes taken from one of the sample drivers
 * Sun's Writing Device Drivers doesn't mention them,
 * but without at least some of them the driver won't compile
 */
#include <sys/types.h>
#include <sys/file.h>
#include <sys/errno.h>
#include <sys/uio.h>
#include <sys/open.h>
#include <sys/cred.h>
#include <sys/kmem.h>
#include <sys/conf.h>
#include <sys/cmn_err.h>
#include <sys/ksynch.h>
#include <sys/modctl.h>
#include <sys/stat.h>
#include <sys/buf.h>

/*
 * these are mentioned in the book
 */
#include <sys/ddi.h>
#include <sys/sunddi.h>

/*
 * these are needed because it's a disk driver
 * or because we may keep statistics
 */
#include <sys/dkio.h>
#include <sys/vtoc.h>
#include <sys/kstat.h>

/*
 * these are needed here for unusual reasons
 */
#include <sys/varargs.h>
#include <sys/sunndi.h>
#include <sys/ndi_impldefs.h>
#include <sys/utsname.h>

/*
 * these are specific to AoE
 */
#include "etherproto.h"	/* for Adrive */
#include "ata.h"
#include "aoed.h"

static char *adpfx(dev_t, Adrive *, char *, int);

/*
 * device number to Adrive
 */
Adrive *
devtodrive(dev, s)
dev_t dev;
char *s;
{
	Adrive *ep;
	int i;

	i = DEVTOINST(dev);
	if ((ep = ddi_get_soft_state(aoedrives, i)) == NULL && s != NULL)
		cmn_err(CE_CONT, "?%s%d part %d: %s: no state\n",
			adrvname, i, DEVTOPART(dev), s);
	return (ep);
}

static char *aoeerrs[] = {
	"No error",
	"Invalid command",
	"Invalid argument",
	"Device unavailable",
	"Config string already set",
	"Invalid protocol version"
};
#define	MAXAOEERR	(sizeof(aoeerrs)/sizeof(aoeerrs[0]))

char *
aoeerrstr(e, buf, len)
int e;
char *buf;
int len;
{
	char *s;

	if (buf == NULL)
		return ("--nullbuf--");
	if (e >= 0 && e < MAXAOEERR)
		s = aoeerrs[e];
	else
		s = "(unknown)";
	snprintf(buf, len, "(%d) %s", e, s);
	return (buf);
}

static struct {
	int v;
	char *s;
} atabits[] = {
	ATAER_AMNF,	"AMNF",
	ATAER_TK0NF,	"TK0NF",
	ATAER_ABRT,	"ABRT",
	ATAER_MCR,	"MCR",
	ATAER_IDNF,	"IDNF",
	ATAER_MC,	"MC",
	ATAER_UNC,	"UNC",
	0,		NULL
};

char *
ataerrstr(e, buf, len)
int e;
char *buf;
int len;
{
	int i, n;
	char *p;

	if (buf == NULL)
		return ("--nullbuf--");
	snprintf(buf, len, "0x%02x", e & 0xff);
	p = buf + strlen(buf);
	for (i = 0; atabits[i].v; i++) {
		if ((e & atabits[i].v) == 0)
			continue;
		n = strlen(atabits[i].s);
		if (p + 1 + n + 1 >= buf + len)	/* +1 for ' ', +1 for NUL */
			continue;
		*p++ = ' ';
		strcpy(p, atabits[i].s);
		p += n;
	}
	return (buf);
}

#define	PFXLEN	100	/* long enough for what adpfx prints */

void
admsg(dev_t dev, Adrive *ep, char *fmt, ...)
{
	char pfx[PFXLEN];
	char msg[300];
	va_list ap;

	va_start(ap, fmt);
	vsnprintf(msg, sizeof(msg), fmt, ap);
	va_end(ap);
	cmn_err(CE_CONT, "?%s: %s\n", adpfx(dev, ep, pfx, sizeof(pfx)), msg);
}

/*
 * when gcc was the only available compiler
 * so varargs was not feasible,
 * this was used all over
 * now it's called only from admsg
 * maybe it should be folded in
 */
static char *
adpfx(dev, ep, pfx, len)
dev_t dev;
Adrive *ep;
char *pfx;
int len;
{
	static char xpfx[PFXLEN];
	int p;

	if (pfx == NULL)
		pfx = xpfx;
	if (ep == NULL)
		ep = ddi_get_soft_state(aoedrives, DEVTOINST(dev));
	p = DEVTOPART(dev);
	if (ep == NULL)
		snprintf(pfx, len, "%s%d s%d (unattached)",
			adrvname, DEVTOINST(dev), p);
	else if (dev == NODEV)
		snprintf(pfx, len, "%s%d %d/%d/%d",
			adrvname, ep->instance, ep->chan, ep->major, ep->minor);
	else if (p == WHOLEDISKPART)
		snprintf(pfx, len, "%s%d (whole) %d/%d/%d",
			adrvname, ep->instance, ep->chan, ep->major, ep->minor);
	else
		snprintf(pfx, len, "%s%d s%d %d/%d/%d",
			adrvname, ep->instance, p, ep->chan, ep->major, ep->minor);
	return (pfx);
}

/*
 * partition names used in /devices, kstat names, etc
 * -- Solaris overlays wd on h, creating it only with EFI label;
 * or on q, the `zeroth' DOS partiton.
 * we just keep it separate and offer it all the time,
 * since it makes life simpler to have it available internally
 */
static char *pnames[] = {
	"a", "b", "c", "d", "e", "f", "g", "h",
#if defined(_SUNOS_VTOC_16)
	"i", "j", "k", "l", "m", "n", "o", "p",
#endif
	"wd",
	"r", "s", "t", "u",
};

#define	MAXPNAME	(sizeof(pnames)/sizeof(pnames[0]))

char *
adpname(p)
int p;
{
	if (p < 0 || p >= MAXPNAME)
		return ("???");	/* shouldn't happen */
	return (pnames[p]);
}

/*
 * zap labels of all but the specified type
 * --do VTOC first in case it's embedded
 */
void
zaplabels(ep, t)
Adrive *ep;
int t;
{
	if ((t & AVLABVTOC) == 0)
		zapvtoc(ep);
	if ((t & AVLABDOS) == 0)
		zapdos(ep);
	if ((t & AVLABEFI) == 0)
		zapefi(ep);
}
/*
 * in Solaris 9 and newer, this boils down to
 *    ddi_prop_update_int64(dev, dip, name, val);
 * but since that routine doesn't exist in
 * older systems we fake it up here
 *
 * it is our hope that if run on a pre-update_int64
 * system, this will either quietly create a
 * property of unreachable type,
 * or fail gracefully
 */

static int propllencode(prop_handle_t *, void *, uint_t);

int
putllprop(dev, dip, name, val)
dev_t dev;
dev_info_t *dip;
char *name;
uint64_t val;
{
	return (ddi_prop_update_common(dev, dip,
		DDI_PROP_SYSTEM_DEF|DDI_PROP_TYPE_INT64,
		name, &val, 1, propllencode));
}

/*
 * restricted version for now:
 * encode exactly one integer
 */

#define	I64SIZE	sizeof(int64_t)

static int
propllencode(ph, d, n)
prop_handle_t *ph;
void *d;
uint_t n;
{
	if (n != 1)
		return (DDI_PROP_CANNOT_ENCODE);
	ph->ph_data = kmem_alloc(I64SIZE, KM_SLEEP);
	ph->ph_size = I64SIZE;
	ph->ph_save_pos = ph->ph_data;
	bcopy(d, ph->ph_data, I64SIZE);
	ph->ph_cur_pos = (char *)ph->ph_data + I64SIZE;
	return (DDI_PROP_RESULT_OK);
}

/*
 * which Solaris version is this?
 * -- no structured interface,
 * so this is just a hack
 * to support other hacks
 */
int
solversion()
{
	char *p, *pp;
	int n;

	p = utsname.release;
	if (*p++ != '5' || *p++ != '.')
		return (0);		/* unknown */
	pp = p;
	n = stoi(&pp);
	if (pp == p)
		return (0);		/* unknown */
	return (n);
}
