/*
 * Copyright 2006, 2007 Coraid, Inc.
 * See accompanying file doc/LICENSE for conditions
 */

/*
 * AoE disk driver
 * old-fashioned (VTOC) partition tables
 */

/*
 * these includes taken from one of the sample drivers
 * Sun's Writing Device Drivers doesn't mention them,
 * but without at least some of them the driver won't compile
 */
#include <sys/types.h>
#include <sys/file.h>
#include <sys/errno.h>
#include <sys/uio.h>
#include <sys/open.h>
#include <sys/cred.h>
#include <sys/kmem.h>
#include <sys/conf.h>
#include <sys/ksynch.h>
#include <sys/modctl.h>
#include <sys/stat.h>
#include <sys/buf.h>

/*
 * these are mentioned in the book
 */
#include <sys/ddi.h>
#include <sys/sunddi.h>

/*
 * these are needed because it's a disk driver
 * or because we may keep statistics
 */
#include <sys/dkio.h>
#include <sys/vtoc.h>
#include <sys/kstat.h>

/*
 * these are specific to AoE
 */
#include "etherproto.h"	/* for Adrive */
#include "ata.h"	/* for Adrive */
#include "aoed.h"

/*
 * Sun convention:
 * backup VTOC labels in last track (in last alternate cylinder)
 * at first five odd sectors
 */
#define	BKLOFF	1
#define	BKLSPC	2
#define	NBKL	5

static int readvtoc(Adrive *, diskaddr_t, struct dk_label *);
static int writevtoc(Adrive *);
static int dklvalid(Adrive *, struct dk_label *);
static uint16_t dklsum(struct dk_label *);
static int sanevtoc(Adrive *, struct vtoc *);

/*
 * try to extract real VTOC from disk label
 * -- don't bother looking for backups if
 * encapsulated label; geometry is too uncertain
 */

int
fetchvtoc(ep)
Adrive *ep;
{
	struct dk_label dkl;
	diskaddr_t cs;
	int i;

	if (readvtoc(ep, DK_LABEL_LOC, &dkl) <= 0) {
		mutex_enter(&ep->lock);
		if (ep->geom.dkg_acyl <= 0) {
			mutex_exit(&ep->lock);
			return (-1);
		}
		cs = ep->geom.dkg_ncyl + ep->geom.dkg_acyl;
		cs *= ep->geom.dkg_nhead * ep->geom.dkg_nsect;
		cs -= ep->geom.dkg_nsect;
		cs += BKLOFF;
		mutex_exit(&ep->lock);
		for (i = 0; i < NBKL; cs += BKLSPC, i++)
			if (readvtoc(ep, cs, &dkl) > 0)
				break;
		if (i >= NBKL)
			return (-1);	/* no backup found */
		admsg(NODEV, ep, "primary VTOC label invalid, using backup %d\n",
			i);
	}
	/*
	 * extract vtoc;
	 * overwrite geom
	 */
	mutex_enter(&ep->lock);
	bzero(&ep->vtoc, sizeof(ep->vtoc));
	ep->vtoc.v_version = dkl.dkl_vtoc.v_version;
	bcopy(dkl.dkl_vtoc.v_volume, ep->vtoc.v_volume, LEN_DKL_VVOL);
	bcopy(dkl.dkl_asciilabel, ep->vtoc.v_asciilabel, LEN_DKL_ASCII);
	ep->vtoc.v_nparts = dkl.dkl_vtoc.v_nparts;
#if defined(_SUNOS_VTOC_16)
	ep->vtoc.v_sectorsz = dkl.dkl_vtoc.v_sectorsz;
#else /* _SUNOS_VTOC_8 */
	ep->vtoc.v_sectorsz = ATASEC;
#endif
	cs = dkl.dkl_nhead * dkl.dkl_nsect;
	for (i = 0; i < ep->vtoc.v_nparts; i++) {
#if defined(_SUNOS_VTOC_16)
		ep->vtoc.v_part[i].p_tag = dkl.dkl_vtoc.v_part[i].p_tag;
		ep->vtoc.v_part[i].p_flag = dkl.dkl_vtoc.v_part[i].p_flag;
		ep->vtoc.v_part[i].p_start = dkl.dkl_vtoc.v_part[i].p_start;
		ep->vtoc.v_part[i].p_size = dkl.dkl_vtoc.v_part[i].p_size;
#else /* _SUNOS_VTOC_8 */
		ep->vtoc.v_part[i].p_tag = dkl.dkl_vtoc.v_part[i].p_tag;
		ep->vtoc.v_part[i].p_flag = dkl.dkl_vtoc.v_part[i].p_flag;
		ep->vtoc.v_part[i].p_start = dkl.dkl_map[i].dkl_cylno * cs;
		ep->vtoc.v_part[i].p_size = dkl.dkl_map[i].dkl_nblk;
#endif
		ep->vtoc.timestamp[i] = dkl.dkl_vtoc.v_timestamp[i];
		ep->ptab[FIRSTSOLPART + i].base =
			ep->vtoc.v_part[i].p_start + ep->slaboff;
		ep->ptab[FIRSTSOLPART + i].len =
			ep->vtoc.v_part[i].p_size;
	}
	ep->vtoc.v_sanity = dkl.dkl_vtoc.v_sanity;
	bzero(&ep->geom, sizeof(ep->geom));
	ep->geom.dkg_ncyl = dkl.dkl_ncyl;
	ep->geom.dkg_acyl = dkl.dkl_acyl;
#if defined(_SUNOS_VTOC_16)
	ep->geom.dkg_bcyl = dkl.dkl_bcyl;
#endif
	ep->geom.dkg_nhead = dkl.dkl_nhead;
	ep->geom.dkg_nsect = dkl.dkl_nsect;
	ep->geom.dkg_intrlv = dkl.dkl_intrlv;
	ep->geom.dkg_apc = dkl.dkl_apc;
	ep->geom.dkg_rpm = dkl.dkl_rpm;
	ep->geom.dkg_pcyl = dkl.dkl_pcyl;
	ep->geom.dkg_write_reinstruct = dkl.dkl_write_reinstruct;
	ep->geom.dkg_read_reinstruct = dkl.dkl_read_reinstruct;
	ep->vflags |= AVLABVTOC;
	mutex_exit(&ep->lock);
	return (1);
}

/*
 * clear any extant VTOC labels
 */
void
zapvtoc(ep)
Adrive *ep;
{
	struct dk_label dkl;
	daddr_t cs, off, size;
	int i;

	if (readvtoc(ep, ep->slaboff + DK_LABEL_LOC, &dkl) > 0)
		zeroblock(ep, DK_LABEL_LOC);
	mutex_enter(&ep->lock);
	ep->vflags &=~ AVLABVTOC;
	if (ep->geom.dkg_acyl <= 0
	||  ep->slabsize <= 0) {
		mutex_exit(&ep->lock);
		return;
	}
	cs = ep->geom.dkg_ncyl + ep->geom.dkg_acyl;
	cs *= ep->geom.dkg_nhead * ep->geom.dkg_nsect;
	cs -= ep->geom.dkg_nsect;
	cs += BKLOFF;
	off = ep->slaboff;
	size = ep->slabsize;
	mutex_exit(&ep->lock);
	for (i = 0; i < NBKL; cs += BKLSPC, i++) {
		if (cs < size && readvtoc(ep, cs, &dkl) > 0)
			zeroblock(ep, cs + off);
	}
}

/*
 * try one location for VTOC label
 */

static int
readvtoc(ep, bno, dkp)
Adrive *ep;
diskaddr_t bno;
struct dk_label *dkp;
{
	int e;

	mutex_enter(&ep->lock);
	if (bno >= ep->slabsize) {
		mutex_exit(&ep->lock);
		return (0);
	}
	bno += ep->slaboff;
	mutex_exit(&ep->lock);
	if ((e = adrwkern(ep, B_READ, bno, dkp, sizeof(*dkp))) != 0) {
		admsg(NODEV, ep, "sector %lld: error (%d) reading %s VTOC disk label\n",
			bno, e, bno == DK_LABEL_LOC ? "primary" : "backup");
		return (-1);
	}
	if (dklvalid(ep, dkp) == 0)
		return (0);
	return (1);
}

/*
 * generate fake VTOC
 * fake geometry info was already supplied by drive-init code
 */

#define	FIRSTPART	0
#define	FULLPART	2

int
fakevtoc(ep)
Adrive *ep;
{
	Blkno maxbno;
	int i;
	char model[LEN_ATAMODEL + 1 + LEN_ATAFWVER + 1];
	char *p;
	Part zpart;

	mutex_enter(&ep->lock);
	if ((maxbno = ep->slabsize) <= 0) {
		maxbno = ep->geom.dkg_ncyl;
		maxbno *= ep->geom.dkg_nhead * ep->geom.dkg_nsect;
	}
	if (maxbno > DK_MAX_BLOCKS) {
		mutex_exit(&ep->lock);
		return (0);
	}
	bzero(&zpart, sizeof(zpart));
	for (i = FIRSTSOLPART; i < FIRSTSOLPART + NSOLPART; i++)
		ep->ptab[i] = zpart;
	ep->ptab[FIRSTSOLPART + FIRSTPART].base = ep->slaboff;
	ep->ptab[FIRSTSOLPART + FIRSTPART].len = maxbno;
	ep->ptab[FIRSTSOLPART + FULLPART].base = ep->slaboff;
	ep->ptab[FIRSTSOLPART + FULLPART].len = maxbno;
	bzero(&ep->vtoc, sizeof(ep->vtoc));
	ep->vtoc.v_sanity = VTOC_SANE;
	ep->vtoc.v_version = V_VERSION;
	ep->vtoc.v_sectorsz = ATASEC;
#if NDKMAP < V_NUMPAR
	ep->vtoc.v_nparts = NDKMAP;
#else
	ep->vtoc.v_nparts = V_NUMPAR;
#endif
	ep->vtoc.v_part[FIRSTPART].p_tag = V_ROOT;
	ep->vtoc.v_part[FIRSTPART].p_start = 0;
	ep->vtoc.v_part[FIRSTPART].p_size = maxbno;
	ep->vtoc.v_part[FULLPART].p_tag = V_BACKUP;
	ep->vtoc.v_part[FULLPART].p_start = 0;
	ep->vtoc.v_part[FULLPART].p_size = maxbno;
	/*
	 * model = atamodel "-" atafwver
	 * with trailing blanks excised
	 * and internal blanks replaced with -
	 */
	bcopy(ep->atamodel, model, LEN_ATAMODEL);
	p = model + LEN_ATAMODEL - 1;
	for (; p >= model && *p == ' '; p--)
		;
	p++;
	*p++ = '-';
	bcopy(ep->atafwver, p, LEN_ATAFWVER);
	p += LEN_ATAFWVER - 1;
	for (; p >= model && *p == ' '; p--)
		;
	p[1] = 0;
	for (; p >= model; p--)
		if (*p == ' ' || *p == '\t')
			*p = '-';
	snprintf(ep->vtoc.v_asciilabel, sizeof(ep->vtoc.v_asciilabel),
		"%s cyl %d alt %d hd %d sec %d",
		model,
		ep->geom.dkg_ncyl, ep->geom.dkg_acyl,
		ep->geom.dkg_nhead, ep->geom.dkg_nsect);
	ep->vflags |= AVLABVTOC;
	mutex_exit(&ep->lock);
	return (1);
}

int
vtocioc(ep, cmd, arg, mode)
Adrive *ep;
int cmd;
intptr_t arg;
int mode;
{
	struct vtoc v;
	struct vtoc32 v32;
	int i;

	if ((ep->vflags & AVLABVTOC) == 0)
		return (ENOTSUP);
	switch (cmd) {
	case DKIOCGVTOC:
		switch (mode & FMODELS) {
		case DDI_MODEL_LP64:
			if (ddi_copyout(&ep->vtoc, (caddr_t)arg, sizeof(ep->vtoc), mode) < 0)
				return (EFAULT);
			return (0);
	
		case DDI_MODEL_ILP32:
			vtoctovtoc32(ep->vtoc, v32);
			if (ddi_copyout(&v32, (caddr_t)arg, sizeof(v32), mode) < 0)
				return (EFAULT);
			return (0);
	
		default:
			return (EINVAL);
		}

	case DKIOCSVTOC:
		switch (mode & FMODELS) {
		case DDI_MODEL_LP64:
			if (ddi_copyin((caddr_t)arg, &v, sizeof(v), mode) < 0)
				return (EFAULT);
			break;

		case DDI_MODEL_ILP32:
			if (ddi_copyin((caddr_t)arg, &v32, sizeof(v32), mode) < 0)
				return (EFAULT);
			vtoc32tovtoc(v32, v);
			break;

		default:
			return (EINVAL);
		}
		if (sanevtoc(ep, &v) == 0)
			return (EINVAL);
		bcopy(&v, &ep->vtoc, sizeof(v));
		for (i = 0; i < ep->vtoc.v_nparts && i < NSOLPART; i++) {
			ep->ptab[FIRSTSOLPART + i].base =
				ep->vtoc.v_part[i].p_start + ep->slaboff;
			ep->ptab[FIRSTSOLPART + i].len =
				ep->vtoc.v_part[i].p_size;
		}
		mutex_exit(&ep->lock);
		zaplabels(ep, AVLABVTOC);
		if (writevtoc(ep) <= 0) {
			mutex_enter(&ep->lock);
			return (EIO);
		}
		mutex_enter(&ep->lock);
		return (0);

	default:
		return (ENOTTY);
	}
}

/*
 * update disk partition table from in-core data
 * return > 0 if at least one label was written
 * check sanity only of things that could cause grief right here,
 * for safety; sanevtoc ought to be called before accepting
 * data supplied via ioctl, to check sanity more generally
 */
static int
writevtoc(ep)
Adrive *ep;
{
	struct dk_label dkl;
	int e, i, retok;
	diskaddr_t cs, maxbno;

	bzero(&dkl, sizeof(dkl));
	mutex_enter(&ep->lock);
	if ((ep->vflags & AVLABVTOC) == 0) {
		mutex_exit(&ep->lock);
		return (0);	
	}
	dkl.dkl_vtoc.v_version = ep->vtoc.v_version;
	bcopy(ep->vtoc.v_volume, dkl.dkl_vtoc.v_volume, LEN_DKL_VVOL);
#if defined(_SUNOS_VTOC_16)
	bcopy(ep->vtoc.v_asciilabel, dkl.dkl_vtoc.v_asciilabel, LEN_DKL_ASCII);
#else
	bcopy(ep->vtoc.v_asciilabel, dkl.dkl_asciilabel, LEN_DKL_ASCII);
#endif
	if ((dkl.dkl_vtoc.v_nparts = ep->vtoc.v_nparts) > V_NUMPAR) {
		mutex_exit(&ep->lock);
		return (-1);
	}
#if defined(_SUNOS_VTOC_16)
	dkl.dkl_vtoc.v_sectorsz = ep->vtoc.v_sectorsz;
#endif
	if ((cs = ep->geom.dkg_nhead * ep->geom.dkg_nsect) == 0) {
		mutex_exit(&ep->lock);
		return (-1);
	}
	for (i = 0; i < ep->vtoc.v_nparts; i++) {
#if defined(_SUNOS_VTOC_16)
		dkl.dkl_vtoc.v_part[i].p_tag = ep->vtoc.v_part[i].p_tag;
		dkl.dkl_vtoc.v_part[i].p_flag = ep->vtoc.v_part[i].p_flag;
		dkl.dkl_vtoc.v_part[i].p_start = ep->vtoc.v_part[i].p_start;
		dkl.dkl_vtoc.v_part[i].p_size = ep->vtoc.v_part[i].p_size;
#else
		dkl.dkl_vtoc.v_part[i].p_tag = ep->vtoc.v_part[i].p_tag;
		dkl.dkl_vtoc.v_part[i].p_flag = ep->vtoc.v_part[i].p_flag;
		dkl.dkl_map[i].dkl_cylno = ep->vtoc.v_part[i].p_start/cs;
		dkl.dkl_map[i].dkl_nblk = ep->vtoc.v_part[i].p_size;
#endif
		dkl.dkl_vtoc.v_timestamp[i] = ep->vtoc.timestamp[i];
	}
	dkl.dkl_vtoc.v_sanity = ep->vtoc.v_sanity;
	dkl.dkl_ncyl = ep->geom.dkg_ncyl;
	dkl.dkl_acyl = ep->geom.dkg_acyl;
#if defined(_SUNOS_VTOC_16)
	dkl.dkl_bcyl = ep->geom.dkg_bcyl;
#endif
	dkl.dkl_nhead = ep->geom.dkg_nhead;
	dkl.dkl_nsect = ep->geom.dkg_nsect;
	dkl.dkl_intrlv = ep->geom.dkg_intrlv;
	dkl.dkl_apc = ep->geom.dkg_apc;
	dkl.dkl_rpm = ep->geom.dkg_rpm;
	dkl.dkl_pcyl = ep->geom.dkg_pcyl;
	dkl.dkl_write_reinstruct = ep->geom.dkg_write_reinstruct;
	dkl.dkl_read_reinstruct = ep->geom.dkg_read_reinstruct;
	mutex_exit(&ep->lock);
	dkl.dkl_magic = DKL_MAGIC;
	dkl.dkl_cksum = 0;
	dkl.dkl_cksum = dklsum(&dkl);
	retok = -1;
	e = adrwkern(ep, B_WRITE, ep->slaboff + DK_LABEL_LOC, &dkl, sizeof(dkl));
	if (e == 0)
		retok = 1;
	else
		admsg(NODEV, ep, "sector %lld: error (%d) writing primary VTOC label\n",
			(diskaddr_t)DK_LABEL_LOC, i);
	mutex_enter(&ep->lock);
	if (ep->geom.dkg_acyl <= 0) {
		mutex_exit(&ep->lock);
		return (retok);	/* no reserved cylinder -> no backups */
	}
	cs = ep->geom.dkg_ncyl + ep->geom.dkg_acyl;
	cs *= ep->geom.dkg_nhead * ep->geom.dkg_nsect;
	cs -= ep->geom.dkg_nsect;
	cs += ep->slaboff;
	maxbno = ep->slaboff + ep->slabsize;
	mutex_exit(&ep->lock);
	for (i = BKLOFF; i < BKLOFF + NBKL*BKLSPC; i += BKLSPC) {
		if (cs >= maxbno)
			break;
		e = adrwkern(ep, B_WRITE, cs + i, &dkl, sizeof(dkl));
		if (e == 0)
			retok = 1;
		else
			admsg(NODEV, ep, "sector %lld: error (%d) writing backup VTOC label\n",
				cs + i, e);
	}
	return (retok);
}

static int
dklvalid(ep, lp)
Adrive *ep;
struct dk_label *lp;
{
	daddr_t cs, size;
	int i;
	diskaddr_t base, len;

	mutex_enter(&ep->lock);
	size = ep->slabsize;
	mutex_exit(&ep->lock);
	if (lp->dkl_magic != DKL_MAGIC
	||  dklsum(lp) != 0
	||  lp->dkl_vtoc.v_sanity != VTOC_SANE
	||  lp->dkl_vtoc.v_nparts > V_NUMPAR)
		return (0);
	if (lp->dkl_nsect <= 0
	||  lp->dkl_nhead <= 0
	||  lp->dkl_pcyl <= 0
	||  lp->dkl_ncyl <= 0)
		return (0);
	cs = lp->dkl_nhead * lp->dkl_nsect;
	if (cs * lp->dkl_pcyl > size
	||  lp->dkl_ncyl + lp->dkl_acyl > lp->dkl_pcyl)
		return (0);
	/*
	 * Sun sample driver checks that each partition
	 * begins on a cylinder boundary
	 * for now let's just check that each
	 * is within legal bounds
	 */
	for (i = 0; i < lp->dkl_vtoc.v_nparts; i++) {
#if defined(_SUNOS_VTOC_16)
		base = lp->dkl_vtoc.v_part[i].p_start;
		len = lp->dkl_vtoc.v_part[i].p_size;
#else /* _SUNOS_VTOC_8 */
		base = lp->dkl_map[i].dkl_cylno * cs;
		len = lp->dkl_map[i].dkl_nblk;
#endif
		if (base > ep->slabsize || base + len > size)
			return (1);
	}
	return (1);
}

/*
 * compute the checksum of a disk label
 */
static uint16_t
dklsum(lp)
struct dk_label *lp;
{
	uint16_t sum;
	uint16_t *sp;
	int n;

	sum = 0;
	sp = (uint16_t *)lp;
	n = sizeof(*lp) / sizeof(uint16_t);
	while (--n >= 0)
		sum ^= *sp++;
	return (sum);
}

static int
sanevtoc(ep, vp)
Adrive *ep;
struct vtoc *vp;
{
	Blkno cs, maxbno;
	int i;

	if (vp->v_sanity != VTOC_SANE
	||  vp->v_nparts > V_NUMPAR)
		return (0);
#if !defined(_SUNOS_VTOC_16)
	/*
	 * in old-style disk label,
	 * partitions are stored with cylinder offsets;
	 * oops
	 */
	if ((cs = ep->geom.dkg_nhead * ep->geom.dkg_nsect) == 0)
		return (0);
#endif
	for (i = 0; i < vp->v_nparts; i++) {
#if !defined(_SUNOS_VTOC_16)
		if ((vp->v_part[i].p_start % cs) != 0)
			return (0);
#endif
		if (vp->v_part[i].p_start + vp->v_part[i].p_size > ep->slabsize)
			return (0);
	}
	return (1);
}
