/*
 * Copyright 2005 Coraid, Inc.
 * See accompanying file doc/LICENSE for conditions
 */

/*
 * stream module for AoE communication
 * pushed on the communication channel
 *
 * -- dammit, it know about Ethernet,
 * so it can fill in headers and pad to minimum length
 * there is probably a proper way to make DLPI do that
 */

/*
 * these includes taken from one of the sample drivers
 * Sun's Writing Device Drivers doesn't mention them,
 * but without at least some of them the driver won't compile
 */
#include <sys/types.h>
#include <sys/file.h>
#include <sys/errno.h>
#include <sys/open.h>
#include <sys/cred.h>
#include <sys/kmem.h>
#include <sys/conf.h>
#include <sys/cmn_err.h>
#include <sys/ksynch.h>
#include <sys/modctl.h>
#include <sys/stat.h>
#include <sys/stream.h>

/*
 * these are mentioned in the book
 */
#include <sys/ddi.h>
#include <sys/sunddi.h>

/*
 * these are local to AoE
 */
#include "aoeproto.h"
#include "aoecomm.h"
#include "aoecommproto.h"
#include "etherproto.h"

/*
 * device registry state, per major block
 */

#ifndef WORD_BIT
#define	WORD_BIT	(NBBY*sizeof(unsigned int))
#endif

typedef struct Aoemaj {
	int major;
	struct Aoemaj *next;
	unsigned int bits[(AOEMINMAX + WORD_BIT - 1)/WORD_BIT];
} Aoemaj;

#define	MINSET(xp, i)	((xp)->bits[(i)/WORD_BIT] |= (1<<((i)%WORD_BIT)))
#define	MINCLR(xp, i)	((xp)->bits[(i)/WORD_BIT] &=~ (1<<((i)%WORD_BIT)))
#define	MINISSET(xp, i)	(((xp)->bits[(i)/WORD_BIT] & (1<<((i)%WORD_BIT))) != 0)

#define	NOCOOKIE	(Aoecookie)0

/*
 * state per instance
 */

typedef struct Aoechan {
	queue_t *wq;
	Aoemaj *majlist;
	krwlock_t lock;
	char locmac[LEN_EADDR];
	int chan;
	int proto;
	void (*receiver)(int, mblk_t *);
	int state;
	int maxdata;
} Aoechan;

#define	ACSINIT	0
#define	ACSACT	1

#define	MAXCHAN	10

static Aoechan *aoechan[MAXCHAN];
static kmutex_t chantlock;
static Aoecookie pendcookie[MAXCHAN];
static kmutex_t pendlock;
static void (*logger)(int, int, mblk_t *);
static krwlock_t loggerlock;

/*
 * entry points and internal functions
 */

static int aoecommopen(queue_t *, dev_t *, int, int, cred_t *);
static int aoecommclose(queue_t *, int, cred_t *);
static int aoecommrput(queue_t *, mblk_t *);
static int aoecommrsrv(queue_t *);
static int aoecommwsrv(queue_t *);

static void setchparms(Aoechan *, char *, queue_t *);
static void ackreply(queue_t *, int);
static int bcastsend(mblk_t *);
static mblk_t *fixheader(mblk_t *, Aoechan *);
static Aoechan *numtochan(int);
static void doenab(Aoechan *, int, int);
static void dodisab(Aoechan *, int, int);
static int doquery(Aoechan *, int, int);

/*
 * configuration data structures
 */

#define	EMTU	1514		/* default Ethernet MTU */
#define	RMSGS	80		/* size of network input queue */
#define	CMTU	LEN_ACMAX	/* longest possible command message */

static struct module_info aoecommrinfo = {
	0xa0e0,			/* arbitrary? */
	"aoecomm",
	0, INFPSZ,
	RMSGS*EMTU, EMTU	/* inward network messages */
};

static struct module_info aoecommwinfo = {
	0xa0e0,
	"aoecomm",
	0, INFPSZ,
	2*CMTU, CMTU/2		/* commands, occasional network message */
};

static struct qinit aoecommrinit = {
	aoecommrput,	/* put */
	aoecommrsrv,	/* serv */
	aoecommopen, aoecommclose,
	NULL,		/* qadmin */
	&aoecommrinfo
};

static struct qinit aoecommwinit = {
	putq, aoecommwsrv,
	aoecommopen, aoecommclose,
	NULL,		/* qadmin */
	&aoecommwinfo
};

static struct streamtab aoecomm_info = {
	&aoecommrinit, &aoecommwinit,
	NULL, NULL		/* muxrinit, muxwinit */
};

static struct fmodsw aoecomm_fsw = {
	"aoecomm", &aoecomm_info, D_NEW|D_MP
};

static struct modlstrmod modlstrmod = {
	&mod_strmodops, "AoE comm module v1.4", &aoecomm_fsw
};

static struct modlinkage modlinkage = {
	MODREV_1, &modlstrmod, NULL
};

/*
 * module loading and unloading
 */
int
_init()
{
	int e, i;

	if ((e = mod_install(&modlinkage)) != 0)
		return (e);
	mutex_init(&chantlock, NULL, MUTEX_DRIVER, NULL);
	mutex_init(&pendlock, NULL, MUTEX_DRIVER, NULL);
	rw_init(&loggerlock, NULL, RW_DRIVER, NULL);
	for (i = 0; i < MAXCHAN; i++)
		pendcookie[i] = NOCOOKIE;
	return (e);
}

int
_fini()
{
	int e;

	if ((e = mod_remove(&modlinkage)) != 0)
		return (e);
	mutex_destroy(&chantlock);
	mutex_destroy(&pendlock);
	rw_destroy(&loggerlock);
	return (e);
}

int
_info(mp)
struct modinfo *mp;
{
	return (mod_info(&modlinkage, mp));
}

static int
aoecommopen(q, devp, flag, sflag, crp)
queue_t *q;
dev_t *devp;
int flag, sflag;
cred_t *crp;
{
	Aoechan *ap;

	if (q->q_ptr)
		return (0);	/* already open; shouldn't happen? */
	if ((ap = kmem_zalloc(sizeof(Aoechan), KM_NOSLEEP)) == NULL)
		return (ENXIO);
	rw_init(&ap->lock, NULL, RW_DRIVER, NULL);
	ap->wq = WR(q);
	ap->state = ACSINIT;
	q->q_ptr = ap;
	WR(q)->q_ptr = ap;
	qprocson(q);
	return (0);
}

/*
 * -- what if aoecomm_send is called while we're
 * removing ap from the channel table?
 * chantlock handles that
 */
static int
aoecommclose(q, flag, crp)
queue_t *q;
int flag;
cred_t *crp;
{
	Aoechan *ap;
	Aoemaj *xp;
	int i;

	qprocsoff(q);
	if ((ap = q->q_ptr) == NULL)
		cmn_err(CE_PANIC, "^aoecomm: close with NULL q_ptr");
	mutex_enter(&chantlock);
	for (i = 0; i < MAXCHAN; i++)
		if (aoechan[i] == ap)
			aoechan[i] = NULL;
	mutex_exit(&chantlock);
	mutex_enter(&pendlock);
	if (ap->state == ACSACT)
		pendcookie[ap->chan] = NOCOOKIE;
	mutex_exit(&pendlock);
	if (ap->receiver)
		(*ap->receiver)(ap->chan, NULL);
	while ((xp = ap->majlist) != NULL) {
		ap->majlist = xp->next;
		kmem_free(xp, sizeof(*xp));
	}
	rw_destroy(&ap->lock);
	kmem_free(ap, sizeof(Aoechan));
	q->q_ptr = NULL;
	WR(q)->q_ptr = NULL;
	return (0);
}

/*
 * message arrives from network
 * hand to receiver if there is one;
 * otherwise send upstream
 */

static int
aoecommrput(q, mp)
queue_t *q;
mblk_t *mp;
{
	Aoechan *ap;
	int state;

	if ((ap = q->q_ptr) == NULL)
		cmn_err(CE_PANIC, "^aoecomm: rput with NULL q_ptr");
	if (mp->b_datap->db_type >= QPCTL) {
		putnext(q, mp);
		return (1);
	}
	rw_enter(&ap->lock, RW_READER);
	state = ap->state;
	rw_exit(&ap->lock);
	if (state != ACSACT)
		freemsg(mp);
	else
		putq(q, mp);
	return (1);
}

static int
aoecommrsrv(q)
queue_t *q;
{
	Aoechan *ap;
	mblk_t *mp;
	mblk_t *bp;
	int n;
	void (*func)(int, mblk_t *);

	if ((ap = q->q_ptr) == NULL)
		cmn_err(CE_PANIC, "^aoecomm: rsrv with NULL q_ptr");
	while ((mp = getq(q)) != NULL) {
		if (mp->b_datap->db_type == M_DATA) {
			rw_enter(&ap->lock, RW_READER);
			func = ap->receiver;
			rw_exit(&ap->lock);
			if (func)
				(*func)(ap->chan, mp);
			else
				aoecomm_log(ap->chan, ACLUNSOL, mp);
			continue;
		}
		if (!canputnext(q)) {
			putbq(q, mp);
			break;
		}
		putnext(q, mp);
	}
	return (1);
}

/*
 * message arrives from user process
 * parse as a command,
 * which may send data downstream
 */

static int
aoecommwsrv(q)
queue_t *q;
{
	Aoechan *ap;
	mblk_t *mp, *nmp;
	int len, n, e, mtu;

	if ((ap = q->q_ptr) == NULL)
		cmn_err(CE_PANIC, "^aoecomm: wsrv with NULL q_ptr");
	while ((mp = getq(q)) != NULL) {
		if (mp->b_datap->db_type >= QPCTL) {
			putnext(q, mp);
			continue;
		}
		if (mp->b_datap->db_type != M_DATA) {
			if (!canputnext(q)) {
				putbq(q, mp);
				return (1);
			}
			putnext(q, mp);
			continue;
		}
		n = msgdsize(mp);
		nmp = msgpullup(mp, min(n, LEN_ACMAXSHORT));
		freemsg(mp);
		if ((mp = nmp) == NULL)
			continue;
		len = fraoeshort(mp->b_rptr, AC_LEN);
		if (len != n) {
			freemsg(mp);
			continue;
		}
		switch (fraoechar(mp->b_rptr, AC_TYPE)) {
		case ACINITCHAN:
			if (len < LEN_ACINITCHAN) {
				freemsg(mp);
				continue;
			}
			n = fraoeshort(mp->b_rptr, AC_CHANNEL);
			if (n < 0 || n >= MAXCHAN) {
				ackreply(q, EINVAL);
				freemsg(mp);
				continue;
			}
			e = 0;
			mutex_enter(&chantlock);
			if (numtochan(n) != NULL)
				e = EBUSY;
			mutex_exit(&chantlock);
			mutex_enter(&pendlock);
			if (pendcookie[n] == NOCOOKIE
			||  fraoelong(mp->b_rptr, AC_COOKIE) != pendcookie[n])
				e = ENXIO;
			mutex_exit(&pendlock);
			if (fraoeshort(mp->b_rptr, AC_MAXDATA) < AOEASEC)
				e = EINVAL;
			if (e == 0) {
				rw_enter(&ap->lock, RW_WRITER);
				if (ap->state != ACSINIT)
					e = EBUSY;
				rw_exit(&ap->lock);
			}
			if (e == 0) {
				setchparms(ap, (char *)mp->b_rptr, RD(q));
				mutex_enter(&chantlock);
				aoechan[n] = ap;
				mutex_exit(&chantlock);
			}
			ackreply(q, e);
			freemsg(mp);
			continue;

		default:
			freemsg(mp);
			continue;
		}
	}
	return (0);
}

/*
 * q should be the read queue
 */
static void
setchparms(ap, p, q)
Aoechan *ap;
char *p;
queue_t *q;
{
	int e, mtu;
	intptr_t v;

	ap->chan = fraoeshort(p, AC_CHANNEL);
	ap->proto = fraoeshort(p, AC_PROTO);
	bcopy(p + AC_LOCADDR, ap->locmac, LEN_EADDR);
	ap->state = ACSACT;
	ap->maxdata = fraoeshort(p, AC_MAXDATA);
	if (ap->maxdata == 0)
		ap->maxdata = AOEASEC * AOEAMAXSEC;
	mtu = ap->maxdata + LEN_ETHER + MINLEN_AOEA;
	if (mtu < EMTU)
		mtu = EMTU;
	v = mtu * RMSGS;
	if ((e = strqset(q, QHIWAT, 0, v)) != 0)
		cmn_err(CE_CONT,
			"?aoecomm: strqset failed to set QHIWAT for mtu %d: %d\n",
			mtu, e);
}

/*
 * send back INITACK with code
 */
static void
ackreply(q, code)
queue_t *q;
int code;
{
	mblk_t *mp;

	if ((mp = allocb(LEN_ACINITACK, 0)) == NULL)
		return;
	mp->b_wptr += LEN_ACINITACK;
	bzero(mp->b_rptr, LEN_ACINITACK);
	toaoechar(mp->b_rptr, AC_TYPE, ACINITACK);
	toaoeshort(mp->b_rptr, AC_LEN, LEN_ACINITACK);
	toaoechar(mp->b_rptr, AC_STATUS, code);
	qreply(q, mp);
}

/*
 * external call: store a cookie for chan,
 * if not already active
 */
int
aoecomm_initchan(chan, cookie)
int chan;
Aoecookie cookie;
{
	Aoechan *ap;

	if (chan < 0 || chan >= MAXCHAN)
		return (-1);
	mutex_enter(&chantlock);
	ap = numtochan(chan);
	mutex_exit(&chantlock);
	if (ap)
		return (-1);
	mutex_enter(&pendlock);
	pendcookie[chan] = cookie;
	mutex_exit(&pendlock);
	return (1);
}

/*
 * establish communications to disk driver for chan
 */
int
aoecomm_initdriver(chan, receiver)
int chan;
void (*receiver)(int, mblk_t *);
{
	Aoechan *ap;
	int ret;

	ret = -1;
	mutex_enter(&chantlock);
	ap = numtochan(chan);
	mutex_exit(&chantlock);
	if (ap == NULL)
		return (-1);
	rw_enter(&ap->lock, RW_WRITER);
	if (ap->receiver == NULL || receiver == NULL
	||  ap->receiver == receiver) {
		ap->receiver = receiver;
		ret = 1;
	}
	rw_exit(&ap->lock);
	return (ret);
}

/*
 * external call:
 * return max data length for this channel
 */
int
aoecomm_maxdata(chan)
int chan;
{
	Aoechan *ap;
	int ret;

	ret = -1;
	mutex_enter(&chantlock);
	ap = numtochan(chan);
	mutex_exit(&chantlock);
	if (ap == NULL)
		return (-1);
	rw_enter(&ap->lock, RW_READER);
	ret = ap->maxdata;
	rw_exit(&ap->lock);
	return (ret);
}

/*
 * pass a log message to the logger,
 * if any
 */
int
aoecomm_initlog(func)
void (*func)(int, int, mblk_t *);
{
	int ret;

	ret = -1;
	rw_enter(&loggerlock, RW_WRITER);
	if (logger == NULL || func == NULL
	||  logger == func) {
		logger = func;
		ret = 1;
	}
	rw_exit(&loggerlock);
	return (ret);
}

void
aoecomm_log(chan, code, mp)
int chan;
int code;
mblk_t *mp;
{
	void (*func)(int, int, mblk_t *);

	if (mp == NULL)
		return;		/* safety */
	rw_enter(&loggerlock, RW_READER);
	func = logger;
	rw_exit(&loggerlock);
	if (func)
		(*func)(chan, code, mp);
	else
		freemsg(mp);
}

/*
 * forward a message out to the network
 * -- ordering is a little odd so that all references
 * to ap are protected by chantlock;
 * fixes a race when closing
 */
int
aoecomm_send(chan, mp)
int chan;
mblk_t *mp;
{
	Aoechan *ap;
	queue_t *q;

	if (chan == ACHWILD)
		return (bcastsend(mp));
	mutex_enter(&chantlock);
	if ((ap = numtochan(chan)) == NULL)
		q = NULL;
	else {
		q = ap->wq;
		mp = fixheader(mp, ap);
	}
	mutex_exit(&chantlock);
	if (mp == NULL)
		return (-1);
	if (q == NULL || !canputnext(q)) {
		freemsg(mp);
		return (-1);
	}
	putnext(q, mp);
	return (1);
}

/*
 * send a copy of mp to each active channel
 * return > 0 if at least one succeeded
 */
static int
bcastsend(mp)
mblk_t *mp;
{
	int i;
	Aoechan *ap;
	queue_t *q;
	mblk_t *cp;
	int ret;

	ret = -1;
	for (i = 0; i < MAXCHAN; i++) {
		cp = NULL;
		mutex_enter(&chantlock);
		if ((ap = numtochan(i)) != NULL
		&&   (q = ap->wq) != NULL
		&&   (cp = copymsg(mp)) != NULL)
			cp = fixheader(cp, ap);
		mutex_exit(&chantlock);
		if (cp == NULL)
			continue;
		if (!canputnext(q)) {
			freemsg(cp);
			continue;
		}
		putnext(q, cp);
		ret = 1;
	}
	freemsg(mp);
	return (ret);
}
		
/*
 * fill in source address, protocol in MAC header
 * pad with zeroes to minimum datagram length
 * one would think the Ethernet device driver would do the latter,
 * but empirically some drivers pad with random system data
 */

static mblk_t *
fixheader(mp, ap)
mblk_t *mp;
Aoechan *ap;
{
	int n;
	mblk_t *bp;

	if (mp->b_wptr < mp->b_rptr + LEN_ETHER) {
		bp = msgpullup(mp, LEN_ETHER);
		freemsg(mp);
		if ((mp = bp) == NULL)
			return (NULL);
	}
	bcopy(ap->locmac, mp->b_rptr + ETHER_SRC, LEN_EADDR);
	toetherprot(mp->b_rptr, ETHER_PROT, ap->proto);
	if (mp->b_cont == NULL && mp->b_rptr + MINLEN_ETHER < mp->b_wptr)
		return (mp);
	if ((n = msgdsize(mp)) < MINLEN_ETHER
	&&  (bp = allocb(MINLEN_ETHER - n, 0)) != NULL) {
		n = MINLEN_ETHER - n;
		bp->b_wptr += n;
		bzero(bp->b_rptr, n);
		linkb(mp, bp);
	}
	return (mp);
}

/*
 * maintain list of enabled devices
 * not used by aoecomm itself
 * -- again, protected by chantlock
 */
int
aoecomm_devenab(chan, maj, min, cmd)
int chan, maj, min, cmd;
{
	Aoechan *ap;
	int answer;
	int i;

	answer = ADDISAB;
	if (maj < 0 || min < 0
	||  (maj > AOEMAJMAX && maj != AOEMAJWILD)
	||  (min > AOEMINMAX && min != AOEMINWILD))
		return (answer);
	mutex_enter(&chantlock);
	switch (cmd) {
	case ADENAB:
		if (maj == AOEMAJWILD)
			break;
		if ((ap = numtochan(chan)) == NULL)
			break;
		doenab(ap, maj, min);
		answer = ADENAB;
		break;

	case ADDISAB:
		if ((ap = numtochan(chan)) == NULL)
			break;
		dodisab(ap, maj, min);
		break;

	case ADQUERY:
		if (chan == ACHWILD) {
			for (i = 0; i < MAXCHAN; i++) {
				if ((ap = aoechan[chan]) != NULL
				&&  (answer = doquery(ap, maj, min)) != ADDISAB)
					break;
			}
		} else {
			if ((ap = numtochan(chan)) == NULL)
				break;
			answer = doquery(ap, maj, min);
		}
		break;

	/* default: fall through, return ADDISAB */
	}
	mutex_exit(&chantlock);
	return (answer);
}

/*
 * caller is responsible for locking chantlock
 * -- assume ADCHGLOB is out of range,
 * hence no special check needed
 */
static Aoechan *
numtochan(chan)
int chan;
{
	Aoechan *ap;

	if (chan < 0 || chan >= MAXCHAN)
		return (NULL);
	ap = aoechan[chan];
	return (ap);
}

static void
doenab(ap, maj, min)
Aoechan *ap;
int maj, min;
{
	Aoemaj *xp;

	for (xp = ap->majlist; xp; xp = xp->next)
		if (xp->major == maj)
			break;
	if (xp == NULL) {
		if ((xp = kmem_zalloc(sizeof(*xp), KM_NOSLEEP)) == NULL)
			return;
		xp->major = maj;
		xp->next = ap->majlist;
		ap->majlist = xp;
	}
	if (min == AOEMINWILD) {
		for (min = 0; min <= AOEMINMAX; min++)
			MINSET(xp, min);
	} else
		MINSET(xp, min);
}

static void
dodisab(ap, maj, min)
Aoechan *ap;
int maj, min;
{
	Aoemaj *xp;

	for (xp = ap->majlist; xp; xp = xp->next) {
		if (xp->major != maj && maj != AOEMAJWILD)
			continue;
		if (min == AOEMINWILD)
			bzero(xp->bits, sizeof(xp->bits));
		else
			MINCLR(xp, min);
		if (maj != AOEMAJWILD)
			break;
	}
}

static int
doquery(ap, maj, min)
Aoechan *ap;
int maj, min;
{
	Aoemaj *xp;
	int i;

	for (xp = ap->majlist; xp; xp = xp->next) {
		if (xp->major != maj && maj != AOEMAJWILD)
			continue;
		if (min == AOEMINWILD) {
			for (i = 0; i < AOEMINMAX; i++)
				if (MINISSET(xp, i))
					return (ADENAB);
		} else if (MINISSET(xp, min))
			return (ADENAB);
		if (maj != AOEMAJWILD)
			return (ADDISAB);
	}
	return (ADDISAB);
}
