/*
 * Copyright 2005 Coraid, Inc.
 * See accompanying file doc/LICENSE for conditions
 */

/*
 * aoe configuration-confirmation driver
 * exists mostly to protect aoecomm line discipline
 * against rogues:
 * anyone may push a line discipline on a stream,
 * so require someone setting up an aoecomm channel
 * to prove they can open a special file too,
 * because special files can be given specific permissions
 */

/*
 * these includes taken from one of the sample drivers
 * Sun's Writing Device Drivers doesn't mention them,
 * but without at least some of them the driver won't compile
 */
#include <sys/types.h>
#include <sys/file.h>
#include <sys/errno.h>
#include <sys/uio.h>
#include <sys/open.h>
#include <sys/cred.h>
#include <sys/kmem.h>
#include <sys/conf.h>
#include <sys/cmn_err.h>
#include <sys/ksynch.h>
#include <sys/modctl.h>
#include <sys/stat.h>

/*
 * these are mentioned in the book
 */
#include <sys/stream.h>
#include <sys/ddi.h>
#include <sys/sunddi.h>

/*
 * these are specific to the AoE drivers
 */
#include "aoeproto.h"
#include "aoecommproto.h"
#include "aoecomm.h"
#include "etherproto.h"

/*
 * minor devices
 */

#define	CTLDEV	0
#define	MONDEV	1

/*
 * driver's own state
 * statically allocated because
 * there is only one instance of us
 * but keep it in a structure
 * against future need
 */

#define	CINST	0	/* sole instance allowed */

typedef struct Aoemon {
	int chan;
	unsigned int nlost;
	int code;
	mblk_t *msg;
	struct Aoemon *next;
} Aoemon;

#define	NMONBUF	32

typedef struct {
	kmutex_t clock;		/* protect CTLDEV */
	kmutex_t mlock;		/* protect MONDEV */
	kcondvar_t mcv;
	int monsleep;
	dev_info_t *dip;
	Aoemon *monbuf;
	Aoemon *wmon, *rmon;
	int nmonbuf;
} Aoectlstate;

static Aoectlstate aoectlstate;

/*
 * entry points and internal functions
 */

static int aoectlgetinfo(dev_info_t *, ddi_info_cmd_t, void *, void **);
static int aoectlattach(dev_info_t *, ddi_attach_cmd_t);
static int aoectldetach(dev_info_t *, ddi_detach_cmd_t);

static int aoectlopen(dev_t *, int, int, cred_t *);
static int aoectlclose(dev_t, int, int, cred_t *);
static int aoectlwrite(dev_t, struct uio *, cred_t *);
static int aoectlread(dev_t, struct uio *, cred_t *);

static void aoectlintr(int, int, mblk_t *);
static int aoectlmoninit(Aoectlstate *, int);
static void aoectlmonfree(Aoectlstate *);

/*
 * configuration data structures
 */

static struct cb_ops aoectl_cb = {
	aoectlopen, aoectlclose,
	nulldev, nulldev, nodev,	/* strategy, print, dump */
	aoectlread, aoectlwrite, nodev,	/* read, write, ioctl */
	nodev, nodev, nodev,	/* devmap, mmap, segmap */
	nochpoll, ddi_prop_op,
	NULL,			/* streamtab */
	D_MP, CB_REV,
	nodev, nodev		/* aread, awrite */	
};

static struct dev_ops aoectl_ops = {
	DEVO_REV, 0,
	aoectlgetinfo, nulldev, nulldev, aoectlattach, aoectldetach,
	nodev,		/* reset */
	&aoectl_cb, NULL,	/* cb_ops, bus_ops */
	NULL		/* power */
};

static struct modldrv modldrv = {
	&mod_driverops, "AoE config setup v1.4", &aoectl_ops
};

static struct modlinkage modlinkage = {
	MODREV_1, &modldrv, NULL
};

static char _depends_on[] = "strmod/aoecomm";

/*
 * module loading and unloading
 */
int
_init()
{
	return (mod_install(&modlinkage));
}

int
_fini()
{
	return (mod_remove(&modlinkage));
}

int
_info(mp)
struct modinfo *mp;
{
	return (mod_info(&modlinkage, mp));
}

/*
 * autoconfiguration
 */

static int
aoectlattach(dip, cmd)
dev_info_t *dip;
ddi_attach_cmd_t cmd;
{
	Aoectlstate *cfp;

	if (ddi_get_instance(dip) != CINST || cmd != DDI_ATTACH)
		return (DDI_FAILURE);
	if (ddi_create_minor_node(dip, "ctl", S_IFCHR, CTLDEV, DDI_PSEUDO, 0) != DDI_SUCCESS
	||  ddi_create_minor_node(dip, "mon", S_IFCHR, MONDEV, DDI_PSEUDO, 0) != DDI_SUCCESS)
		return (DDI_FAILURE);
	cfp = &aoectlstate;
	cfp->dip = dip;
	mutex_init(&cfp->clock, NULL, MUTEX_DRIVER, NULL);
	mutex_init(&cfp->mlock, NULL, MUTEX_DRIVER, NULL);
	cv_init(&cfp->mcv, NULL, CV_DRIVER, NULL);
	ddi_report_dev(dip);
	return (DDI_SUCCESS);
}

static int
aoectldetach(dip, cmd)
dev_info_t *dip;
ddi_detach_cmd_t cmd;
{
	Aoectlstate *cfp;
	int i;

	i = ddi_get_instance(dip);
	if (cmd != DDI_DETACH)
		return (DDI_FAILURE);
	cfp = &aoectlstate;
	ddi_remove_minor_node(dip, NULL);
	mutex_destroy(&cfp->clock);
	mutex_destroy(&cfp->mlock);
	cv_destroy(&cfp->mcv);
	return (DDI_SUCCESS);
}

static int
aoectlgetinfo(dip, cmd, arg, resp)
dev_info_t *dip;
ddi_info_cmd_t cmd;
void *arg;
void **resp;
{
	switch (cmd) {
	case DDI_INFO_DEVT2INSTANCE:
		*resp = CINST;
		return (DDI_SUCCESS);

	case DDI_INFO_DEVT2DEVINFO:
		*resp = (void *)aoectlstate.dip;
		return (DDI_SUCCESS);

	default:
		cmn_err(CE_CONT, "?aoectl: getinfo unknown cmd %d, arg 0x%lx\n", cmd, (long)arg);
		return (DDI_FAILURE);
	}
}

/*
 * device operations
 */

static int
aoectlopen(devp, flag, type, crp)
dev_t *devp;
int flag, type;
cred_t *crp;
{
	Aoectlstate *cfp;
	int n;
	int e;

	cfp = &aoectlstate;
	e = 0;
	switch (getminor(*devp)) {
	case CTLDEV:
		break;

	case MONDEV:
		if ((n = ddi_prop_get_int(DDI_DEV_T_ANY, cfp->dip, 0, "monbufs", -1)) <= 0)
			n = NMONBUF;
		mutex_enter(&cfp->mlock);
		if (cfp->monbuf != NULL)
			e = EBUSY;
		else if (aoecomm_initlog(aoectlintr) < 0)
			e = EBUSY;
		else if (aoectlmoninit(cfp, n) < 0) {
			aoecomm_initlog(NULL);
			e = EBUSY;
		}
		mutex_exit(&cfp->mlock);
		break;

	default:
		e = ENXIO;
		break;
	}
	return (e);
}

static int
aoectlclose(dev, flag, type, crp)
dev_t dev;
int flag, type;
cred_t *crp;
{
	Aoectlstate *cfp;
	int n;

	cfp = &aoectlstate;
	switch (getminor(dev)) {
	case MONDEV:
		mutex_enter(&cfp->mlock);
		aoecomm_initlog(NULL);
		aoectlmonfree(cfp);
		mutex_exit(&cfp->mlock);
		break;
	}
	return (0);
}

static int
aoectlwrite(dev, uio, crp)
dev_t dev;
struct uio *uio;
cred_t *crp;
{
	Aoectlstate *cfp;
	char buf[LEN_ACMAX];
	int e, oresid, len;
	mblk_t *mp;

	if (getminor(dev) != CTLDEV)
		return (ENODEV);
	cfp = &aoectlstate;
	if ((oresid = uio->uio_resid) < LEN_ACHEAD)
		return (EINVAL);
	if ((e = uiomove((caddr_t)buf, LEN_ACHEAD, UIO_WRITE, uio)) != 0)
		return (e);
	len = fraoeshort(buf, AC_LEN);
	if (len < LEN_ACHEAD || len > LEN_ACMAX)
		return (EINVAL);
	if ((e = uiomove((caddr_t)(buf + LEN_ACHEAD), len - LEN_ACHEAD, UIO_WRITE, uio)) != 0)
		return (e);
	mutex_enter(&cfp->clock);
	e = 0;
	switch (fraoechar(buf, AC_TYPE)) {
	case ACINITCHAN:
		if (len < LEN_ACINITCHAN)
			e = EINVAL;
		else if (aoecomm_initchan(fraoeshort(buf, AC_CHANNEL), fraoelong(buf, AC_COOKIE)) <= 0)
			e = ENXIO;
		break;

	case ACSEND:
		if (len < MINLEN_ACSEND || len > LEN_ACSEND) {
			e = EINVAL;
			break;
		}
		len -= AC_MESSAGE;
		if ((mp = allocb(len, 0)) == NULL) {
			e = ENOMEM;
			break;
		}
		bcopy(buf + AC_MESSAGE, mp->b_wptr, len);
		mp->b_wptr += len;
		if (aoecomm_send(fraoeshort(buf, AC_CHANNEL), mp) <= 0)
			e = ENXIO;
		break;

	case ACDEVENAB:
		if (len < LEN_ACDEVENAB) {
			e = EINVAL;
			break;
		}
		if (aoecomm_devenab(fraoeshort(buf, AC_CHANNEL),
		      fraoeshort(buf, AC_MAJOR), fraoechar(buf, AC_MINOR),
		      fraoechar(buf, AC_ENABCMD)) == ADDISAB)
			uio->uio_resid = oresid;	/* write returns 0 */
		break;

	default:
		e = EINVAL;
		break;
	}
	mutex_exit(&cfp->clock);
	return (e);
}

/*
 * the buffering scheme works as follows:
 * monbuf is treated as a ring; to advance to the
 * next message, follow the next link
 *
 * cfp->rmon is the next message to be read;
 * cfp->wmon the next to be written
 * if rmon == wmon the buffer is empty;
 * if rmon == wmon->next it's full
 * (so the capacity is one less than it appears)
 * if a message arrives when the buffer is full,
 * the `lost' counter of the unused sentinel is incremented;
 * the first message correctly received will
 * include the missed-message counter
 */

static int
aoectlread(dev, uio, crp)
dev_t dev;
struct uio *uio;
cred_t *crp;
{
	Aoectlstate *cfp;
	char buf[LEN_ACMAXSHORT];
	Aoemon am;
	int msglen;
	int e;
	mblk_t *mp;

	if (getminor(dev) != MONDEV)
		return (ENODEV);
	cfp = &aoectlstate;
	mutex_enter(&cfp->mlock);
	while (cfp->rmon == cfp->wmon) {
		cfp->monsleep = 1;
		if (cv_wait_sig(&cfp->mcv, &cfp->mlock) == 0) {
			mutex_exit(&cfp->mlock);
			return (EINTR);
		}
	}
	mutex_exit(&cfp->mlock);
	e = 0;
	for (;;) {
		mutex_enter(&cfp->mlock);
		if (cfp->rmon == cfp->wmon) {
			mutex_exit(&cfp->mlock);
			break;
		}
		bcopy(cfp->rmon, &am, sizeof(Aoemon));
		msglen = AC_MESSAGE;
		if (am.msg)
			msglen += msgdsize(am.msg);
		if (msglen > uio->uio_resid) {
			mutex_exit(&cfp->mlock);
			break;
		}
		cfp->rmon->nlost = 0;
		cfp->rmon->msg = NULL;
		cfp->rmon = am.next;
		mutex_exit(&cfp->mlock);
		bzero(buf, sizeof(buf));
		toaoechar(buf, AC_TYPE, ACLOG);
		toaoeshort(buf, AC_LEN, msglen);
		toaoeshort(buf, AC_CHANNEL, am.chan);
		toaoechar(buf, AC_LTYPE, am.code);
		toaoeshort(buf, AC_NLOST, am.nlost);
		if ((e = uiomove(buf, AC_MESSAGE, UIO_READ, uio)) != 0) {
			if (am.msg)
				freemsg(am.msg);
			break;
		}
		for (mp = am.msg; mp; mp = mp->b_cont) {
			if (mp->b_datap->db_type != M_DATA)
				continue;
			e = uiomove(mp->b_rptr, mp->b_wptr - mp->b_rptr, UIO_READ, uio);
			if (e != 0)
				break;
		}
		if (am.msg)
			freemsg(am.msg);
		if (e != 0)
			break;
	}
	return (e);
}

/*
 * message arrives;
 * add to the buffer if possible
 */
static void
aoectlintr(chan, code, mp)
int chan;
int code;
mblk_t *mp;
{
	Aoectlstate *cfp;
	Aoemon *amp;

	cfp = &aoectlstate;
	mutex_enter(&cfp->mlock);
	if (cfp->monbuf == NULL) {
		mutex_exit(&cfp->mlock);
		freemsg(mp);
		return;
	}
	if ((amp = cfp->wmon) == NULL)
		cmn_err(CE_PANIC, "?aoectlintr null wmon\n");
	if (amp->next == cfp->rmon) {
		amp->nlost++;
		freemsg(mp);
	} else {
		amp->msg = mp;
		amp->chan = chan;
		amp->code = code;
		cfp->wmon = amp->next;
	}
	if (cfp->monsleep) {
		cfp->monsleep = 0;
		cv_signal(&cfp->mcv);
	}
	mutex_exit(&cfp->mlock);
}

/*
 * allocate and set up monitor buffer
 */
static int
aoectlmoninit(cfp, n)
Aoectlstate *cfp;
int n;
{
	Aoemon *buf;
	int i;

	if (n <= 0)
		return (-1);
	n++;		/* room for the between-messages sentinel */
	if ((buf = kmem_zalloc(n * sizeof(Aoemon), KM_NOSLEEP)) == NULL)
		return (-1);
	for (i = n - 1; i > 0; i--)
		buf[i - 1].next = &buf[i];
	buf[n - 1].next = buf;
	cfp->nmonbuf = n;
	cfp->monbuf = buf;
	cfp->wmon = cfp->rmon = buf;
	return (1);
}

/*
 * clear monitor buffer
 */
static void
aoectlmonfree(cfp)
Aoectlstate *cfp;
{
	int i;
	Aoemon *amp;

	if (cfp->monbuf) {
		amp = cfp->monbuf;
		for (i = 0; i < cfp->nmonbuf; amp++, i++)
			if (amp->msg)
				freemsg(amp->msg);
		kmem_free(cfp->monbuf, cfp->nmonbuf * sizeof(Aoemon));
	}
	cfp->monbuf = NULL;
}
