/*
 * Copyright 2005-2007 Coraid, Inc.
 * See accompanying file doc/LICENSE for conditions
 */

/*
 * AoE disk driver
 * main entry points
 *
 * Simple ATA disk driver
 * using AoE to reach the drive
 * --it would be better to write an ATA host-adapter driver,
 * and use the existing Solaris ATA code;
 * but ATA HBA drivers are not documented, are they?
 */

/*
 * these includes taken from one of the sample drivers
 * Sun's Writing Device Drivers doesn't mention them,
 * but without at least some of them the driver won't compile
 */
#include <sys/types.h>
#include <sys/file.h>
#include <sys/errno.h>
#include <sys/uio.h>
#include <sys/open.h>
#include <sys/cred.h>
#include <sys/kmem.h>
#include <sys/conf.h>
#include <sys/cmn_err.h>
#include <sys/ksynch.h>
#include <sys/modctl.h>
#include <sys/stat.h>
#include <sys/buf.h>

static char _depends_on[] = "strmod/aoecomm";

/*
 * these are mentioned in the book
 */
#include <sys/ddi.h>
#include <sys/sunddi.h>

/*
 * these are needed because it's a disk driver
 * or because we may keep statistics
 * or to define special undocumented ioctls
 */
#include <sys/dkio.h>
#include <sys/vtoc.h>
#include <sys/kstat.h>
#include <sys/dktp/dadkio.h>	/* ATA-specific ioctls */
#include <sys/scsi/scsi.h>	/* SCSI-specific ioctls */

/*
 * these are specific to AoE
 */
#include "aoeproto.h"
#include "aoecomm.h"
#include "aoecommproto.h"	/* for aoecomm_devenab commands */
#include "etherproto.h"		/* for Adrive */
#include "ata.h"		/* for Adrive */
#include "aoed.h"

void *aoedrives;
void *aoechannels;
static int aoemaxchan;		/* only needed for cleanup */

char *adrvname;

/*
 * entry points and internal functions
 */

static int adgetinfo(dev_info_t *, ddi_info_cmd_t, void *, void **);
static int adattach(dev_info_t *, ddi_attach_cmd_t);
static int addetach(dev_info_t *, ddi_detach_cmd_t);
static int adopen(dev_t *, int, int, cred_t *);
static int adclose(dev_t, int, int, cred_t *);
static int adstrategy(struct buf *);
static int adprint(dev_t, char *);
static int adread(dev_t, struct uio *, cred_t *);
static int adwrite(dev_t, struct uio *, cred_t *);
static int adioctl(dev_t, int, intptr_t, int, cred_t *, int *);

static int createnode(dev_info_t *, char *, int, minor_t, int);
static int openwait(Adrive *);
static int readptab(Adrive *);
static void mkstrprop(Adrive *, char *, char *);
static void setdevid(Adrive *);
static void mkkstats(Adrive *, int);
static void instkstat(Adrive *, kstat_t *, char *);
static int sanegeom(Adrive *, struct dk_geom *);

/*
 * here only for debugging;
 * to use it, #define DEBUG_PROP 1
 */
#ifdef DEBUG_PROP
static int AOEddi_prop_op(dev, dip, op, f, name, vp, lp)
dev_t dev;
dev_info_t *dip;
ddi_prop_op_t op;
int f;
char *name;
caddr_t vp;
int *lp;
{
	int e;

	admsg(dev, NULL, "aoe prop: %s, len %d, op %d\n",
		name, *lp, op);
	e = ddi_prop_op(dev, dip, op, f, name, vp, lp);
	if (e != DDI_PROP_SUCCESS)
		admsg(dev, NULL, "aoe prop: return %d\n", e);
	else if (*lp == 4)
		admsg(dev, NULL, "aoe prop: %s = %d\n", name, *(int *)vp);
	else if (*lp == 8)
		admsg(dev, NULL, "aoe prop: %s = %lld\n", name, *(long long *)vp);
	return (e);
}
#define	ddi_prop_op	AOEddi_prop_op
#endif

/*
 * configuration data structures
 */

static struct cb_ops ad_cb = {
	adopen, adclose,
	adstrategy, adprint, nodev,	/* strategy, print, dump */
	adread, adwrite, adioctl,
	nodev, nodev, nodev,	/* devmap, mmap, segmap */
	nochpoll,
	ddi_prop_op,
	NULL,			/* streamtab */
	D_MP|D_64BIT, CB_REV,
	nodev, nodev		/* aread, awrite */	
};

static struct dev_ops ad_ops = {
	DEVO_REV, 0,
	adgetinfo, nulldev, nulldev, adattach, addetach,
	nodev,		/* reset */
	&ad_cb, NULL,	/* cb_ops, bus_ops */
	NULL		/* power */
};

static struct modldrv modldrv = {
	&mod_driverops, "AoE disk driver v1.4", &ad_ops
};

static struct modlinkage modlinkage = {
	MODREV_1, &modldrv, NULL
};

/*
 * module loading and unloading
 */
int
_init()
{
	int e;

	ddi_soft_state_init(&aoedrives, sizeof(Adrive), 0);
	ddi_soft_state_init(&aoechannels, sizeof(Achan), 0);
	if ((e = mod_install(&modlinkage)) != DDI_SUCCESS) {
		cmn_err(CE_CONT, "?aoed: _init: mod_install error %d\n", e);
		ddi_soft_state_fini(&aoechannels);
		ddi_soft_state_fini(&aoedrives);
	}
	return (e);
}

int
_fini()
{
	int e, i;
	Achan *cp;
	timeout_id_t tid;

	if ((e = mod_remove(&modlinkage)) != DDI_SUCCESS) {
		cmn_err(CE_CONT, "?aoed: _fini: mod_remove error %d\n", e);
		return (e);
	}
	for (i = 0; i < aoemaxchan; i++) {
		if ((cp = ddi_get_soft_state(aoechannels, i)) == NULL)
			continue;
		aoecomm_initdriver(i, NULL);
		if ((cp->flag & ACLINIT) != 0) {
			mutex_enter(&cp->lock);
			tid = cp->tid;
			cp->tid = 0;
			mutex_exit(&cp->lock);
			if (tid)
				untimeout(tid);
			mutex_destroy(&cp->lock);
		}
	}
	ddi_soft_state_fini(&aoechannels);
	ddi_soft_state_fini(&aoedrives);
	return (e);
}

int
_info(mp)
struct modinfo *mp;
{
	return (mod_info(&modlinkage, mp));
}

#define	NOID	(-1)

static int
adattach(dip, cmd)
dev_info_t *dip;
ddi_attach_cmd_t cmd;
{
	Adrive *ep;
	Achan *cp;
	int e, i;
	char *pname;
	char rname[5 + 4 + 1];	/* big enough for "whole,raw\0" */

	if (adrvname == NULL)
		adrvname = ddi_get_name(dip);
	i = ddi_get_instance(dip);
	if (cmd != DDI_ATTACH) {
		cmn_err(CE_CONT, "?%s%d: attach command %d not implemented\n",
			adrvname, i, cmd);
		return (DDI_FAILURE);
	}
	if ((e = ddi_soft_state_zalloc(aoedrives, i)) != 0) {
		cmn_err(CE_CONT, "?%s%d: attach: state alloc failed (%d)\n",
			adrvname, i, e);
		return (DDI_FAILURE);
	}
	ep = ddi_get_soft_state(aoedrives, i);
	ep->dip = dip;
	ep->instance = i;
	ep->devmajor = UNKMAJOR;	/* used only for adrw */
	if (solversion() < 10)
		ep->cflags |= ACOLDDKEFI;
	if ((ep->chan = ddi_prop_get_int(DDI_DEV_T_ANY, dip, 0, "aoechan", NOID)) == NOID
	||  (ep->major = ddi_prop_get_int(DDI_DEV_T_ANY, dip, 0, "aoemaj", NOID)) == NOID
	||  (ep->minor = ddi_prop_get_int(DDI_DEV_T_ANY, dip, 0, "aoemin", NOID)) == NOID) {
		cmn_err(CE_CONT, "?%s%d: attach: aoechan, aoemaj, or aoemin property missing\n",
			adrvname, ep->instance);
		ddi_soft_state_free(aoedrives, ep->instance);
		return (DDI_FAILURE);
	}
	if ((ep->mintime = ddi_prop_get_int(DDI_DEV_T_ANY, dip, 0, "timeout", 0)) > 0)
		ep->mintime = drv_usectohz(ep->mintime*1000);
	else
		ep->mintime = TMO_RWMIN;
	if ((ep->maxmaxbuf = ddi_prop_get_int(DDI_DEV_T_ANY, dip, 0, "maxbuf", 0)) <= 0)
		ep->maxmaxbuf = MAXMAXBUF;
	ep->confhd = ddi_prop_get_int(DDI_DEV_T_ANY, dip, 0, "hd", 0);
	ep->confsec = ddi_prop_get_int(DDI_DEV_T_ANY, dip, 0, "sec", 0);
	ep->maxdata = ddi_prop_get_int(DDI_DEV_T_ANY, dip, 0, "maxdata", 0);
	if (ep->maxdata < AOEASEC)
		ep->maxdata = 0;
	else
		ep->cflags |= ACHARDMAX;
	if (ddi_prop_get_int(DDI_DEV_T_ANY, dip, 0, "disksort", 1) != 0)
		ep->cflags |= ACDISKSORT;
	if ((cp = ddi_get_soft_state(aoechannels, ep->chan)) == NULL) {
		if (ddi_soft_state_zalloc(aoechannels, ep->chan) != 0
		||  (cp = ddi_get_soft_state(aoechannels, ep->chan)) == NULL) {
			cmn_err(CE_CONT, "?%s%d: attach: cannot make aoechan %d\n",
				adrvname, ep->instance, ep->chan);
			ddi_soft_state_free(aoedrives, ep->instance);
			return (DDI_FAILURE);
		}
	}
	if (aoemaxchan <= ep->chan)
		aoemaxchan = ep->chan + 1;
	if ((cp->flag & ACLINIT) == 0) {
		mutex_init(&cp->lock, NULL, MUTEX_DRIVER, NULL);
		cp->flag |= ACLINIT;
	}
	for (i = 0; i < NPART; i++) {
		pname = adpname(i);
		if (createnode(dip, pname, S_IFBLK,
		    makeminor(ep->instance, i), ep->instance) < 0)
			return (DDI_FAILURE);
		snprintf(rname, sizeof(rname), "%s,raw", pname);
		if (createnode(dip, rname, S_IFCHR,
		    makeminor(ep->instance, i), ep->instance) < 0)
			return (DDI_FAILURE);
	}
	if ((e = ddi_prop_create(DDI_DEV_T_NONE, dip, 0, DDI_KERNEL_IOCTL, "", 0)) != DDI_SUCCESS)
		cmn_err(CE_CONT, "?%s%d: attach: warning: cannot make kernel-ioctl property (%d)",
			adrvname, ep->instance, e);
	mutex_init(&ep->lock, NULL, MUTEX_DRIVER, NULL);
	cv_init(&ep->cv, NULL, CV_DRIVER, NULL);
	ddi_report_dev(dip);
	return (DDI_SUCCESS);
}

/*
 * subroutine to create a minor node,
 * with error recovery
 * separated out mainly for readability
 */
static int
createnode(dip, name, type, min, inst)
dev_info_t *dip;
char *name;
int type;
minor_t min;
int inst;
{
	int e;

#ifndef ENUMERATED_DEV
#define	ENUMERATED_DEV	0
#endif
	if ((e = ddi_create_minor_node(dip, name, type,
	    min, DDI_PSEUDO, ENUMERATED_DEV)) != DDI_SUCCESS) {
		cmn_err(CE_CONT, "?%s%d: attach: create node %s failed (%d)\n",
			adrvname, inst, name, e);
		ddi_remove_minor_node(dip, NULL);
		ddi_soft_state_free(aoedrives, inst);
		return (-1);
	}
	return (1);
}

static int
addetach(dip, cmd)
dev_info_t *dip;
ddi_detach_cmd_t cmd;
{
	Adrive *ep;
	int i;
	Acmd *dp;

	i = ddi_get_instance(dip);
	if (cmd != DDI_DETACH) {
		cmn_err(CE_CONT, "?%s%d: detach: command %d not implemented\n",
			adrvname, i, cmd);
		return (DDI_FAILURE);
	}
	if ((ep = ddi_get_soft_state(aoedrives, i)) == NULL) {
		cmn_err(CE_CONT, "?%s%d: detach: get state failed\n",
			adrvname, i);
		return (DDI_FAILURE);
	}
	ddi_remove_minor_node(dip, NULL);
	mutex_enter(&ep->lock);
	if (ep->fact != NULL) {
		cmn_err(CE_CONT, "?%s%d: detach: buffers outstanding\n",
			adrvname, i);
		mutex_exit(&ep->lock);
		return (DDI_FAILURE);	/* shouldn't happen */
	}
	adkilldrive(ep->chan, ep);
	acclear(ep);
	mutex_exit(&ep->lock);
	if (ep->kp)
		kstat_delete(ep->kp);
	if (ep->knp)
		kstat_delete(ep->knp);
	if (ep->kcp)
		kstat_delete(ep->kcp);
	for (i = 0; i < NPART; i++)
		if (ep->kpp[i])
			kstat_delete(ep->kpp[i]);
	cv_destroy(&ep->cv);
	mutex_destroy(&ep->lock);
	ddi_soft_state_free(aoedrives, ep->instance);
	return (DDI_SUCCESS);
}

static int
adgetinfo(dip, cmd, arg, resp)
dev_info_t *dip;
ddi_info_cmd_t cmd;
void *arg;
void **resp;
{
	Adrive *ep;
	dev_t devarg;
	int i;

	switch (cmd) {
	case DDI_INFO_DEVT2INSTANCE:
		devarg = (dev_t)arg;
		*resp = (void *)(uintptr_t)DEVTOINST(getminor(devarg));
		return (DDI_SUCCESS);

	case DDI_INFO_DEVT2DEVINFO:
		devarg = (dev_t)arg;
		i = DEVTOINST(getminor(devarg));
		if ((ep = ddi_get_soft_state(aoedrives, i)) == NULL) {
			cmn_err(CE_CONT, "?%s: getinfo: dev %d,%d -> info: no state\n",
				adrvname, getmajor(devarg), getminor(devarg));
			return (DDI_FAILURE);
		}
		*resp = (void *)ep->dip;
		return (DDI_SUCCESS);

	default:
		cmn_err(CE_CONT, "?%s: getinfo unknown cmd %d, arg 0x%lx\n",
			adrvname, cmd, (long)arg);
		return (DDI_FAILURE);
	}
}

/*
 * device operations
 */

static int
adopen(devp, flag, type, crp)
dev_t *devp;
int flag, type;
cred_t *crp;
{
	Adrive *ep;
	int pno, e;
	diskaddr_t nb;
	longlong_t nbyte;

	if ((ep = devtodrive(*devp, "open")) == NULL)
		return (ENXIO);
	mutex_enter(&ep->lock);
	ep->devmajor = getmajor(*devp);
	/*
	 * forbid open if drive not enabled
	 * do it here despite the check in driveinit,
	 * so disable prevents future opens even if
	 * some subdevice remains open
	 */
	if (aoecomm_devenab(ep->chan, ep->major, ep->minor, ADQUERY) == 0) {
		mutex_exit(&ep->lock);
		return (ENXIO);
	}
	pno = DEVTOPART(*devp);
	mkkstats(ep, pno);
	/*
	 * if non-blocking open, don't wait for regular open work,
	 * but call driveinit anyway
	 * to reset the contents of ep
	 * if blocking, openwait calls driveinit if necessary
	 */
	if (flag & (FNDELAY|FNONBLOCK)) {
		if (ep->state == ADCLOSED)
			(void) driveinit(ep, 1);
		nb = 0;
	} else {
		if ((e = openwait(ep)) != 0) {
			mutex_exit(&ep->lock);
			return (e);
		}
		if ((nb = ep->ptab[pno].len) <= 0) {
			mutex_exit(&ep->lock);
			return (ENXIO);
		}
	}
	switch (type) {
	case OTYP_BLK:
		ep->oblk |= 1<<pno;
		break;

	case OTYP_CHR:
		ep->ochr |= 1<<pno;
		break;

	case OTYP_LYR:
		ep->olyr++;
		break;

	default:
		cmn_err(CE_CONT, "?%s%d: internal: unexpected open type %d\n",
			adrvname, ep->instance, type);
		mutex_exit(&ep->lock);
		return (ENXIO);
	}
	if (nb > 0) {
		e = ddi_prop_update_int(*devp, ep->dip, "nblocks", nb);
		if (e != DDI_PROP_SUCCESS)
			cmn_err(CE_CONT, "?%s%d: make nblocks prop: error %d\n",
				adrvname, ep->instance, e);
		e = putllprop(*devp, ep->dip, "Nblocks", nb);
		if (e != DDI_PROP_SUCCESS)
			cmn_err(CE_CONT, "?%s%d: make Nblocks prop: error %d\n",
				adrvname, ep->instance, e);
		nbyte = nb * DEV_BSIZE;
		e = ddi_prop_update_int(*devp, ep->dip, "size", nbyte);
		if (e != DDI_PROP_SUCCESS)
			cmn_err(CE_CONT, "?%s%d: make size prop: error %d\n",
				adrvname, ep->instance, e);
		e = putllprop(*devp, ep->dip, "Size", nbyte);
		if (e != DDI_PROP_SUCCESS)
			cmn_err(CE_CONT, "?%s%d: make Size prop: error %d\n",
				adrvname, ep->instance, e);
	}
	mutex_exit(&ep->lock);
	return (0);
}

/*
 * wait for open to complete
 * called from the open routine if not a non-blocking open;
 * may be called from other places if open didn't block
 */
static int
openwait(ep)
Adrive *ep;
{
	int i;

	if (ep->state == ADCLOSED) {
		if (driveinit(ep, 0) < 0)
			return (ENXIO);
	}
	if (ep->state < ADGATA) {
		if (cv_wait_sig(&ep->cv, &ep->lock) == 0) {
			acclear(ep);
			ep->state = ADCLOSED;
			return (EINTR);
		} else if (ep->state < ADGATA) {
			acclear(ep);
			ep->state = ADCLOSED;
			return (ENXIO);
		}
	}
	if (ep->state == ADGATA) {
		setdevid(ep);
		ep->state = ADWPTAB;
		mutex_exit(&ep->lock);
		i = readptab(ep);
		mutex_enter(&ep->lock);
		if (i < 0) {	/* couldn't even fake one */
			acclear(ep);
			ep->state = ADCLOSED;
			return (ENXIO);
		}
		ep->state = ADREADY;
		mkstrprop(ep, ep->atamodel, "identity-model");
		mkstrprop(ep, ep->atafwver, "identity-fwver");
		mkstrprop(ep, ep->ataserial, "identity-serial");
		cv_broadcast(&ep->cv);
	}
	if (ep->state < ADREADY) {
		if (cv_wait_sig(&ep->cv, &ep->lock) == 0) {
			acclear(ep);
			ep->state = ADCLOSED;
			return (EINTR);
		} else if (ep->state < ADREADY) {
			acclear(ep);
			ep->state = ADCLOSED;
			return (ENXIO);
		}
	}
	return (0);
}

/*
 * read partition table
 * return -1 if couldn't even try;
 * 0 if non found, fake version supplied;
 * 1 if real version found
 */

static int
readptab(ep)
Adrive *ep;
{
	mutex_enter(&ep->lock);
	if (ep->state < ADGATA || ep->nblocks <= 0) {
		mutex_exit(&ep->lock);
		return (-1);
	}
	ep->vflags &=~ AVLABEL;
	ep->slaboff = 0;
	ep->slabsize = 0;
	mutex_exit(&ep->lock);
	if (fetchvtoc(ep) > 0)
		return (1);
	if (fetchefi(ep) > 0)
		return (1);
	if (fetchdos(ep) > 0) {
#if defined(_FIRMWARE_NEEDS_FDISK)
		if (ep->slabsize > 0 && fetchvtoc(ep) <= 0)
			fakevtoc(ep);
#endif
		return (1);
	}
	ep->slaboff = 0;	/* in case scrambled by fetchdos */
	ep->slabsize = ep->nblocks;
	admsg(NODEV, ep, "no disk label\n");
	if (fakevtoc(ep) > 0)
		return (0);
	return (1);
}

/*
 * trim trailing blanks from str,
 * then set up as the named property
 * str is assumed to be of length < MAXSTRLEN
 * and NUL-terminated
 */

#define	MAXSTRLEN	50

static void
mkstrprop(ep, str, prop)
Adrive *ep;
char *str;
char *prop;
{
	char buf[MAXSTRLEN+1];
	char *p;
	int e;

	strncpy(buf, str, MAXSTRLEN);
	buf[MAXSTRLEN] = 0;
	for (p = buf + strlen(buf) - 1; p > buf; --p)
		if (*p != ' ')
			break;
	*++p = 0;
	for (p = buf; *p && *p == ' '; p++)
		;
	e = ddi_prop_update_string(DDI_DEV_T_NONE, ep->dip, prop, p);
	if (e != DDI_PROP_SUCCESS)
		cmn_err(CE_CONT, "?%s%d cannot make prop %s: error %d\n",
			adrvname, ep->instance, prop, e);
}

/*
 * make device-id token
 * -- DEVID_ATA_SERIAL seems to be the closest match
 * to what we really are; there's no generic DEVID_ATA
 * -- ID data is model "=" serial, with blanks trimmed
 */
static void
setdevid(ep)
Adrive *ep;
{
	char idbuf[LEN_ATAMODEL + 1 + LEN_ATASERIAL];
	char *p, *q;
	int e;

	q = idbuf;
	for (p = ep->atamodel; *p && *p == ' '; p++)
		;
	for (; *p; p++)
		*q++ = *p;
	while (q > idbuf && q[-1] == ' ')
		q--;
	*q++ = '=';
	for (p = ep->ataserial; *p; p++)
		*q++ = *p;
	while (q > idbuf && q[-1] == ' ')
		q--;
	if (ep->vflags & AVDEVID) {
		ep->vflags &=~ AVDEVID;
		ddi_devid_unregister(ep->dip);
		ddi_devid_free(ep->devid);
	}
	e = ddi_devid_init(ep->dip, DEVID_ATA_SERIAL, q - idbuf, idbuf, &ep->devid);
	if (e != DDI_SUCCESS) {
#if NOTDEF		/* < Solaris 9: DEVID_ATA_SERIAL doesn't exist */
		admsg(NODEV, ep, "cannot create devid: error %d\n", e);
#endif
		return;
	}
	e = ddi_devid_register(ep->dip, ep->devid);
	if (e != DDI_SUCCESS) {
		ddi_devid_free(ep->devid);
		admsg(NODEV, ep, "cannot register devid: error %d\n", e);
		return;
	}
	ep->vflags |= AVDEVID;
}

/*
 * create kstat data if needed
 * for this drive and partition
 * done in open so iostat output
 * won't be polluted by a lot of disks
 * that were attached but never used
 * the kstats are deleted in the detach code
 */

/*
 * this table had better be in sync
 * with the list in aoed.h
 */
static char *knnames[NADKS] = {
/* 0 */	"atasnd",
/* 1 */	"atarcv",
/* 2 */	"errwmsg",
/* 3 */	"errtimeout",
/* 4 */	"erraoe",
/* 5 */	"errata",
/* 6 */	"errtrunc",
/* 7 */	"retries",
/* 8 */ "softerr",
};

static char *kcnames[NADKC] = {
/* 0 */	"cylsize",
/* 1 */ "maxdata",
/* 2 */ "mibsize",
/* 3 */ "maxbuf",
/* 4 */	"timeout",
};

static void
mkkstats(ep, pno)
Adrive *ep;
int pno;
{
	char name[100];	/* room enough for "drivername[,stats]\0" */
	int i;
	kstat_named_t *np;

	if (ep->kp == NULL) {
		snprintf(name, sizeof(name), "%s%d", adrvname, ep->instance);
		ep->kp = kstat_create(adrvname, ep->instance, name, "disk",
			KSTAT_TYPE_IO, 1, KSTAT_FLAG_PERSISTENT);
		instkstat(ep, ep->kp, name);
	}
	if (ep->knp == NULL) {
		snprintf(name, sizeof(name), "%s%d,stats", adrvname, ep->instance);
		ep->knp = kstat_create(adrvname, ep->instance, name, "misc",
			KSTAT_TYPE_NAMED, NADKS, 0);
		if (ep->knp) {
			np = ep->knp->ks_data;
			for (i = 0; i < NADKS; i++) {
				if (knnames[i] == NULL)	/* safety */
					continue;
				kstat_named_init(&np[i], knnames[i],
					KSTAT_DATA_ULONG);
			}
		}
		instkstat(ep, ep->knp, name);
	}
	if (ep->kcp == NULL) {
		snprintf(name, sizeof(name), "%s%d,conf", adrvname, ep->instance);
		ep->kcp = kstat_create(adrvname, ep->instance, name, "misc",
			KSTAT_TYPE_NAMED, NADKC, 0);
		if (ep->kcp) {
			np = ep->kcp->ks_data;
			for (i = 0; i < NADKC; i++) {
				if (kcnames[i] == NULL)	/* safety */
					continue;
				kstat_named_init(&np[i], kcnames[i],
					KSTAT_DATA_ULONG);
			}
		}
		instkstat(ep, ep->kcp, name);
	}
	if (ep->kpp[pno] == NULL) {
		snprintf(name, sizeof(name), "%s%d,%s",
			adrvname, ep->instance, adpname(pno));
		ep->kpp[pno] = kstat_create(adrvname, ep->instance, name, "partition",
			KSTAT_TYPE_IO, 1, KSTAT_FLAG_PERSISTENT);
		instkstat(ep, ep->kpp[pno], name);
	}
}

static void
instkstat(ep, kp, name)
Adrive *ep;
kstat_t *kp;
char *name;
{
	if (kp == NULL)
		admsg(NODEV, ep, "open: cannot make kstat %s\n", name);
	else {
		kp->ks_lock = &ep->lock;
		kstat_install(kp);
	}
}

static int
adclose(dev, flag, type, crp)
dev_t dev;
int flag, type;
cred_t *crp;
{
	Adrive *ep;

	if ((ep = devtodrive(dev, "close")) == NULL)
		return (ENXIO);
	mutex_enter(&ep->lock);
	switch (type) {
	case OTYP_BLK:
		ep->oblk &=~ (1<<DEVTOPART(dev));
		break;

	case OTYP_CHR:
		ep->ochr &=~ (1<<DEVTOPART(dev));
		break;

	case OTYP_LYR:
		if (ep->olyr > 0)	/* safety */
			ep->olyr--;
		break;
	}
	if (ep->oblk == 0 && ep->ochr == 0 && ep->olyr == 0) {
		if (ep->state > ADGATA) {
			ep->closing = 1;
			while (ep->fact || ep->q.av_forw || ep->nact > 0)
				cv_wait(&ep->cv, &ep->lock);
		}
		ep->nact = 0;		/* safety */
		acclear(ep);
		if (ep->vflags & AVDEVID) {
			ep->vflags &=~ AVDEVID;
			ddi_devid_unregister(ep->dip);
			ddi_devid_free(ep->devid);
		}
		ep->closing = 0;
		ep->state = ADCLOSED;
	}
	mutex_exit(&ep->lock);
	return (0);
}

static int
adstrategy(bp)
struct buf *bp;
{
	Adrive *ep;
	int e, part;
	long bcount;
	diskaddr_t bno, base, len;
	kstat_t *kp;

	bcount = bp->b_bcount;
	if ((bcount % DEV_BSIZE) != 0
	||  (ep = devtodrive(bp->b_edev, "strat")) == NULL) {
		bioerror(bp, ENXIO);
		biodone(bp);
		return (0);
	}
	mutex_enter(&ep->lock);
	if (ep->closing) {
		bioerror(bp, ENXIO);
		biodone(bp);
		mutex_exit(&ep->lock);
		return (0);
	}
	if (ep->state < ADGATA) {
		if ((e = openwait(ep)) != 0) {
			bioerror(bp, e);
			biodone(bp);
			mutex_exit(&ep->lock);
			return (0);
		}
	}
	if (ep->cylsize <= 0) {	/* shouldn't happen? */
		bioerror(bp, ENXIO);
		biodone(bp);
		mutex_exit(&ep->lock);
		return (0);
	}
	part = DEVTOPART(getminor(bp->b_edev));
	base = ep->ptab[part].base;
	len = ep->ptab[part].len;
	if (bp->b_lblkno + (bcount / DEV_BSIZE) > len)
		bcount = (len - bp->b_lblkno) * DEV_BSIZE;
	if (bcount == 0) {
		bp->b_resid = bp->b_bcount;
		biodone(bp);
		mutex_exit(&ep->lock);
		return (0);
	} else if (bcount < 0) {
		bioerror(bp, ENXIO);
		biodone(bp);
		mutex_exit(&ep->lock);
		return (0);
	}
	if ((bp->b_private = kmem_zalloc(sizeof(Abuf), KM_NOSLEEP)) == NULL) {
		bioerror(bp, ENOMEM);
		biodone(bp);
		mutex_exit(&ep->lock);
		return (0);
	}
	bptoabuf(bp)->bcount = bcount;
	bp_mapin(bp);
	bptoabuf(bp)->bno = bp->b_lblkno + base;
	bp->b_resid = (Blkno)bptoabuf(bp)->bno/ep->cylsize;	/* for disksort */
	if ((kp = ep->kp) != NULL)
		kstat_waitq_enter(KSTAT_IO_PTR(kp));
	if ((kp = ep->kpp[part]) != NULL)
		kstat_waitq_enter(KSTAT_IO_PTR(kp));
	if (ep->cflags & ACDISKSORT)
		disksort(&ep->q, bp);
	else {
		/*
		 * no disksort:
		 * just append to list
		 * -- disksort doesn't use q.av_back,
		 * but our use is believed harmless
		 */
		if (ep->q.av_back)
			ep->q.av_back->av_forw = bp;
		else
			ep->q.av_forw = bp;
		ep->q.av_back = bp;
		bp->av_forw = NULL;
	}
	adstart(ep);
	mutex_exit(&ep->lock);
	return (0);
}

static int
adprint(dev, str)
dev_t dev;
char *str;
{
	admsg(dev, devtodrive(dev, NULL), "%s\n", str);
	return (0);
}

static int
adread(dev, uio, crp)
dev_t dev;
struct uio *uio;
cred_t *crp;
{
	return (physio(adstrategy, NULL, dev, B_READ, minphys, uio));
}

static int
adwrite(dev, uio, crp)
dev_t dev;
struct uio *uio;
cred_t *crp;
{
	return (physio(adstrategy, NULL, dev, B_WRITE, minphys, uio));
}

/*
 * internally-used routines to read or write absolute disk addresses
 * ep had better be unlocked, and sleeping had better be allowed
 *
 * -- originally written to set up a uio with segment UIO_SYSSEG
 * and call physio, but that always returns EFAULT;
 * evidently physio doesn't expect to handle kernel-space IO
 * -- EXDEV is an intentionally crazy value to make it obvious
 * when that particular unexpected condition occurs
 * -- in kernel I/O, a short transfer is always an error;
 * use EXDEV as an intentionally-unique value, unlikely to
 * be a normal I/O error, so we can tell
 * -- in user I/O just tell the user about the short transfer
 */

int
zeroblock(ep, bno)
Adrive *ep;
diskaddr_t bno;
{
	char buf[DEV_BSIZE];
	int e;

	bzero(buf, sizeof(buf));
	if ((e = adrwkern(ep, B_WRITE, bno, buf, DEV_BSIZE)) != 0)
		admsg(NODEV, ep, "sector %lld: error zeroing: %d\n",
			bno, e);
	return (e);
}

int
adrwkern(ep, rw, bno, buf, len)
Adrive *ep;
int rw;
diskaddr_t bno;
void *buf;
int len;
{
	struct buf *bp;
	int e;

	if ((bp = getrbuf(KM_SLEEP)) == NULL)
		return (ENOMEM);
	bp->b_flags |= rw;
	bp->b_edev = makedevice(ep->devmajor,
		makeminor(ep->instance, WHOLEDISKPART));
	bp->b_lblkno = bno;
	bp->b_un.b_addr = buf;
	bp->b_bcount = len;
	bp->b_bufsize = len;	/* does this matter? */
	(void) adstrategy(bp);
	if ((e = biowait(bp)) == 0
	&&  bp->b_resid != 0)		
		e = EXDEV;
	freerbuf(bp);
	return (e);
}

int
adrwuser(ep, rw, bno, buf, len, resp)
Adrive *ep;
int rw;
diskaddr_t bno;
void *buf;
int len;
int *resp;
{
	struct uio uio;
	iovec_t iovec;
	dev_t dev;
	int err;

	bzero(&uio, sizeof(uio));
	uio.uio_iov = &iovec;
	uio.uio_iovcnt = 1;
	uio.uio_loffset = bno * DEV_BSIZE;
	uio.uio_segflg = UIO_USERSPACE;
	uio.uio_fmode = FREAD|FWRITE;
	uio.uio_resid = len;
	bzero(&iovec, sizeof(iovec));
	iovec.iov_base = buf;
	iovec.iov_len = len;
	dev = makedevice(ep->devmajor, makeminor(ep->instance, WHOLEDISKPART));
	err = physio(adstrategy, NULL, dev, rw, minphys, &uio);
	if (resp)
		*resp = uio.uio_resid;
	return (err);
}

static int
adioctl(dev, cmd, arg, mode, crp, rvalp)
dev_t dev;
int cmd;
intptr_t arg;
int mode;
cred_t *crp;
int *rvalp;
{
	int err;
	Adrive *ep;
	int i;
	uint64_t addr;
	diskaddr_t bno;
	union {
		struct dk_cinfo i;
		struct dk_allmap m;
		struct dk_allmap32 m32;
		struct part_info p;
		int32_t i32;
		int64_t i64;
		struct dk_minfo mi;
		struct dk_geom g;
	} argval;
#define	Ii	argval.i
#define	Im	argval.m
#define	Im32	argval.m32
#define	Ip	argval.p
#define	Ig	argval.g
#define	I32	argval.i32
#define	I64	argval.i64
#define	Imi	argval.mi
#define	Ig	argval.g

	if ((ep = devtodrive(dev, "ioctl")) == NULL)
		return (ENXIO);
	mutex_enter(&ep->lock);
	/*
	 * wait for drive to be ready, in case opened non-blocking
	 * in principle we could handle certain ioctls anyway,
	 * e.g. CINFO
	 * in practice, doing so causes format to think the drive exists
	 */
	if ((err = openwait(ep)) != 0) {
		mutex_exit(&ep->lock);
		return (err);
	}
	switch (cmd) {
	case DKIOCINFO:
		/* 64-bit unchanged */
		bzero(&Ii, sizeof(Ii));
		strncpy(Ii.dki_cname, adrvname, sizeof(Ii.dki_cname));
		Ii.dki_ctype = DKC_AOED;
		Ii.dki_flags = DKI_FMTVOL;
		strncpy(Ii.dki_dname, adrvname, sizeof(Ii.dki_dname));
		/*
		 * it would make more sense to report chan/major/minor
		 * as cnum/unit/slave; but:
		 * -- format(1M) requires that cnum/unit is unique
		 * and ignores later drives that report the same value
		 * (this was discovered empirically)
		 * -- to avoid other bugs in format, DKC_AOED == DKC_DIRECT
		 * so it seems wise to have a large, hopefully unique
		 * controller number in case this system
		 * has directly-attached ATA disks too
		 */
#if NOTDEF
		Ii.dki_cnum = ep->chan;
		Ii.dki_unit = ep->major;
		Ii.dki_slave = ep->minor;
#else
		Ii.dki_cnum = CTLOFFSET;
		Ii.dki_unit = ep->instance;
		Ii.dki_slave = 0;
#endif
		Ii.dki_partition = DEVTOPART(getminor(dev));
		Ii.dki_maxtransfer = AOEMAXXFR;
		if (ddi_copyout(&Ii, (caddr_t)arg, sizeof(Ii), mode) < 0)
			err = EFAULT;
		break;

	case DKIOCGAPART:
		if (ep->nblocks > DK_MAX_BLOCKS) {
			err = ENOTSUP;
			break;
		}
		if (ep->slabsize <= 0) {
			err = EIO;
			break;
		}
		if ((bno = ep->geom.dkg_nhead * ep->geom.dkg_nsect) == 0)
			bno = ep->cylsize;
		switch (mode & FMODELS) {
		case DDI_MODEL_LP64:
			bzero(&Im, sizeof(Im));
			for (i = 0; i < NSOLPART; i++) {
				Im.dka_map[i].dkl_cylno =
					(ep->ptab[i].base - ep->slaboff)/bno;
				Im.dka_map[i].dkl_nblk = ep->ptab[i].len;
			}
			if (ddi_copyout(&Im, (caddr_t)arg, sizeof(Im), mode) < 0)
				err = EFAULT;
			break;

		case DDI_MODEL_ILP32:
			bzero(&Im32, sizeof(Im32));
			for (i = 0; i < NSOLPART; i++) {
				Im32.dka_map[i].dkl_cylno =
					(ep->ptab[i].base - ep->slaboff)/bno;
				Im32.dka_map[i].dkl_nblk = ep->ptab[i].len;
			}
			if (ddi_copyout(&Im32, (caddr_t)arg, sizeof(Im32), mode) < 0)
				err = EFAULT;
			break;

		default:
			err = EINVAL;
			break;
		}
		break;

	case DKIOCSAPART:
		if (ep->nblocks > DK_MAX_BLOCKS) {
			err = ENOTSUP;
			break;
		}
		if (ep->slabsize <= 0) {
			err = EIO;
			break;
		}
		if ((bno = ep->geom.dkg_nhead * ep->geom.dkg_nsect) == 0)
			bno = ep->cylsize;
		switch (mode & FMODELS) {
		case DDI_MODEL_LP64:
			if (ddi_copyin((caddr_t)arg, &Im, sizeof(Im), mode) < 0) {
				err = EFAULT;
				break;
			}
			for (i = 0; i < NSOLPART; i++) {
				ep->ptab[i].base = Im.dka_map[i].dkl_cylno * bno;
				ep->ptab[i].base += ep->slaboff;
				ep->ptab[i].len = Im.dka_map[i].dkl_nblk;
			}
			break;

		case DDI_MODEL_ILP32:
			if (ddi_copyin((caddr_t)arg, &Im32, sizeof(Im32), mode) < 0) {
				err = EFAULT;
				break;
			}
			for (i = 0; i < NSOLPART; i++) {
				ep->ptab[i].base = Im32.dka_map[i].dkl_cylno * bno;
				ep->ptab[i].base += ep->slaboff;
				ep->ptab[i].len = Im32.dka_map[i].dkl_nblk;
			}
			break;

		default:
			err = EINVAL;
			break;
		}
		break;

	/*
	 * officially `for intel platform only'
	 * why?  it's of general use
	 */
	case DKIOCPARTINFO:
		i = DEVTOPART(getminor(dev));
		Ip.p_start = ep->ptab[i].base;
		Ip.p_length = ep->ptab[i].len;
		if (ddi_copyout(&Ip, (caddr_t)arg, sizeof(Ip), mode) < 0)
			err = EFAULT;
		break;

	case DKIOCREMOVABLE:
		switch (mode & FMODELS) {
		case DDI_MODEL_LP64:
			I64 = 0;
			if (ddi_copyout(&I64, (caddr_t)arg, sizeof(I64), mode) < 0)
				err = EFAULT;
			break;

		case DDI_MODEL_ILP32:
			I32 = 0;
			if (ddi_copyout(&I32, (caddr_t)arg, sizeof(I32), mode) < 0)
				err = EFAULT;
			break;

		default:
			err = EINVAL;
			break;
		}
		break;

	case DKIOCGMEDIAINFO:
		/* 64-bit unchanged */
		Imi.dki_media_type = DK_FIXED_DISK;
		Imi.dki_lbsize = DEV_BSIZE;
		Imi.dki_capacity = ep->nblocks;
		if (ddi_copyout(&Imi, (caddr_t)arg, sizeof(Imi), mode) < 0)
			err = EFAULT;
		break;

	case DKIOCGGEOM:
	case DKIOCG_PHYGEOM:
		/* 64-bit unchanged */
		if (ep->nblocks > DK_MAX_BLOCKS) {
			err = ENOTSUP;
			break;
		}
		bcopy(&ep->geom, &Ig, sizeof(Ig));
		if (cmd == DKIOCG_PHYGEOM) {
			bno = Ig.dkg_nhead * Ig.dkg_nsect;
			if (bno)
				Ig.dkg_pcyl = ep->nblocks / bno;
			Ig.dkg_ncyl = Ig.dkg_pcyl;
			Ig.dkg_acyl = 0;
		}
		if (ddi_copyout(&Ig, (caddr_t)arg, sizeof(Ig), mode) < 0)
			err = EFAULT;
		break;

	case DKIOCSGEOM:
		/* 64-bit unchanged */
		if (ep->nblocks > DK_MAX_BLOCKS)
			err = ENOTSUP;
		else if (ddi_copyin((caddr_t)arg, &Ig, sizeof(Ig), mode) < 0)
			err = EFAULT;
		else if (sanegeom(ep, &Ig) < 0)
			err = EINVAL;
		else {
			bcopy(&Ig, &ep->geom, sizeof(Ig));
			ep->cylsize = ep->geom.dkg_nhead * ep->geom.dkg_nsect;
		}
		break;

	/*
	 * VTOC-label-specific ioctls
	 */
	case DKIOCGVTOC:
	case DKIOCSVTOC:
		err = vtocioc(ep, cmd, arg, mode);
		break;

	/*
	 * fdisk (DOS-label) ioctls
	 */
	case DKIOCGMBOOT:
	case DKIOCSMBOOT:
		err = dosioc(ep, cmd, arg, mode);
		break;

	/*
	 * EFI-specific ioctls
	 */
	case DKIOCGETEFI:
	case DKIOCSETEFI:
	case DKIOCPARTITION:
		err = efiioc(ep, cmd, arg, mode);
		break;


	/*
	 * special ATA-specific ioctls
	 */
	case DIOCTL_RWCMD:
	case DIOCTL_GETMODEL:
	case DIOCTL_GETSERIAL:
		err = dadkioctl(ep, cmd, arg, mode);
		break;

	/*
	 * special SCSI-specific ioctls (sic)
	 */
	case USCSICMD:
		err = aduscsi(ep, cmd, arg, mode);
		break;

	default:
		err = ENOTTY;
		break;
	}
	mutex_exit(&ep->lock);
	return (err);
}

static int
sanegeom(ep, gp)
Adrive *ep;
struct dk_geom *gp;
{
	if (gp->dkg_nhead == 0 || gp->dkg_nsect == 0
	||  gp->dkg_pcyl == 0 || gp->dkg_ncyl == 0)
		return (0);
	if (gp->dkg_ncyl + gp->dkg_acyl + gp->dkg_bcyl > gp->dkg_pcyl)
		return (0);
	return (1);
}

/*
 * fudge nhead, nsect, pcyl if necessary
 * to be reasonable (modern ATA disks report bogus geometry)
 * while fitting within quirky Solaris constraints
 * in particular FAKEHD and FAKESEC were determined empirically
 * to be values acceptable to newfs
 *
 * called only if the disk is small enough to support
 * VTOC labelling; otherwise none of it matters
 */

#define	FAKEHD	64
#define	FAKESEC	890
#define	MAXCYL	65535
#define	MAXSEC	65535
#define	MAXHD	65535

void
fudgegeom(ep)
Adrive *ep;
{
	Blkno cylsize;
	Blkno nhead, nsect, pcyl;

	if (ep->confhd > 0 && ep->confsec > 0) {
		/*
		 * values supplied in driver conf file
		 * override all else
		 */
		ep->geom.dkg_nhead = ep->confhd;
		ep->geom.dkg_nsect = ep->confsec;
		ep->geom.dkg_pcyl = (Blkno)ep->nblocks/(ep->confhd*ep->confsec);
		return;
	}
	if (ep->geom.dkg_nhead == 0)
		ep->geom.dkg_nhead = FAKEHD;
	if (ep->geom.dkg_nsect == 0)
		ep->geom.dkg_nsect = FAKESEC;
	cylsize = ep->geom.dkg_nhead * ep->geom.dkg_nsect;
	if (cylsize * ep->geom.dkg_pcyl >= ep->nblocks)
		return;		/* drive returned usable values */
	nhead = FAKEHD;
	nsect = FAKESEC;
	pcyl = (Blkno)ep->nblocks / (FAKEHD * FAKESEC);
	while (pcyl > MAXCYL && nsect <= MAXSEC/2) {
		nsect *= 2;
		pcyl /= 2;
	}
	while (pcyl > MAXCYL && nhead <= MAXSEC/2) {
		nhead *= 2;
		pcyl /= 2;
	}
	if (pcyl > MAXCYL) {
		admsg(NODEV, ep, "disk size %lld too big to invent geometry\n",
			ep->nblocks);
		return;		/* with drive-supplied values */
	}
	ep->geom.dkg_nhead = nhead;
	ep->geom.dkg_nsect = nsect;
	ep->geom.dkg_pcyl = pcyl;
}
