/*
 * Copyright 2005-2007 Coraid, Inc.
 * See accompanying file doc/LICENSE for conditions
 */

/*
 * common definitions for AoE disk driver
 */

/*
 * recent gcc emits code unusable in kernel
 * for certain 64-bit operations in 32-bit mode;
 * use Blkno in those spots
 * -- use it only where it matters, and define
 * GCCBOTCH only when using the troublesome compiler;
 * otherwise the driver may eventually fail on disks
 * of size > 2^32 sectors (1TB)
 */
#ifndef GCCBOTCH
typedef diskaddr_t Blkno;
#else
typedef daddr32_t Blkno;
#endif

typedef unsigned int Tag;

/*
 * minor device number breakout
 *	there are NPART minor devices:
 *	NSOLPART VTOC or EFI slices
 *	the WHOLEDISKPART partition, not taken from disk label
 *	NDOSPART DOS partitions
 * --it is assumed that 1<<(NPART) fits in Pbits,
 * the type used for oblk/ochr
 * --to fit a silly constraint in Sun's read_vtoc library,
 * the SOLPARTs must be first, followed immediately
 * by the WHOLEDISKPART
 */

#define	NSOLPART	NDKMAP
#define	NDOSPART	4
#define	FIRSTSOLPART	0
#define	WHOLEDISKPART	(FIRSTSOLPART+NSOLPART)
#define	FIRSTDOSPART	(WHOLEDISKPART+1)
#define	NPART		(FIRSTDOSPART+NDOSPART)

typedef int Pbits;

#define	DEVTOINST(d)	(getminor(d)/NPART)
#define	DEVTOPART(d)	(getminor(d)%NPART)
#define	makeminor(inst, part)	(((inst)*NPART) + ((part)%NPART))

#define	UNKMAJOR	(major_t)(~0)

/*
 * a pending AoE command
 * -- assumption: AOEHEADLEN is enough to hold
 * the header (not data sector) part of any
 * AoE message to be sent as a command
 * the current value works fine for AoEr8
 * as long as there are no config strings sent
 */
#define	AOEHEADLEN	32

typedef struct Acmd {
	struct Acmd *next, *prev;
	Tag tag;
	char aoehd[AOEHEADLEN];
	int aoelen;
	struct Adrive *ep;
	struct buf *bp;		/* nonzero only for I/O commands */
	int off, len;		/* within bp */
	diskaddr_t bno;		/* first for this transfer */
	kcondvar_t *cvp;
	clock_t started, expires;
	int ntries, maxtries;
	int exptime;
	int state;		/* just for debugging */
} Acmd;

#define	CSLOOSE	0	/* on no list */
#define	CSFREE	1	/* belongs to free list */
#define	CSCHAN	2	/* belongs to channel */
/*
 * driver's own state
 * one Adrive per instance (== actual EtherDrive)
 */

typedef struct {
	diskaddr_t base;
	diskaddr_t len;
} Part;

#define	NOPTIME	16

typedef struct Adrive {
	/*
	 * general device driver state
	 */
	dev_info_t *dip;
	int instance;
	int cflags;	/* configuration constants */
	int vflags;	/* flags that may change during operation */
	ddi_devid_t devid;
	kmutex_t lock;
	kcondvar_t cv;
	major_t devmajor;
	kstat_t *kp, *knp, *kcp;
	Tag nexttag;
	/*
	 * general per-drive state
	 */
	int state;
	int closing;
	Pbits oblk, ochr;
	int olyr;
	int chan, major, minor;
	char mac[LEN_EADDR];
	Acmd *cfree;
	int mintime;
	int nact;
	/*
	 * info about AoE drive server or the drive itself
	 */
	int maxbuf;		/* AoE max outstanding messages */
	int maxmaxbuf;		/* max maxbuf value, may be reconfigured */
	int maxdata;		/* longest ATA data segment allowed */
	char ataserial[LEN_ATASERIAL+1];
	char atafwver[LEN_ATAFWVER+1];
	char atamodel[LEN_ATAMODEL+1];
	/*
	 * pending I/O operations
	 */
	struct diskhd q;	/* pending buffers; sorted by disksort */
	struct buf *fact, *lact;	/* active buffers */
	/*
	 * disk geometry and partition info
	 */
	diskaddr_t nblocks;
	diskaddr_t slaboff;	/* offset to Solaris label */
	diskaddr_t slabsize;	/* size of Solaris-label part of disk */
	Part ptab[NPART];
	kstat_t *kpp[NPART];
	/* meaningful only for VTOC label */
	Blkno confhd, confsec;	/* values from config file, if any */
	Blkno cylsize;
	struct dk_geom geom;
	struct vtoc vtoc;
} Adrive;

#define	NOID	(-1)	/* invalid channel, major, minor */
#define	MAXMAXBUF	256	/* max outstanding commands */

/*
 * states,
 * each representing a step along the setup path
 */
#define	ADCLOSED 0
#define	ADWAOE	1	/* waiting for AoE Query-Config response */
#define	ADWATA	2	/* waiting for ATA IDENTIFY reply */
#define	ADGATA	3	/* have ATA IDENTIFY; ready to read part table */
#define	ADWPTAB	4	/* waiting for part table to be read */
#define	ADREADY	5	/* all done, ready for I/O */

/*
 * cflags: set when Adrive is created,
 * may be read without locking
 */
#define	ACOLDDKEFI	01	/* use Solaris 9 dk_efi_t */
#define	ACHARDMAX	02	/* hard-specified maxdata */
#define	ACDISKSORT	04	/* use disksort; else just FIFO */

/*
 * vflags: may change; lock before reading or writing
 */
#define	AVLABEFI	01	/* has EFI label */
#define	AVLABVTOC	02	/* has VTOC label (possibly encapsulated) */
#define	AVLABDOS	04	/* has DOS partition table */

#define	AVLABEL		(AVLABEFI|AVLABVTOC|AVLABDOS)

#define	AVDEVID		0100	/* devid was created */

/*
 * one Achan per channel:
 * a few state flags,
 * queue of outstanding commands
 */
typedef struct {
	int flag;
	kmutex_t lock;
	timeout_id_t tid;
	Acmd *head, *tail;
} Achan;

#define	ACLINIT	01	/* locks were set up */
#define	ACLOPEN	02	/* receiver registered */

/*
 * one Abuf per struct buf,
 * to hold our local context
 * allocated on the fly; stored in b_private
 */

typedef struct {
	int ncmds;		/* outstanding for this buffer */
	int nextoff;
	size_t bcount;		/* total for this transfer */
	diskaddr_t bno;
} Abuf;

#define	bptoabuf(bp)	((Abuf *)((bp)->b_private))

/*
 * timeout values
 * n.b. all are in clock ticks,
 * because that's what most of the system uses
 */

#define	TMO_CLOCK	drv_usectohz(1000*20)	/* how often timer code runs */
#define	TMO_CLOCKHOLD	drv_usectohz(1000*10)	/* stall between back-to-back timeouts */
#define	TMO_OPEN	drv_usectohz(1000*1000*10)	/* for open to complete */
#define	TMO_QC		drv_usectohz(1000*500)	/* query-config response */
#define	TMO_IDENTIFY	drv_usectohz(1000*500)	/* ATA IDENTIFY */
#define	TMO_RWMIN	drv_usectohz(1000*200)	/* read/write minimum value */

#define	RETRY_QC	5
#define	RETRY_IDENTIFY	5
#define	RETRY_RW	10

/*
 * max transfer size in sectors
 * advisory only; returned by DKIOCINFO
 */

#define	AOEMAXXFR	128

/*
 * tags belonging to this driver never have TAGOFFSET set
 * (it would make more sense just to set the bit, but
 * the pre-existing Linux driver did it the other way)
 */
#define	TAGOFFSET	0x80000000

/*
 * controller/drive type, returned by some ioctls
 * -- in principle ought to be unique to aoed;
 * in practice has to be something for which there is
 * working code in format, as format is just broken
 * for disks of unrecognized type
 * CTLOFFSET is just a magic number to avoid mixing
 * our disks up with real ATA disks 
 */
#define	DKC_AOED	DKC_DIRECT
#define	CTLOFFSET	100

/*
 * aoed-specific kstats:
 *	aoedXXX,stats	per-drive statistics
 *	aoedXXX,conf	per-drive config data
 * all values are type ULOG;
 * aoed.c/mkkstats has private tables declaring the names
 */

/*
 * ADKS == ,stats
 */
#define	ADKS(ep, i)	((kstat_named_t *)((ep)->knp->ks_data))[i].value.ul

#define	ADKS_WATA	0	/* ATA messages written */
#define	ADKS_RATA	1	/* ATA messages read */
#define	ADKS_ERRWMSG	2	/* couldn't write message */
#define	ADKS_ERRTIMEOUT	3	/* request failed due to timeout */
#define	ADKS_ERRAOE	4	/* AoE-level error */
#define	ADKS_ERRATA	5	/* ATA error */
#define	ADKS_ERRTRUNC	6	/* truncated read data */
#define	ADKS_RETRIES	7	/* retried transmissions */
#define	ADKS_SOFTERR	8	/* corrected ECC errors */

#define	NADKS	9

/*
 * ADKC == ,conf
 */
#define	ADKC(ep, i)	((kstat_named_t *)((ep)->kcp->ks_data))[i].value.ul

#define	ADKC_CYLSIZE	0	/* computed cylinder size */
#define	ADKC_MAXDATA	1	/* maximum data-segment size */
#define	ADKC_MBSIZE	2	/* drive size in megabytes */
#define	ADKC_MAXBUF	3	/* AoE window size--max buffers outstanding */
#define	ADKC_TIMEOUT	4	/* ATA IO timeout, ms */

#define	NADKC	5

/*
 * internal globals
 * --use #define to rename them all to make them
 * explicit property of this driver
 * does it matter?  I don't know,
 * but I don't want to clutter the code up
 * with long prefixes either
 */

/* aoed.c */

#define	zeroblock	aoed_zeroblock
extern int zeroblock(Adrive *, diskaddr_t);
#define	adrwkern	aoed_adrwkern
extern int adrwkern(Adrive *, int, diskaddr_t, void *, int);
#define	adrwuser	aoed_adrwuser
extern int adrwuser(Adrive *, int, diskaddr_t, void *, int, int *);
#define	fudgegeom	aoed_fudgegeom
extern void fudgegeom(Adrive *);

#define	aoedrives	aoed_drives
extern void *aoedrives;
#define	aoechannels	aoed_channels
extern void *aoechannels;
#define	adrvname	aoed_drvname
extern char *adrvname;

/* adcmd.c */

#define	adlookup	aoed_adlookup
extern Acmd *adlookup(int, char *);
#define	adstore		aoed_adstore
extern void adstore(int, Acmd *);
#define	adremove	aoed_adremove
extern void adremove(int, Acmd *);
#define	adcremove	aoed_adcremove
extern void adcremove(Achan *, Acmd *);
#define	adkilldrive	aoed_adkilldrive
extern void adkilldrive(int, Adrive *);
#define	adkillbuf	aoed_adkillbuf
extern void adkillbuf(int, struct buf *);
#define	acalloc		aoed_acalloc
extern Acmd *acalloc(Adrive *, int, int);
#define	acfree		aoed_acfree
extern void acfree(Adrive *, Acmd *);
#define	acinit		aoed_acinit
extern void acinit(Adrive *, int);
#define	acclear		aoed_acclear
extern void acclear(Adrive *);

/* addos.c */

#define	fetchdos	aoed_fetchdos
extern int fetchdos(Adrive *);
#define	zapdos		aoed_zapdos
extern void zapdos(Adrive *);
#define	dosioc		aoed_dosioc
extern int dosioc(Adrive *, int, intptr_t, int);

/* adefi.c */

#define	fetchefi	aoed_fetchefi
extern int fetchefi(Adrive *);
#define	zapefi		aoed_zapefi
extern void zapefi(Adrive *);
#define	efiioc		aoed_efiioc
extern int efiioc(Adrive *, int, intptr_t, int);

/* addadk.c */

#define	dadkioctl	aoed_dadkioctl
extern int dadkioctl(Adrive *, int, intptr_t, int);

/* adio.c */

#define	driveinit	aoed_driveinit
extern int driveinit(Adrive *, int);
#define	adstart	aoed_adstart
extern void adstart(Adrive *);

/* adsubr.c */

#define	devtodrive	aoed_devtodrive
extern Adrive *devtodrive(dev_t, char *);
#define	aoeerrstr	aoed_aoeerrstr
extern char *aoeerrstr(int, char *, int);
#define	ataerrstr	aoed_ataerrstr
extern char *ataerrstr(int, char *, int);
#define	admsg		aoed_admsg
extern void admsg(dev_t, Adrive *, char *, ...);
#define	adpname		aoed_adpname
extern char *adpname(int);
#define	zaplabels	aoed_zaplabels
extern void zaplabels(Adrive *, int);
#define	putllprop	aoed_putllprop
extern int putllprop(dev_t, dev_info_t *, char *, uint64_t);
#define	solversion	aoed_solversion
extern int solversion(void);

/* adtrace.c */

#define	trace		aoed_trace
extern void trace(int, void *, int, int);

/* aduscsi.c */

#define	aduscsi		aoed_aduscsi
extern int aduscsi(Adrive *, int, intptr_t, int);

/* advtoc.c */

#define	fetchvtoc	aoed_fetchvtoc
extern int fetchvtoc(Adrive *);
#define	zapvtoc		aoed_zapvtoc
extern void zapvtoc(Adrive *);
#define	vtocioc		aoed_vtocioc
extern int vtocioc(Adrive *, int, intptr_t, int);
#define	fakevtoc	aoed_fakevtoc
extern int fakevtoc(Adrive *);
