/*
 * Copyright 2005 Coraid, Inc.
 * See accompanying file doc/LICENSE for conditions
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>	/* for open */
#include "aoecmd.h"
#include "aoeproto.h"
#include "aoecommproto.h"
#include "etherproto.h"

typedef struct {
	unsigned int chan;
	unsigned int maj;
	unsigned int min;
} Devid;

static char *ctldev = AOECTL;
static char *progname = "aoectl";

typedef struct {
	char *cmd;
	int (*func)(char **);
	char *argstr;
} Cmd;

#define	MAXCHAN	100	/* highest possible channel number */

static void usage(void);
static int cmdprobe(char **);
static int probeone(int, Devid *, char *, char *);
static int cmdset(char **);
static int cmdlist(char **);
static int listchan(int, Devid *);
static int listmajors(int, Devid *);
static int listminors(int, Devid *);
static int queryone(int, Devid *);
static int strtodev(char *, Devid *);
static char *devtostr(Devid *);
static int openctl(void);

#define	DEVIDARG	"aoechan/aoemaj/aoemin ..."

static Cmd cmds[] = {
	"probe",	cmdprobe,	DEVIDARG,
	"list",		cmdlist,	DEVIDARG,
	"enable",	cmdset,		DEVIDARG,
	"disable",	cmdset,		DEVIDARG,
	NULL
};
	
int
main(argc, argv)
int argc;
char **argv;
{
	Cmd *cp;

	if (argc <= 0)
		usage();
	progname = *argv++;
	if (*argv && strcmp(*argv, "-f") == 0) {
		argv++;
		if ((ctldev = *argv++) == NULL)
			usage();
	}
	if (*argv == NULL)
		usage();
	for (cp = cmds; cp->cmd; cp++)
		if (strcmp(*argv, cp->cmd) == 0)
			return ((*cp->func)(argv));
	usage();
	return (1);	/* bogus, usage never returns */
}

static void
usage()
{
	Cmd *cp;

	fprintf(stderr, "usage: %s [-f ctldev] cmd arg ...\n", progname);
	fprintf(stderr, "commands:\n");
	for (cp = cmds; cp->cmd; cp++)
		fprintf(stderr, "\t%s%s%s\n",
			cp->cmd, cp->argstr ? " " : "", cp->argstr ? cp->argstr : "");
	exit(1);
}

static int
cmdprobe(av)
char **av;
{
	Devid d;
	int fd;
	int i, ret;

	if ((fd = openctl()) < 0)
		return (1);
	ret = 0;
	if (av[1] == NULL) {
		d.chan = ACHWILD;
		d.maj = AOEMAJWILD;
		d.min = AOEMINWILD;
		if (probeone(fd, &d, av[0], NULL) < 0)
			ret = 1;
	} else {
		for (i = 1; av[i]; i++) {
			if (strtodev(av[i], &d) < 0) {
				fprintf(stderr, "%s %s %s: bad device spec ignored\n",
					progname, av[0], av[i]);
				ret = 1;
				continue;
			}
		if (probeone(fd, &d, av[0], av[i]) < 0)
			ret = 1;
		}
	}
	close(fd);
	return (ret);
}

static int
probeone(fd, dp, cmd, arg)
int fd;
Devid *dp;
char *cmd, *arg;
{
	char buf[AC_MESSAGE + LEN_ETHER + MINLEN_AOEI];
	char *p;
	int n, len;

	p = buf;
	len = sizeof(buf);
	memset(p, 0, len);
	toaoechar(p, AC_TYPE, ACSEND);
	toaoeshort(p, AC_LEN, len);
	toaoeshort(p, AC_CHANNEL, dp->chan);
	p += AC_MESSAGE;	/* to MAC header */
	memset(p + ETHER_DST, 0xff, LEN_EADDR);	/* broadcast address */
	/* ETHER_SRC, ETHER_PROT filled in by aoecomm */
	p += LEN_ETHER;		/* to AoE message */
	toaoechar(p, AOE_VF, aoemkvf(0));
	toaoeshort(p, AOE_MAJOR, dp->maj);
	toaoechar(p, AOE_MINOR, dp->min);
	toaoechar(p, AOE_COMMAND, AOECID);
	toaoelong(p, AOE_TAG, 0);
	toaoechar(p, AOEI_VCC, aoeimkvcc(0, AOEICGET));
	toaoeshort(p, AOEI_CLEN, 0);
	if ((n = write(fd, buf, len)) != n) {
		fprintf(stderr, "%s %s%s%s: write returned %d/%d: %s\n",
			progname, cmd, arg ? " " : "", arg ? arg : "",
			n, len, strerror(errno));
		return (-1);
	}
	return (1);
}

static int
cmdset(av)
char **av;
{
	Devid d;
	int fd;
	char buf[LEN_ACDEVENAB];
	int i, n;
	int cmd, okret, ret;

	if (strcmp(av[0], "enable") == 0) {
		cmd = ADENAB;
		okret = sizeof(buf);
	} else if (strcmp(av[0], "disable") == 0) {
		cmd = ADDISAB;
		okret = 0;
	} else
		usage();
	if ((fd = openctl()) < 0)
		return (1);
	ret = 0;
	for (i = 1; av[i]; i++) {
		if (strtodev(av[i], &d) < 0) {
			fprintf(stderr, "%s %s %s: bad device spec ignored\n",
				progname, av[0], av[i]);
			ret = 1;
			continue;
		}
		memset(buf, 0, sizeof(buf));
		toaoechar(buf, AC_TYPE, ACDEVENAB);
		toaoeshort(buf, AC_LEN, LEN_ACDEVENAB);
		toaoeshort(buf, AC_CHANNEL, d.chan);
		toaoeshort(buf, AC_MAJOR, d.maj);
		toaoechar(buf, AC_MINOR, d.min);
		toaoechar(buf, AC_ENABCMD, cmd);
		if ((n = write(fd, buf, sizeof(buf))) < 0) {
			fprintf(stderr, "%s %s %s: unexpected return from write: %d/%d: %s\n",
				progname, av[0], av[i],
				n, sizeof(buf), strerror(errno));
			ret = 1;
		} else if (n != okret) {
			fprintf(stderr, "%s: cannot %s %s\n", progname, av[0], av[i]);
			ret = 1;
		}
	}
	close(fd);
	return (ret);
}

static int
cmdlist(av)
char **av;
{
	Devid d;
	int fd;
	int i, n;

	if ((fd = openctl()) < 0)
		return (1);
	if (av[1] == NULL) {
		d.chan = ACHWILD;
		d.maj = AOEMAJWILD;
		d.min = AOEMINWILD;
		n = listchan(fd, &d);
	} else {
		n = 0;
		for (i = 1; av[i]; i++) {
			if (strtodev(av[i], &d) < 0) {
				fprintf(stderr, "%s %s %s: bad device spec ignored\n",
					progname, av[0], av[i]);
				continue;
			}
			n += listchan(fd, &d);
		}
	}
	close(fd);
	if (n == 0)
		return (1);
	return (0);
}

static int
listchan(fd, dp)
int fd;
Devid *dp;
{
	Devid nd;
	int n;

	if (dp->chan != ACHWILD)
		return (listmajors(fd, dp));
	nd.maj = AOEMAJWILD;
	nd.min = AOEMINWILD;
	n = 0;
	for (nd.chan = 0; nd.chan < MAXCHAN; nd.chan++) {
		if (queryone(fd, &nd) <= 0)
			continue;
		dp->chan = nd.chan;
		n += listmajors(fd, dp);
	}
	return (n);
}
	
static int
listmajors(fd, dp)
int fd;
Devid *dp;
{
	int n, m, min;

	if (dp->maj != AOEMAJWILD)
		return (listminors(fd, dp));
	n = 0;
	for (dp->maj = 0; dp->maj < AOEMAJMAX; dp->maj++) {
		min = dp->min;
		dp->min = AOEMINWILD;
		m = queryone(fd, dp);
		dp->min = min;
		if (m > 0)
			n += listminors(fd, dp);
	}
	return (n);
}

static int
listminors(fd, dp)
int fd;
Devid *dp;
{
	int n;

	n = 0;
	if (dp->min != AOEMINWILD) {
		if (queryone(fd, dp) > 0) {
			printf("%s\n", devtostr(dp));
			n++;
		}
	} else {
		for (dp->min = 0; dp->min < AOEMINMAX; dp->min++)
			if (queryone(fd, dp) > 0) {
				printf("%s\n", devtostr(dp));
				n++;
			}
	}
	return (n);
}

static int
queryone(fd, dp)
int fd;
Devid *dp;
{
	char buf[LEN_ACDEVENAB];
	int n;

	memset(buf, 0, sizeof(buf));
	toaoechar(buf, AC_TYPE, ACDEVENAB);
	toaoeshort(buf, AC_LEN, LEN_ACDEVENAB);
	toaoeshort(buf, AC_CHANNEL, dp->chan);
	toaoeshort(buf, AC_MAJOR, dp->maj);
	toaoechar(buf, AC_MINOR, dp->min);
	toaoechar(buf, AC_ENABCMD, ADQUERY);
	if ((n = write(fd, buf, sizeof(buf))) < 0) {
		fprintf(stderr, "%s query %d/%d/%d: unexpected return from write: %d/%d: %s\n",
			progname, dp->chan, dp->maj, dp->min,
			n, sizeof(buf), strerror(errno));
		return (-1);
	}
	return (n);
}

/*
 * crack s into chan/maj/min
 * with sanity checks,
 * and allowing wildcards for the latter two
 * or allowing missing trailing parts to be omitted
 * do not change the original s
 */

static int
strtodev(s, dp)
char *s;
Devid *dp;
{
	char *p, *ep;
	int i, n;

	if (s == NULL)
		return (-1);
	dp->chan = strtoul(s, &ep, 0);
	if (ep == s)
		dp->chan = ACHWILD;
	if (*ep == 0) {
		dp->maj = AOEMAJWILD;
		dp->min = AOEMINWILD;
		return (1);
	}
	if (*ep != '/')
		return (-1);
	s = ep + 1;
	if (*s == 0) {
		dp->maj = AOEMAJWILD;
		dp->min = AOEMINWILD;
		return (1);
	}
	if (*s == '/') {
		dp->maj = AOEMAJWILD;
		ep = s + 1;
	} else {
		dp->maj = strtoul(s, &ep, 0);
		if (ep == s)
			return (-1);
	}
	if (dp->maj > AOEMAJMAX && dp->maj != AOEMAJWILD)
		return (-1);
	if (*ep == 0) {
		dp->min = AOEMINWILD;
		return (1);
	}
	if (*ep != '/')
		return (-1);
	s = ep + 1;
	if (*s == 0) {
		dp->min = AOEMINWILD;
		return (1);
	}
	dp->min = strtoul(s, &ep, 0);
	if (ep == s || *ep != 0)
		return (-1);
	if (dp->min > AOEMINMAX && dp->min != AOEMINWILD)
		return (-1);
	return (1);
}

static char *
devtostr(dp)
Devid *dp;
{
	static char buf[100];

	snprintf(buf, sizeof(buf), "%d/%d/%d", dp->chan, dp->maj, dp->min);
	return (buf);
}

static int
openctl()
{
	int fd;

	if ((fd = open(ctldev, 1)) < 0) {
		fprintf(stderr, "%s: open %s: %s\n",
			progname, ctldev, strerror(errno));
		return (-1);
	}
	return (fd);
}
