/*
 * write a simple, standard EFI disk label
 * because Sun's tools can't
 * for non-SCSI disks
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>	/* open */
#include <sys/param.h>	/* DEV_BSIZE */
#include <sys/vtoc.h>
#include <sys/dkio.h>
#include <sys/efi_partition.h>

#define	LTSTD		0
#define	LTDEBUG		1
#define	LTNOPART	2

#define	LFUNK		0
#define	LVTOC		1
#define	LEFI		2

static char *progname = "aoelabinit";
static int force;
static int ltype = LTSTD;
static int dontwrite;
static int lformat = LFUNK;

static char **doargs(char **);
static void usage(void);
static int hasoldlabel(int, char *);
static int putlabel(int, char *, int, int);
static int putvtoc(int, char *, int);
static int zaplabels(int, diskaddr_t, char *);
static int readvtoc(int, struct vtoc *);
static int writevtoc(int, struct vtoc *, char *);
static int putefi(int, char *, int);
static void dumpvtoc(struct vtoc *);
static void dumpefi(struct dk_gpt *);
static char *ptname(int);
static void mkvtoc(struct vtoc *, diskaddr_t, int);
static void mkefi(struct dk_gpt *, int);
static void setefipart(struct dk_part *, diskaddr_t, diskaddr_t, diskaddr_t);
static char *vterror(int);
static diskaddr_t disksize(int, char *);

int
main(argc, argv)
int argc;
char **argv;
{
	int fd;
	int errs;

	if (*argv)
		progname = *argv++;
	argv = doargs(argv);
	errs = 0;
	for (; *argv; argv++) {
		if ((fd = open(*argv, O_RDONLY|O_NONBLOCK)) < 0) {
			fprintf(stderr, "%s: %s: cannot open: %s\n",
				progname, *argv, strerror(errno));
			errs++;
			continue;
		}
		if (force == 0 && hasoldlabel(fd, *argv)) {
			fprintf(stderr, "%s: %s: may have old label, use -f to force\n",
				progname, *argv);
			close(fd);
			continue;
		}
		if (putlabel(fd, *argv, lformat, ltype) < 0)
			errs++;
		close(fd);
	}
	if (errs)
		return (1);
	return (0);
}

static char **
doargs(av)
char **av;
{
	char *p;

	for (; *av && **av == '-'; av++)
	for (p = *av + 1; *p; p++)
	switch (*p) {
	case 'e':
		lformat = LEFI;
		continue;

	case 'f':
		force = 1;
		continue;

	case 'n':
		ltype = LTNOPART;
		continue;

	case 'N':
		dontwrite = 1;
		continue;

	case 's':
		ltype = LTSTD;
		continue;

	case 't':
		ltype = LTDEBUG;
		continue;

	case 'v':
		lformat = LVTOC;
		continue;

	default:
		usage();
	}
	if (*av == NULL)
		usage();
	return (av);
}

static void
usage()
{
	fprintf(stderr, "usage: %s [-efvN] rawdevice ...\n", progname);
	exit(1);
}

static int
hasoldlabel(fd, dev)
int fd;
char *dev;
{
	struct vtoc v;
	struct dk_gpt *vp;
	int e;

	if ((e = readvtoc(fd, &v)) == 0)
		return (1);
	if (e != ENOTSUP) {
		fprintf(stderr, "%s: %s: unexpected error reading old VTOC: %s\n",
			progname, dev, strerror(e));
		return (1);		/* assume old label might be there */
	}
	if ((e = efi_alloc_and_read(fd, &vp)) >= 0) {
		efi_free(vp);
		return (1);
	}
	if (e != VT_EINVAL) {
		fprintf(stderr, "%s: %s: unexpected error reading old EFI label: %s\n",
			progname, dev, vterror(e));
		return (1);
	}
	return (0);
}

static int
putlabel(fd, dev, format, type)
int fd;
char *dev;
int format;
int type;
{
	diskaddr_t dsize;

	switch (format) {
	case LVTOC:
		return (putvtoc(fd, dev, type));

	case LEFI:
		return (putefi(fd, dev, type));

	case LFUNK:
		if ((dsize = disksize(fd, dev)) == 0)
			return (-1);
		else if (dsize <= DK_MAX_BLOCKS)
			return (putvtoc(fd, dev, type));
		else
			return (putefi(fd, dev, type));

	default:
		fprintf(stderr, "%s: internal error: label format %d\n",
			dev, format);
		return (1);
	}
}

/*
 * assume the device driver supplies a default label;
 * we just adjust its partition-table contents
 * ENOTSUP means this disk cannot handle VTOC,
 * usually because too big
 */
static int
putvtoc(fd, dev, type)
int fd;
char *dev;
int type;
{
	struct vtoc v;
	struct dk_geom g;
	int i, e;
	diskaddr_t realsize, size;

	if (ioctl(fd, DKIOCGGEOM, &g) < 0) {
		fprintf(stderr, "%s: cannot get geometry parameters, so can't write VTOC: %s\n",
			dev, strerror(errno));
		return (-1);
	}
	if ((realsize = disksize(fd, dev)) == 0)
		return (-1);
	if (realsize > DK_MAX_BLOCKS) {
		fprintf(stderr, "%s: warning: %lld sectors, too big for VTOC, truncating to %lld\n",
			dev, size, (diskaddr_t)DK_MAX_BLOCKS);
		realsize = DK_MAX_BLOCKS;
	}
	size = g.dkg_nhead * g.dkg_nsect;
	size *= g.dkg_ncyl;
	if (size > realsize)
		size = realsize;
	if ((e = readvtoc(fd, &v)) != 0) {
		if (e != VT_ENOTSUP) {
			fprintf(stderr, "%s: error reading old VTOC: %s\n",
				dev, strerror(e));
			return (-1);
		}
		if ((fd = zaplabels(fd, realsize, dev)) < 0)
			return (-1);
		if ((e = readvtoc(fd, &v)) != 0) {
			fprintf(stderr, "%s: error reading old VTOC (retry): %s\n",
				dev, strerror(e));
			return (-1);
		}
	}
	memset(v.v_part, 0, sizeof(v.v_part));
	mkvtoc(&v, size, type);
	if (dontwrite)
		dumpvtoc(&v);
	else if ((e = writevtoc(fd, &v, dev)) < 0) {
		fprintf(stderr, "%s: error writing VTOC: %s\n", dev, strerror(e));
		return (-1);
	}
	return (0);
}

/*
 * zap EFI label, to allow (default) VTOC label to peek through
 * use DKIOCSETEFI to guarantee absolute block numbers;
 * also happens to wipe out any VTOC label already there
 *
 * must close and reopen so driver will use the new label too
 */
static int
zaplabels(fd, size, dev)
int fd;
diskaddr_t size;
char *dev;
{
	char buf[DEV_BSIZE];
	dk_efi_t dke;

	memset(buf, 0, sizeof(buf));
	dke.dki_length = sizeof(buf);
	dke.dki_data_64 = (uint64_t)buf;
	dke.dki_lba = 1;
	if (ioctl(fd, DKIOCSETEFI, &dke) < 0) {
		fprintf(stderr, "%s: error clearing primary EFI label: %s\n",
			dev, strerror(errno));
		return (-1);
	}
	dke.dki_lba = size - 1;
	if (ioctl(fd, DKIOCSETEFI, &dke) < 0) {
		fprintf(stderr, "%s: error clearing backup EFI label: %s\n",
			dev, strerror(errno));
		return (-1);
	}
	close(fd);
	if ((fd = open(dev, O_RDONLY|O_NONBLOCK)) < 0) {
		fprintf(stderr, "%s: failed to reopen after clearing EFI label: %s\n",
			dev, strerror(errno));
		return (-1);
	}
	return (fd);
}

/*
 * there's an official read_vtoc/write_vtoc library interface
 * but we have to bust through it to fetch geometry
 * to compute size of disk data area anyway
 */
static int
readvtoc(fd, vp)
int fd;
struct vtoc *vp;
{
	if (ioctl(fd, DKIOCGVTOC, vp) >= 0)
		return (0);
	return (errno);
}

static int
writevtoc(fd, vp, name)
int fd;
struct vtoc *vp;
char *name;
{
	if (ioctl(fd, DKIOCSVTOC, vp) >= 0)
		return (0);
	return (errno);
}

static int
putefi(fd, dev, type)
int fd;
char *dev;
int type;
{
	int e;
	struct dk_gpt *vp;

	if ((e = efi_alloc_and_init(fd, NDKMAP + 1, &vp)) != 0) {
		fprintf(stderr, "%s: %s: alloc EFI label: %s\n",
			progname, dev, vterror(e));
		return (-1);
	}
	mkefi(vp, type);
	if (dontwrite)
		dumpefi(vp);
	else if ((e = efi_write(fd, vp)) != 0) {
		fprintf(stderr, "%s: %s: write EFI label: %s\n",
			progname, dev, vterror(e));
		efi_free(vp);
		return (-1);
	}
	efi_free(vp);
	return (1);
}

static void
dumpvtoc(vp)
struct vtoc *vp;
{
	int i;

	for (i = 0; i < vp->v_nparts; i++) {
		if (vp->v_part[i].p_size == 0)
			continue;
		printf("%2d %10s %12ld to %12ld = %12ld\n",
			i, ptname(vp->v_part[i].p_tag),
			vp->v_part[i].p_start,
			vp->v_part[i].p_start + vp->v_part[i].p_size - 1,
			vp->v_part[i].p_size);
	}
}

static void
dumpefi(vp)
struct dk_gpt *vp;
{
	int i;

	for (i = 0; i < vp->efi_nparts; i++) {
		if (vp->efi_parts[i].p_size == 0)
			continue;
		printf("%2d %10s %12lld to %12lld = %12lld\n",
			i, ptname(vp->efi_parts[i].p_tag),
			vp->efi_parts[i].p_start,
			vp->efi_parts[i].p_start + vp->efi_parts[i].p_size - 1,
			vp->efi_parts[i].p_size);
	}
}

static struct {
	int v;
	char *n;
} ptnames[] = {
	V_UNASSIGNED, "unassigned",
	V_BOOT,	"boot",
	V_ROOT, "root",
	V_SWAP, "swap",
	V_USR, "usr",
	V_BACKUP, "backup",
	V_STAND, "stand",
	V_VAR, "var",
	V_HOME, "home",
	V_ALTSCTR, "altsctr",
	V_CACHE, "cache",
	V_RESERVED, "reserved",
	0, NULL
};


static char *
ptname(pt)
int pt;
{
	int i;
	static char buf[40];

	for (i = 0; ptnames[i].n; i++)
		if (ptnames[i].v == pt)
			return (ptnames[i].n);
	snprintf(buf, sizeof(buf), "unknown-%d", pt);
	return (buf);
}

static void
mkvtoc(vp, size, type)
struct vtoc *vp;
diskaddr_t size;
int type;
{
	switch (type) {
	default:
	case LTNOPART:
	case LTDEBUG:		/* for now */
		break;

	case LTSTD:
		vp->v_part[0].p_tag = V_ROOT;
		vp->v_part[0].p_start = 0;
		vp->v_part[0].p_size = size;
		break;
	}
	/*
	 * partition 2 is always the whole disk,
	 * says Solaris convention
	 */
	vp->v_part[2].p_tag = V_BACKUP;
	vp->v_part[2].p_start = 0;
	vp->v_part[2].p_size = size;
}

/*
 * LTSTD lifted from format
 */

#define	RESSIZE	(16 * 1024)
#define	MB	(1024*(1024/512ULL))	/* 1MB worth of 512-byte sectors */
#define	GB	(1024*MB)		/* 1GB */
#define	TB	(1024*GB)		/* 1TB */
#define	TTESTSZ	(20*MB)			/* test slice size */

static void
mkefi(vp, type)
struct dk_gpt *vp;
int type;
{
	diskaddr_t lo, hi, b;

	lo = vp->efi_first_u_lba;
	hi = vp->efi_last_u_lba + 1 - RESSIZE;
	switch (type) {
	default:
	case LTNOPART:
		vp->efi_parts[NDKMAP].p_start = hi;
		vp->efi_parts[NDKMAP].p_size = RESSIZE;
		vp->efi_parts[NDKMAP].p_tag = V_RESERVED;
		break;

	case LTSTD:
		vp->efi_parts[0].p_start = lo;
		vp->efi_parts[0].p_size = hi - lo;
		vp->efi_parts[0].p_tag = V_BACKUP;
		vp->efi_parts[NDKMAP].p_start = hi;
		vp->efi_parts[NDKMAP].p_size = RESSIZE;
		vp->efi_parts[NDKMAP].p_tag = V_RESERVED;
		break;

	case LTDEBUG:
		setefipart(&vp->efi_parts[0], lo, TTESTSZ/2, hi);
		setefipart(&vp->efi_parts[1], TB - TTESTSZ/2, TTESTSZ, hi);
		setefipart(&vp->efi_parts[2], (2 * TB) - TTESTSZ/2, TTESTSZ, hi);
		/*
		 * s3 s4 magic locations
		 * determined by constraints of sparse file in ext3 fs
		 * for testing with vblade
		 */
		setefipart(&vp->efi_parts[3], TB - (4*GB) - TTESTSZ/2, TTESTSZ, hi);
		setefipart(&vp->efi_parts[4], (2*TB) - (4*GB) - TTESTSZ/2, TTESTSZ, hi);
		vp->efi_parts[NDKMAP].p_start = hi;
		vp->efi_parts[NDKMAP].p_size = RESSIZE;
		vp->efi_parts[NDKMAP].p_tag = V_RESERVED;
		break;
	}
}

static void
setefipart(pp, start, size, lim)
struct dk_part *pp;
diskaddr_t start, size, lim;
{
	if (start >= lim)
		return;
	pp->p_start = start;
	if (start + size <= lim)
		pp->p_size = size;
	else
		pp->p_size = lim - start;
	pp->p_tag = V_BACKUP;
}

static char *
vterror(e)
int e;
{
	static char oopsbuf[40];

	switch (e) {
	case VT_EIO:
		return ("I/O error");

	case VT_ERROR:
		return ("Unknown error");

	case VT_EINVAL:
		return ("Invalid label");

	case VT_ENOTSUP:
		return ("Not supported");

	case 0:
		return ("No error");

	default:
		snprintf(oopsbuf, sizeof(oopsbuf), "Unexpected error %d", e);
		return (oopsbuf);
	}
}

static diskaddr_t
disksize(fd, dev)
int fd;
char *dev;
{
	struct dk_minfo m;

	if (ioctl(fd, DKIOCGMEDIAINFO, &m) < 0) {
		fprintf(stderr, "%s: %s: cannot get media info: %s\n",
			progname, dev, strerror(errno));
		return (0);
	}
	return (m.dki_capacity);
}
