/*
 * Copyright 2005 Coraid, Inc.
 * See accompanying file doc/LICENSE for conditions
 */

/*
 * monitor AoE subsystem:
 * enable devices as they identify themselves,
 * log unexpected messages
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>	/* for open */
#include <signal.h>
#include "aoecmd.h"
#include "aoecommproto.h"
#include "aoeproto.h"
#include "etherproto.h"

static char *ctldev = AOECTL;
static char *mondev = AOEMON;
char *progname = "aoemon";
static int noenabs;
static int oknomon;
int debug;
int verbose;

static char **doargs(char **);
static void usage(void);
static int detach(void);
static int commmsg(char *, int, int);
static char *bladeaddr(int, char *, int);
static int aoercv(int, char *, int, int);

int
main(argc, argv)
int argc;
char **argv;
{
	int cfd, mfd;
	char *p;
	int n, un;
	char buf[LEN_ACMAX*10];

	if (argc > 0)
		progname = *argv++;
	if ((p = strrchr(progname, '/')) != NULL)
		progname = p + 1;
	argv = doargs(argv);
	if (*argv)
		usage();
	loginit(progname, debug);
	if ((mfd = open(mondev, 0)) < 0) {
		if (oknomon == 0)
			logmsg(LE|LS, "cannot open %s: %s", mondev, strerror(errno));
		return (1);
	}
	if ((cfd = open(ctldev, 1)) < 0) {
		logmsg(LE|LS, "cannot open %s: %s", ctldev, strerror(errno));
		return (1);
	}
	if (debug == 0 && detach() < 0)
		return (1);
	while ((n = read(mfd, buf, sizeof(buf))) > 0) {
		p = buf;
		while (n > 0 && (un = commmsg(p, n, cfd)) > 0) {
			p += un;
			n -= un;
		}
	}
	logmsg(LE, "read error: %d/%d: %s", n, sizeof(buf), strerror(errno));
	return (1);
}

static char **
doargs(av)
char **av;
{
	char *p;

	for (; *av && **av == '-'; av++)
	for (p = &av[0][1]; *p; p++)
	switch (*p) {
	case 'c':
		if ((ctldev = *++av) == NULL)
			usage();
		continue;

	case 'd':
		debug = 1;
		continue;

	case 'e':
		oknomon = 1;
		continue;

	case 'm':
		if ((mondev = *++av) == NULL)
			usage();
		continue;

	case 'n':
		noenabs = 1;
		continue;

	case 'v':
		verbose = 1;
		continue;

	default:
		usage();
	}
	return (av);
}

static void
usage()
{
	fprintf(stderr, "usage: %s [-ednv] [-c ctldev] [-m mondev]\n", progname);
	exit(1);
}

/*
 * hive off to run in the background
 */

#define	MAXSTDFILE	2

static int
detach()
{
	int i;

	if ((i = fork()) < 0) {
		logmsg(LE|LS, "cannot fork to detach: %s", strerror(errno));
		return (-1);
	}
	if (i != 0) {
		/*
		 * parent
		 */
		exit(0);
	}
	/*
	 * child
	 */
	signal(SIGINT, SIG_IGN);
	signal(SIGQUIT, SIG_IGN);
	for (i = 0; i <= MAXSTDFILE; i++)
		close(i);
	setpgrp();
	return (1);
}

/*
 * process next message from monitor device
 * should be an AoE message from the network
 * return the length consumed
 */

static char *logtype[] = {
	"Unexpected message",
	"Ill-formed message",
	"Can't send command",
	"Command timed out",
	"Received AoE error",
	"Received ATA error",
};

#define	MAXLOGTYPE	(sizeof(logtype)/sizeof(logtype[0]))

static int
commmsg(buf, len, cfd)
char *buf;
int len;
int cfd;
{
	int e, t;
	int mlen;
	char oopsmsg[50];
	char *p;

	if (len < LEN_ACHEAD) {
		logmsg(LN, "invalid length %d", len);
		loghex(LN, buf, len);
		return (0);
	}
	if ((mlen = fraoeshort(buf, AC_LEN)) > len
	||  mlen < LEN_ACHEAD) {
		logmsg(LN, "invalid stored length %d/%d", mlen, len);
		loghex(LN, buf, len);
		return (0);
	}
	if (fraoechar(buf, AC_TYPE) != ACLOG) {
		logmsg(LN, "receive: unknown message type %d len %d",
			fraoechar(buf, AC_TYPE), mlen);
		loghex(LN, buf, mlen);
		return (mlen);
	}
	if (mlen < AC_MESSAGE) {
		logmsg(LN, "ACLOG too short: %d", mlen);
		loghex(LN, buf, mlen);
		return (mlen);
	}
	if ((e = fraoechar(buf, AC_NLOST)) != 0)
		logmsg(LE, "%d log messages lost", e);
	t = fraoechar(buf, AC_LTYPE);
	e = LN;
	if (t == ACLUNSOL && noenabs == 0) {
		e = aoercv(fraoeshort(buf, AC_CHANNEL),
			buf + AC_MESSAGE, mlen - AC_MESSAGE, cfd);
		if (e >= 0) {
			if (verbose == 0 && debug == 0)
				return (mlen);
			e = LD;
		}
	}
	p = bladeaddr(fraoeshort(buf, AC_CHANNEL),
		buf + AC_MESSAGE, mlen - AC_MESSAGE);
	if (t >= 0 && t < MAXLOGTYPE)
		logmsg(e, "%s: %s:", p, logtype[t]);
	else
		logmsg(e, "%s: ACLOG type %d:", p, t);
	logpkt(e, buf + AC_MESSAGE, mlen - AC_MESSAGE);
	return (mlen);
}

/*
 * return a string for the AoE blade address,
 * if buf appears to contain an AoE message
 */
static char *
bladeaddr(chan, buf, len)
int chan;
char *buf;
int len;
{
	static char addr[100];

	len -= LEN_ETHER;
	buf += LEN_ETHER;
	if (len < MINLEN_AOE || aoeversion(fraoechar(buf, AOE_VF)) != AOEV1)
		snprintf(addr, sizeof(addr), "%d/unknown", chan);
	else
		snprintf(addr, sizeof(addr), "%d/%d/%d",
			chan, fraoeshort(buf, AOE_MAJOR), fraoechar(buf, AOE_MINOR));
	return (addr);
}

/*
 * return > 0 if message consumed (it was Query-Config);
 * == 0 if OK but not wanted;
 * < 0 if somehow improper
 */
static int
aoercv(chan, buf, len, cfd)
int chan;
char *buf;
int len;
int cfd;
{
	char *p;
	int alen;
	int n;
	int major, minor;
	int maxdata, maxbuf;
	int clen;
	char ebuf[LEN_ACDEVENAB];

	/*
	 * parse AoE Query-Config response;
	 * extract device identity
	 */
	if (len < LEN_ETHER + MINLEN_AOE) {
		logmsg(LN, "message too short, len %d", len);
		return (-1);
	}
	p = buf + LEN_ETHER;
	alen = len - LEN_ETHER;
	n = fraoechar(p, AOE_VF);
	if (aoeversion(n) != AOEV1) {
		logmsg(LN, "wrong AoE version");
		return (-1);
	}
	if (fraoechar(p, AOE_COMMAND) != AOECID)
		return (0);
	major = fraoeshort(p, AOE_MAJOR);
	minor = fraoechar(p, AOE_MINOR);
	maxdata = fraoechar(p, AOEI_MAXSEC) * AOEASEC;
	maxbuf = fraoeshort(p, AOEI_MAXBUF);
	if ((n & (AOEF_R|AOEF_E)) != AOEF_R) {
		if (verbose || debug)
			logmsg(LI, "foreign Query-Config command for %d/%d/%d",
				chan, major, minor);
		return (1);
	}
	if (alen < AOEI_CSTR
	||  (clen = fraoeshort(p, AOEI_CLEN)) > alen - AOEI_CSTR) {
	/*
	 * vblade test program generates messages with crazy CLEN,
	 * apparently truncated;
	 * but we can tolerate that because we don't use CSTR anyway
	 */
#ifdef BROKEN_VBLADE
		logmsg(LN, "Query-Config response truncated");
		return (-1);
#else
		logmsg(LN, "Query-Config invalid config-str len %d; max allowed %d",
			clen, alen - AOEI_CSTR);
		clen = 0;
#endif
	}
	if (major == AOEMAJWILD || minor == AOEMINWILD) {
		logmsg(LN, "receiver: Query-Config response with wild major/minor");
		return (-1);
	}
	/*
	 * compose and send device-enable message
	 */
	memset(ebuf, 0, sizeof(ebuf));
	toaoechar(ebuf, AC_TYPE, ACDEVENAB);
	toaoeshort(ebuf, AC_LEN, LEN_ACDEVENAB);
	toaoeshort(ebuf, AC_CHANNEL, chan);
	toaoeshort(ebuf, AC_MAJOR, major);
	toaoechar(ebuf, AC_MINOR, minor);
	toaoechar(ebuf, AC_ENABCMD, ADENAB);
	if ((n = write(cfd, ebuf, LEN_ACDEVENAB) != LEN_ACDEVENAB))
		logmsg(LE, "write to enable %d/%d/%d: returned %d/%d: %s",
			chan, major, minor, n, LEN_ACDEVENAB, strerror(errno));
	else {
		n = fraoeshort(p, AOEI_FV);
		logmsg(LI, "enabled %d/%d/%d firmware %d.%d maxbuf %d maxdata %d",
			chan, major, minor, aoefvmaj(n), aoefvmin(n),
			maxbuf, maxdata);
	}
	return (1);
}
