/*
 * Copyright 2005 Coraid, Inc.
 * See accompanying file doc/LICENSE for conditions
 */

/*
 * start up an aoe channel
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>	/* for open */
#include "aoecmd.h"
#include "aoecommproto.h"
#include "aoeproto.h"
#include "etherproto.h"

typedef struct {
	char *name;
	int type;
	void *ptr;
} Option;

#define	OSTR	0
#define	ONUM	1
#define	OSET	2
#define	OCLR	3
#define	OINC	4
#define	ODEC	5
#define	OSPC	6

static char *progname = "aoestart";
static char *edev;
static int eproto = EPROTO_AOE;
static int chan = -1;
static int dobcast = 1;
static int maxdata, mtu;
static char mymac[LEN_EADDR];

static Option opts[] = {
	"ether",	OSTR,	&edev,
	"eproto",	ONUM,	&eproto,
	"aoechan",	ONUM,	&chan,
	"bcast",	OSET,	&dobcast,
	"nobcast",	OCLR,	&dobcast,
	"maxdata",	ONUM,	&maxdata,
	"mtu",		ONUM,	&mtu,
	"jumbo",	OSPC,	NULL,
	"nojumbo",	OSPC,	NULL,
	"verbose",	OINC,	&verbose,
	NULL
};

static int doargs(char **);
static char **special(char **);
static void usage(void);
static int detach(void);
static void commmsg(char *, int);
static void aoercv(char *, int);
static void bcastquery(int, int, char *, int);

int
main(argc, argv)
int argc;
char **argv;
{
	int fd, cfd;
	char *p;

	if (argc < 1 || doargs(&argv[1]) < 0)
		usage();
	if (edev == NULL || chan < 0
	||  eproto < 0 || eproto > 0xffffU
	||  maxdata < 0 || maxdata > 0xffffU
	||  mtu < 0 || mtu > 0xffffU)
		usage();
	if (maxdata == 0 && mtu != 0)
		maxdata = mtu - AOEA_DATA;
	if ((p = strrchr(argv[0], '/')) == NULL)
		p = argv[0];
	else
		p++;
	loginit(p, verbose);
	progname = p;
	if ((fd = comminit(chan, edev, eproto, maxdata, mymac)) < 0) {
		logmsg(LE|LS, "failed to set up aoechan %d dev %s", chan, edev);
		return (1);
	}
	if (verbose)
		logmsg(LD, "%s: local addr %s", edev, ethertoa(mymac, NULL));
	if (dobcast) {
		if ((cfd = open(AOECTL, O_WRONLY)) < 0)
			logmsg(LE|LS, "cannot open %s: %s; no broadcast",
				AOECTL, strerror(errno));
		else {
			bcastquery(cfd, chan, mymac, eproto);
			close(cfd);
		}
	}
	if (achattach(fd, chan) < 0) {
		logmsg(LE|LS, "cannot attach aoechan %d", chan);
		return (1);
	}
	return (0);
}

static int
doargs(av)
char **av;
{
	Option *op;
	char *ep;
	int n;

	for (; *av; av++) {
		for (op = opts; op->name; op++)
			if (strcmp(*av, op->name) == 0)
				break;
		if (op->name == NULL)
			return (-1);
		switch (op->type) {
		case OSET:
			*(int *)op->ptr = 1;
			break;

		case OCLR:
			*(int *)op->ptr = 0;
			break;

		case OINC:
			(*(int *)op->ptr)++;
			break;

		case ODEC:
			(*(int *)op->ptr)--;
			break;

		case OSTR:
			if (*++av == NULL)
				return (-1);
			*(char **)op->ptr = *av;
			break;

		case ONUM:
			if (*++av == NULL)
				return (-1);
			n = strtol(*av, &ep, 0);
			if (*ep != 0)
				return (-1);
			*(int *)op->ptr = n;
			break;

		case OSPC:
			av = special(av);
			break;

		default:
			fprintf(stderr, "aoestart: internal option type error\n");
			return (-1);
		}
	}
	return (1);
}

static char **
special(av)
char **av;
{
	if (strcmp(*av, "jumbo") == 0) {
		maxdata = mtu = 0;
		return (av);
	} else if (strcmp(*av, "nojumbo") == 0) {
		maxdata = 0;
		mtu = MAXLEN_ETHER - LEN_ETHER;
		return (av);
	} else {
		fprintf(stderr, "aoestart: %s: internal option special error\n", *av);
		return (av);
	}
}
		
#define	OPL	6

static void
usage()
{
	int i;
	char *sep;

	fprintf(stderr, "usage: %s command ...", progname);
	for (i = 0; opts[i].name; i++) {
		if ((i % OPL) == 0)
			sep = "\n\t";
		else
			sep = "  ";
		fprintf(stderr, "%s%s", sep, opts[i].name);
		switch (opts[i].type) {
		case OSTR:
			fprintf(stderr, " (str)");
			break;

		case ONUM:
			fprintf(stderr, " (num)");
			break;
		}
	}
	fprintf(stderr, "\nrequired: ether (device) aoechan (channel-number)\n");
	exit(1);
}

static void
bcastquery(fd, chan, mac, proto)
int fd;
int chan;
char *mac;
int proto;
{
	char buf[LEN_ACSEND];
	char *p;
	int len, rlen;

	p = buf;
	len = AC_MESSAGE + LEN_ETHER + MINLEN_AOEI;
	memset(p, 0, len);
	toaoechar(p, AC_TYPE, ACSEND);
	toaoeshort(p, AC_LEN, len);
	toaoeshort(p, AC_CHANNEL, chan);
	p += AC_MESSAGE;	/* to MAC header */
	memset(p + ETHER_DST, 0xff, LEN_EADDR);	/* broadcast address */
	memcpy(p + ETHER_SRC, mymac, LEN_EADDR);
	toetherprot(p, ETHER_PROT, proto);
	p += LEN_ETHER;		/* to AoE message */
	toaoechar(p, AOE_VF, aoemkvf(0));
	toaoeshort(p, AOE_MAJOR, AOEMAJWILD);
	toaoechar(p, AOE_MINOR, AOEMINWILD);
	toaoechar(p, AOE_COMMAND, AOECID);
	toaoelong(p, AOE_TAG, 0);
	toaoechar(p, AOEI_VCC, aoeimkvcc(0, AOEICGET));
	toaoeshort(p, AOEI_CLEN, 0);
	if (debug || verbose) {
		logmsg(LD, "send broadcast query:");
		logpkt(LD, buf + AC_MESSAGE, len - AC_MESSAGE);
	}
	if ((rlen = write(fd, buf, len)) < 0)
		logmsg(LE, "send broadcast failed: %s", strerror(errno));
	else if (rlen != len)
		logmsg(LE, "send broadcast: short write, %d/%d", rlen, len);
}
