/*
 * Copyright 2007 Coraid, Inc.
 * See accompanying file doc/LICENSE for conditions
 */

#define	_FILE_OFFSET_BITS	64	/* Solaris-specific */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <limits.h>
#include <unistd.h>
#include <fcntl.h>	/* for open(2) */
#include <sys/dkio.h>

static char *progname = "aoeunlabel";

#define	SECSIZE	512

typedef int64_t Blkno;

static void usage(void);
static int unlabel(char *);
static int zapefi(char *, int, struct dk_minfo *);
static int zapvtoc(char *, int, struct dk_minfo *);
static int clearsec(char *, int, struct dk_minfo *, Blkno);

int
main(argc, argv)
int argc;
char **argv;
{
	int s;

	if (argc > 0)
		progname = *argv++;
	if (argc < 2)
		usage();
	s = 0;
	while (*argv)
		if (unlabel(*argv++) < 0)
			s = 1;
	return (s);
}

static void
usage()
{
	fprintf(stderr, "usage: %s device ...\n", progname);
	exit(1);
}

static int
unlabel(dev)
char *dev;
{
	int fd;
	struct dk_minfo ds;

	if ((fd = open(dev, O_WRONLY)) < 0) {
		fprintf(stderr, "%s: cannot open %s: %s\n",
			progname, dev, strerror(errno));
		return (-1);
	}
	if (ioctl(fd, DKIOCGMEDIAINFO, &ds) < 0) {
		fprintf(stderr, "%s: %s: cannot read disk size: %s\n",
			progname, dev, strerror(errno));
		close(fd);
		return (-1);
	}
	if (ds.dki_lbsize <= 0 || ds.dki_capacity <= 0) {
		fprintf(stderr, "%s: %s: unreasonable blocksize (%d) or capacity (%lld)\n",
			progname, dev, ds.dki_lbsize, ds.dki_capacity);
		close(fd);
		return (-1);
	}
	if (zapefi(dev, fd, &ds) < 0) {
		close(fd);
		return (-1);
	}
	/*
	 * if there was an EFI label,
	 * may have to reopen to get disk geometry
	 */
	close(fd);
	if ((fd = open(dev, O_WRONLY)) < 0) {
		fprintf(stderr, "%s: cannot reopen %s: %s\n",
			progname, dev, strerror(errno));
		return (-1);
	}
	if (zapvtoc(dev, fd, &ds) < 0) {
		close(fd);
		return (-1);
	}
	close(fd);
	return (1);
}

/*
 * clear EFI partition table
 * -- GPT resides in sectors 1, N-1
 */
static int
zapefi(dev, fd, dp)
char *dev;
int fd;
struct dk_minfo *dp;
{
	if (clearsec(dev, fd, dp, 1) < 0)
		return (-1);
	if (clearsec(dev, fd, dp, dp->dki_capacity - 1) < 0)
		return (-1);
	return (1);
}

/*
 * zap VTOC label
 * -- primary label in sector 0;
 *    backups in last track, at magic offsets
 */
static int
zapvtoc(dev, fd, dp)
char *dev;
int fd;
struct dk_minfo *dp;
{
	struct dk_geom g;
	Blkno b;
	int i;

	if (ioctl(fd, DKIOCGGEOM, &g) < 0) {
		fprintf(stderr, "%s: %s: cannot get VTOC geometry: %s\n",
			progname, dev, strerror(errno));
		return (-1);
	}
	b = g.dkg_ncyl + g.dkg_acyl;
	b *= g.dkg_nhead * g.dkg_nsect;
	b -= g.dkg_nsect;
	for (i = 1; i < 10; i += 2)
		if (clearsec(dev, fd, dp, b + i) < 0)
			break;
	if (clearsec(dev, fd, dp, 0) < 0)
		return (-1);
	return (1);
}

#define	SEEKERR	((off_t)-1)

static int
clearsec(dev, fd, dp, bno)
char *dev;
int fd;
struct dk_minfo *dp;
Blkno bno;
{
	static void *zbuf;
	static int zsize;
	int n;

	if (zbuf == NULL || zsize != dp->dki_lbsize) {
		if (zbuf)
			free(zbuf);
		if ((zbuf = malloc(dp->dki_lbsize)) == NULL) {
			fprintf(stderr, "%s: %s: cannot alloc sector buffer of size %d\n",
				progname, dev, dp->dki_lbsize);
			return (-1);
		}
		zsize = dp->dki_lbsize;
		memset(zbuf, 0, dp->dki_lbsize);
	}
	if (lseek(fd, bno * dp->dki_lbsize, SEEK_SET) == SEEKERR) {
		fprintf(stderr, "%s: %s: seek to sector %lld: %s\n",
			progname, dev, bno, strerror(errno));
		return (-1);
	}
	if ((n = write(fd, zbuf, zsize)) != n) {
		fprintf(stderr, "%s: %s: write error for bno %lld: %d/%d: %s\n",
			progname, dev, bno, n, zsize, strerror(errno));
		return (-1);
	}
	return (1);
}
