/*
 * Copyright 2005 Coraid, Inc.
 * See accompanying file doc/LICENSE for conditions
 */

/*
 * set up communications:
 * open and configure the network interface
 * --for now, assume the interface is Ethernet;
 * maybe that ought to be generalized later
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <setjmp.h>
#include <signal.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/time.h>
#include <stropts.h>
#include <sys/dlpi.h>
#include <kstat.h>
#include <inet/nd.h>
#include "aoecmd.h"
#include "aoecomm.h"
#include "aoeproto.h"
#include "aoecommproto.h"
#include "etherproto.h"

#define	DLREPLY	1024
#define	INITTMO	(3*1000)	/* msec */

#define	RDYSECS	30

#define	NOUNIT	(-1)

#define	ARB	100

static int etheropen(char *, int, char *, int *);
static int dlpisetup(int, int, char *, int *, int, char *);
static int msgout(int, void *, int, void *, int);
static int msgin(int, void *, int, int);
static void msgtmo();
static int etherready(int, char *, char *, int);
static int kstatready(char *, int);
static int nddready(int, char *, int);
static char *ndop(int, char *, char *);
static int rawether(int, char *);
static Aoecookie newcookie(void);

/*
 * record of a past confusion:
 * -- if the caller supplied maxdata == 0
 * (probably the user didn't specify a desired value)
 * and (unlikely) no default was picked up from the device MTU,
 * just leave it at zero;
 * aoecomm allows that, and inserts the AoE-protocol default value
 */

int
comminit(chan, name, proto, maxdata, retaddr)
int chan;
char *name;
int proto;
int maxdata;
char *retaddr;
{
	int efd, cfd;
	Aoecookie cookie;
	char cbuf[LEN_ACINITCHAN];
	char rbuf[LEN_ACINITACK];
	int n;

	if ((efd = etheropen(name, proto, retaddr, &maxdata)) < 0)
		return (-1);
	if (ioctl(efd, I_PUSH, AOECOMM) < 0) {
		logmsg(LE, "comminit: push %s: %s",
			AOECOMM, strerror(errno));
		close(efd);
		return (-1);
	}
	if (verbose)
		logmsg(LD, "%s: opened, aoecomm pushed ok", name);
	if ((cfd = open(AOECTL, O_WRONLY)) < 0) {
		logmsg(LE, "%s: cannot open: %s", AOECTL, strerror(errno));
		close(efd);
		return (-1);
	}
	cookie = newcookie();
	memset(cbuf, 0, sizeof(cbuf));
	toaoechar(cbuf, AC_TYPE, ACINITCHAN);
	toaoeshort(cbuf, AC_LEN, LEN_ACINITCHAN);
	toaoeshort(cbuf, AC_CHANNEL, chan);
	toaoeshort(cbuf, AC_PROTO, proto);
	toaoelong(cbuf, AC_COOKIE, cookie);
	memcpy(cbuf + AC_LOCADDR, retaddr, LEN_EADDR);
	toaoeshort(cbuf, AC_MAXDATA, maxdata);
	if (verbose)
		logmsg(LD, "init aoechan %d mac %s cookie 0x%x proto 0x%x maxdata %d",
			fraoeshort(cbuf, AC_CHANNEL),
			ethertoa(cbuf + AC_LOCADDR, NULL),
			fraoelong(cbuf, AC_COOKIE),
			fraoeshort(cbuf, AC_PROTO),
			fraoeshort(cbuf, AC_MAXDATA));
	n = twrite(cfd, cbuf, LEN_ACINITCHAN, INITTMO);
	close(cfd);
	if (n != LEN_ACINITCHAN) {
		logmsg(LE, "comminit: write INITCHAN to %s: %d/%d: %s",
			AOECTL, n, LEN_ACINITCHAN, strerror(errno));
		close(efd);
		return (-1);
	}
	if ((n = twrite(efd, cbuf, LEN_ACINITCHAN, INITTMO)) != LEN_ACINITCHAN) {
		logmsg(LE, "comminit: write INITCHAN to %s: %d/%d: %s",
			AOECOMM, n, LEN_ACINITCHAN, strerror(errno));
		close(efd);
		return (-1);
	}
	if ((n = tread(efd, rbuf, sizeof(rbuf), INITTMO)) != LEN_ACINITACK) {
		logmsg(LE, "comminit: read INITACK from %s: %d/%d: %s",
			AOECOMM, n, LEN_ACINITACK, strerror(errno));
		close(efd);
		return (-1);
	}
	if (fraoechar(rbuf, AC_TYPE) != ACINITACK
	||  fraoeshort(rbuf, AC_LEN) != LEN_ACINITACK) {
		logmsg(LE, "comminit: read INITACK from %s: garbled",
			AOECOMM);
		loghex(LE, rbuf, n);
		close(efd);
		return (-1);
	}
	if ((n = fraoechar(rbuf, AC_STATUS)) != 0) {
		logmsg(LE, "comminit: %s INITCHAN: %s",
			AOECOMM, strerror(n));
		close(efd);
		return (-1);
	}
	if (verbose)
		logmsg(LD, "aoechan %d ether %s: all init ok", chan, name);
	return (efd);
}

/*
 * Solaris/SVr4 DLPI gore
 * to open an Ethernet interface
 * and configure it to send and receive
 * a particular Ethernet protocol type
 */
static int
etheropen(name, p, ra, maxp)
char *name;
int p;
char *ra;
int *maxp;
{
	char d[ARB], dp[ARB];
	int fd;
	char *up;
	int unit;
	char *fname;

	if (*name == '/')
		strncpy(d, name, sizeof(d) - 1);
	else
		snprintf(d, sizeof(d), "/dev/%s", name);
	d[sizeof(d)-1] = 0;
	strcpy(dp, d);
	/*
	 * split dp into the part before and after numeric suffix
	 * keep d as the original complete name
	 */
	for (up = dp + strlen(dp); up > dp && isdigit(up[-1]); --up)
		;
	if (*up == 0)
		unit = 0;	/* no digits */
	else {
		unit = atoi(up);
		*up = 0;
	}
	if ((fd = open(d, O_RDWR)) >= 0) {
		unit = NOUNIT;
		fname = d;
	} else if ((fd = open(dp, O_RDWR)) >= 0)
		fname = dp;
	else {
		logmsg(LE|LS, "%s: cannot open: %s",
			d, strerror(errno));
		return (-1);
	}
	if (dlpisetup(fd, p, ra, maxp, unit, name) < 0) {
		close(fd);
		return (-1);
	}
	if (etherready(fd, name, fname, unit) <= 0) {
		close(fd);
		return (-1);
	}
	if (rawether(fd, name) < 0) {
		close(fd);
		return (-1);
	}
	return (fd);
}

/*
 * DLPI initialization commands:
 * attach to the PPA, bind the SAP
 */
static int
dlpisetup(fd, proto, ra, maxp, unit, name)
int fd;
int proto;
char *ra;
int *maxp;
int unit;
char *name;
{
	union {
		dl_attach_req_t a;
		dl_bind_req_t b;
		dl_info_req_t i;
	} req;
	union {
		char a[DLREPLY];	/* room for addr */
		dl_ok_ack_t o;
		dl_error_ack_t e;
		dl_bind_ack_t b;
		dl_phys_addr_ack_t p;
		dl_info_ack_t i;
	} ack;
	int n;

	/*
	 * DL_ATTACH_REQ: attach to desired PPA (device unit)
	 * required only if we opened a unitless device
	 */
	if (unit != NOUNIT) {
		req.a.dl_primitive = DL_ATTACH_REQ;
		req.a.dl_ppa = unit;
		if (msgout(fd, &req, sizeof(req.a), NULL, 0) < 0
		|| (n = msgin(fd, &ack, sizeof(ack), 0)) < sizeof(ack.o)) {
			logmsg(LE, "%s: cannot DL_ATTACH: %d/%d bytes",
				name, n, sizeof(ack.o));
			return (-1);
		}
		if (ack.e.dl_primitive == DL_ERROR_ACK) {
			logmsg(LE, "%s: DL_ATTACH error: dlpi %d%s%s\n",
				name, ack.e.dl_errno,
				ack.e.dl_errno == DL_SYSERR ? ": " : "",
				ack.e.dl_errno == DL_SYSERR ?
					strerror(ack.e.dl_unix_errno) : "");
			return (-1);
		}
		if (ack.o.dl_primitive != DL_OK_ACK) {
			logmsg(LE, "%s: DL_ATTACH returned 0x%x",
				name, ack.o.dl_primitive);
			return (-1);
		}
		if (ack.o.dl_correct_primitive != DL_ATTACH_REQ) {
			logmsg(LE, "%s: unexpected ack for cmd 0x%x",
				name, ack.o.dl_correct_primitive);
			return (-1);
		}
	}
	/*
	 * DL_BIND: set Ethernet protocol number
	 */
	memset(&req.b, 0, sizeof(req.b));
	req.b.dl_primitive = DL_BIND_REQ;
	req.b.dl_sap = proto;
	req.b.dl_service_mode = DL_CLDLS;
	if (msgout(fd, &req.b, sizeof(req.b), NULL, 0) < 0
	|| (n = msgin(fd, &ack, sizeof(ack), 0)) < sizeof(ack.b)) {
		logmsg(LE, "%s: cannot DL_BIND: %d/%d",
			name, n, sizeof(ack));
		return (-1);
	}
	if (ack.b.dl_primitive != DL_BIND_ACK) {
		logmsg(LE, "%s: DL_BIND_ACK returned 0x%x",
			name, ack.b.dl_primitive);
		return (-1);
	}
	if (ack.b.dl_addr_offset + ack.b.dl_addr_length > n) {
		logmsg(LE, "%s: DL_BIND_ACK garbled: addr_offset %d, length %d, message len %d",
			name, ack.b.dl_addr_offset,
			ack.b.dl_addr_length, n);
		return (-1);
	}
	/*
	 * assume address returned by bind is [MAC addr||ether proto]
	 */
	if (ack.b.dl_addr_length != LEN_EADDR + 2) {
		logmsg(LE, "%s: DL_BIND_ACK address length unexpected: %d/%d",
			name, ack.b.dl_addr_length, LEN_EADDR + 2);
		return (-1);
	}
	memcpy(ra, ack.a + ack.b.dl_addr_offset, LEN_EADDR);
	if (verbose)
		logmsg(LD, "%s: proto 0x%x MAC %s",
			name, ack.b.dl_sap, ethertoa(ra, NULL));
	/*
	 * DL_INFO: fetch current MTU
	 */
	req.i.dl_primitive = DL_INFO_REQ;
	if (msgout(fd, &req.i, sizeof(req.i), NULL, 0) < 0
	|| (n = msgin(fd, &ack, sizeof(ack), 0)) < sizeof(ack.b)) {
		logmsg(LE, "%s: cannot DL_INFO: %d/%d",
			name, n, sizeof(ack));
		return (-1);
	}
	if (ack.i.dl_primitive != DL_INFO_ACK) {
		logmsg(LE, "%s: DL_INFO_ACK returned 0x%x",
			name, ack.i.dl_primitive);
	} else {
		n = ack.i.dl_max_sdu - AOEA_DATA;
		if (verbose) {
			logmsg(LD, "%s: mac type %d; DLSDU %d == maxdata %d",
				name, ack.i.dl_mac_type,
				ack.i.dl_max_sdu, n);
			if (maxp)
				logmsg(LD, "%s: argument maxdata %d, device %d",
					name, *maxp, n);
		}
		if (maxp && (*maxp > n || *maxp <= 0))
			*maxp = n;
	}
	return (1);
}

static int
msgout(fd, obuf, olen, dbuf, dlen)
int fd;
void *obuf;
int olen;
void *dbuf;
int dlen;
{
	struct strbuf ctl, dat;

	ctl.len = olen;
	ctl.buf = obuf;
	dat.len = dlen;
	dat.buf = dbuf;
	if (putmsg(fd, &ctl, dbuf ? &dat : NULL, 0) < 0) {
		logmsg(LE, "msgout: putmsg: %s", strerror(errno));
		return (-1);
	}
	return (1);
}

static jmp_buf msgjmp;

static int
msgin(fd, ibuf, ilen, tmo)
int fd;
void *ibuf;
int ilen;
int tmo;
{
	struct strbuf ctl;
	int gflg;
	void (*osig)();

	ctl.maxlen = ilen;
	ctl.len = 0;
	ctl.buf = ibuf;
	gflg = 0;
	if (tmo) {
		if (setjmp(msgjmp)) {
			signal(SIGALRM, osig);
			logmsg(LE, "msgin: timeout");
			return (-1);
		}
		osig = signal(SIGALRM, msgtmo);
		alarm(tmo);
	}
	if (getmsg(fd, &ctl, NULL, &gflg) < 0) {
		logmsg(LE, "msgin: getmsg: %d", strerror(errno));
		return (-1);
	}
	if (tmo) {
		alarm(0);
		signal(SIGALRM, osig);
	}
	return (ctl.len);
}

static void
msgtmo()
{
	longjmp(msgjmp, 1);
}

/*
 * wait for Ethernet device to become ready
 * takes a few seconds for some devices on some systems,
 * and the driver won't hold data, so if we don't wait
 * our initial broadcast will be lost
 *
 * the kstat algorithm (look for kstat parameter link_up)
 * works on newer systems and drivers, but not older
 * the ndd algorithm (ndd (device) link_status)
 * works only on older systems, and relies on global state
 * that some other process could change underfoot
 *
 * this routine, and each algorithm routine, returns:
 *	>0	link definitely up
 *	0	link definitely down
 *	-1	can't tell
 */
static int
etherready(fd, name, d, unit)
int fd;
char *name;
char *d;
int unit;
{
	int ans;

	if ((ans = kstatready(name, unit)) >= 0
	||  (ans = nddready(fd, d, unit)) >= 0) {
		if (ans == 0)
			logmsg(LE, "%s: device has no link", name);
		return (ans);
	}
	logmsg(LE, "%s: cannot determine link, assuming up", name);
	return (1);
}

static int
kstatready(name, unit)
char *name;
int unit;
{
	char *kn;
	kstat_ctl_t *kc;
	kstat_t *ks;
	kstat_named_t *kd;
	int i, val;

	if ((kn = strrchr(name, '/')) != NULL)
		kn++;
	else
		kn = name;
	if ((kc = kstat_open()) == NULL) {
		logmsg(LE, "kstat_open: %s", strerror(errno));
		return (-1);
	}
	for (ks = kc->kc_chain; ks; ks = ks->ks_next) {
		if (ks->ks_type != KSTAT_TYPE_NAMED)
			continue;
		if (strcmp(kn, ks->ks_name) != 0)
			continue;
		if (kstat_read(kc, ks, NULL) < 0) {
			logmsg(LE, "kstat %s, class %s: kstat_read: %s",
				ks->ks_name, ks->ks_class, strerror(errno));
			continue;
		}
		if ((kd = kstat_data_lookup(ks, "link_up")) == NULL)
			continue;
		switch (kd->data_type) {
		case KSTAT_DATA_INT32:
		case KSTAT_DATA_UINT32:
		case KSTAT_DATA_INT64:
		case KSTAT_DATA_UINT64:
			break;

		default:	/* unknown data type */
			continue;
		}
		break;		/* this is the kstat we want */
	}
	if (ks == NULL)	{	/* no kstat with link_up for our device */
		if (debug)
			logmsg(LD, "%s: no kstat with link_up found", name);
		kstat_close(kc);
		return (-1);
	}
	for (i = 0; i < RDYSECS; i++) {
		if (i > 0)
			sleep(1);
		if (kstat_read(kc, ks, NULL) < 0) {
			logmsg(LE, "kstat %s, class %s: kstat_read: %s",
				ks->ks_name, ks->ks_class, strerror(errno));
			kstat_close(kc);
			return (-1);
		}
		if ((kd = kstat_data_lookup(ks, "link_up")) == NULL) {
			logmsg(LE, "kstat %s%d, class %s: link_up disappeared",
				ks->ks_name, ks->ks_class);
			kstat_close(kc);
			return (-1);
		}
		switch (kd->data_type) {
		case KSTAT_DATA_INT32:
			val = kd->value.i32;
			break;

		case KSTAT_DATA_UINT32:
			val = kd->value.ui32;
			break;

		case KSTAT_DATA_INT64:
			val = kd->value.i64;
			break;

		case KSTAT_DATA_UINT64:
			val = kd->value.ui64;
			break;

		default:
			logmsg(LE, "kstat %s, class %s: link_up data type changed to %d",
				ks->ks_name, ks->ks_class, kd->data_type);
			return (-1);
		}
		if (val > 0) {
			if (debug)
				logmsg(LD, "%s: kstat %s class %s says link up",
					name, ks->ks_name, ks->ks_class);
			kstat_close(kc);
			return (1);
		}
	}
	if (debug)
		logmsg(LD, "%s: kstat %s class %s never link up",
			name, ks->ks_name, ks->ks_class);
	kstat_close(kc);
	return (0);
}

/*
 * works on most older devices and systems:
 * test link with the equivalent of
 *	ndd -get (device) instance	# and save
 *	ndd -set (device) instance (unit)
 *	ndd -get (device) link_status
 *	ndd -set (device) instance (prior unit)
 * the instance number is global to the whole system,
 * at least in older systems, hence the save and restore
 * beware that this also introduces a potential race over
 * the global instance state!  ugh.
 */
static int
nddready(fd, d, unit)
int fd;
char *d;
int unit;
{
	int i, ans;
	char *p;
	char inst[20];
	char oinst[20];

	snprintf(inst, sizeof(inst), "%d", unit);
	for (i = 0; i < RDYSECS; i++) {
		if (i > 0)
			sleep(1);
		if (unit != NOUNIT) {
			if ((p = ndop(fd, "instance", NULL)) == NULL) {
				logmsg(LE, "ndd %s: warning: cannot get instance", d);
				return (-1);
			}
			strncpy(oinst, p, sizeof(oinst));
			if (ndop(fd, "instance", inst) == NULL) {
				logmsg(LE, "ndd %s: warning: cannot set instance %s",
					d, inst);
				return (-1);
			}
		}
		if ((p = ndop(fd, "link_status", NULL)) != NULL)
			ans = atoi(p);
		else {
			logmsg(LE, "ndd %s: warning: cannot get link_status", d);
			ans = 1;		/* but still reset unit */
		}
		if (unit != NOUNIT) {
			if (ndop(fd, "instance", oinst) == NULL)
				logmsg(LE, "ndd %s: warning: cannot reset instance to %s",
					d, oinst);
		}
		if (ans > 0) {
			if (debug)
				logmsg(LD, "ndd %s%d: link up", d, unit);
			return (1);
		}
	}
	if (debug)
		logmsg(LD, "ndd %s%d: never link up", d, unit);
	return (0);
}

static char *
ndop(fd, pname, arg)
int fd;
char *pname;
char *arg;
{
	struct strioctl sioc;
	static char buf[ARB];
	char *p, *q, *ep;

	memset(buf, 0, sizeof(buf));
	p = buf;
	ep = buf + sizeof(buf);
	q = pname;
	while (*q && p < ep)
		*p++ = *q++;
	if (p >= ep) {
		logmsg(LE, "internal error: ndop overflow");
		return (NULL);
	}
	*p++ = 0;
	if (arg == NULL)
		sioc.ic_cmd = ND_GET;
	else {
		sioc.ic_cmd = ND_SET;
		q = arg;
		while (*q && p < ep)
			*p++ = *q++;
		if (p >= ep) {
			logmsg(LE, "internal error: ndop overflow");
			return (NULL);
		}
		*p++ = 0;
	}
	sioc.ic_timout = -1;
	sioc.ic_len = sizeof(buf);	/* for return value too */
	sioc.ic_dp = buf;
	if (ioctl(fd, I_STR, &sioc) < 0) {
		logmsg(LE, "ndd-%s %s I_STR: %s", arg ? "set" : "get", pname,
			strerror(errno));
		return (NULL);
	}
	/*
	 * ensure NUL-terminated string (sanity check)
	 */
	for (p = buf; p < ep; p++)
		if (*p == 0)
			return (buf);	/* NUL-term, so OK */
	return (NULL);
}

/*
 * set Ethernet device to raw mode
 */
static int
rawether(fd, name)
int fd;
char *name;
{
	struct strioctl sioc;

	sioc.ic_cmd = DLIOCRAW;
	sioc.ic_timout = -1;
	sioc.ic_len = 0;
	sioc.ic_dp = NULL;
	if (ioctl(fd, I_STR, &sioc) < 0) {
		logmsg(LE, "%s: DLIOCRAW: %s",
			name, strerror(errno));
		return (-1);
	}
	return (1);
}

/*
 * invent a random cookie
 * --use /dev/urandom if available,
 * otherwise pseudo-random
 */

static Aoecookie
newcookie()
{
	int fd;
	struct timeval t;
	int i;
	Aoecookie c;

	if ((fd = open("/dev/urandom", 0)) >= 0) {
		i = read(fd, &c, sizeof(c));
		close(fd);
		if (i == sizeof(c))
			return (c);
	}
	(void) gettimeofday(&t, NULL);
	srand(t.tv_usec);
	for (c = rand(), i = (getpid() % 127) + sizeof(c); i > 0; --i)
		c = (c << 8) | ((rand()>>7) & 0xff);
	return (c);
}
