/*
 * Copyright 2005 Coraid, Inc.
 * See accompanying file doc/LICENSE for conditions
 */

/*
 * logging subroutines
 * -- use syslog, but impose our own layer of abstraction
 * so it can change later,
 * and take care over long buffers
 */

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <syslog.h>
#include "aoecmd.h"
#include "aoeproto.h"
#include "ata.h"
#include "etherproto.h"

typedef struct {
	int v;
	char *s;
} Flag;

#define	MAXLINE	250

static char digits[16] = "0123456789abcdef";
static int tostderr;
static char progname[30];

static char *flagstoa(int, Flag *);

void
loginit(name, isdebug)
char *name;
int isdebug;
{
	strncpy(progname, name, sizeof(progname)-1);
	openlog(progname, LOG_PID, LOG_DAEMON);
	tostderr = isdebug;
}

/*
 * interpret an AoE message, with MAC header
 * all very ad-hoc, as is often the case in such stuff;
 * if it is probably valid try to interpret,
 * if not bail out to a hex dump
 */

static char *aoeerrs[] = {
	"(none)",
	"invalid-command",
	"invalid-arg",
	"device-unavailable",
	"config-str-full",
	"invalid-AoE-ver",
};
#define	MAXERR	(sizeof(aoeerrs)/sizeof(aoeerrs[0]))

static Flag aoeflags[] = {
	AOEF_R, "R", AOEF_E, "E",
	0
};

static Flag ataflags[] = {
	AOEAFE, "E", AOEAFD, "D",
	AOEAFA, "A", AOEAFW, "W",
	0
};

static Flag ataerrflags[] = {
	ATAER_AMNF, " AMNF",
	ATAER_TK0NF, " TK0NF",
	ATAER_ABRT, " ABRT",
	ATAER_MCR, " MCR",
	ATAER_IDNF, " IDNF",
	ATAER_MC, " MC",
	ATAER_UNC, " UNC",
	0
};

static char *aoeccmd[] = {
	"get-current",
	"match-whole",
	"match-prefix",
	"set-if-clear",
	"set-forced",
};
#define	MAXCCMD	(sizeof(aoeccmd)/sizeof(aoeccmd[0]))

void
logpkt(type, p, len)
int type;
char *p;
int len;
{
	char b1[LEN_EADDR*3], b2[LEN_EADDR*3];
	char *s;
	int n, nn;

	if (len < LEN_ETHER) {
		logmsg(type, "--invalid MAC header: %d bytes", len);
		loghex(type, p, len);
		return;
	}
	logmsg(type, "dst %s src %s type 0x%x",
		ethertoa(p + ETHER_DST, b1),
		ethertoa(p + ETHER_SRC, b2),
		fretherprot(p, ETHER_PROT));
	p += LEN_ETHER;
	len -= LEN_ETHER;
	n = fraoechar(p, AOE_VF);
	if (len < MINLEN_AOE || aoeversion(n) != AOEV1) {
		logmsg(type, "--invalid AoE header: %d bytes", len);
		loghex(type, p, len);
		return;
	}
	if ((nn = fraoeshort(p, AOE_MAJOR)) == AOEMAJWILD)
		strcpy(b1, "(wild)");
	else
		snprintf(b1, sizeof(b1), "%d", nn);
	if ((nn = fraoechar(p, AOE_MINOR)) == AOEMINWILD)
		strcpy(b2, "(wild)");
	else
		snprintf(b2, sizeof(b2), "%d", nn);
	nn = fraoechar(p, AOE_ERROR);
	if (nn < 0 || nn >= MAXERR)
		s = "(unknown)";
	else
		s = aoeerrs[nn];
	logmsg(type, "\tAoE V%d flag 0x%x [%s] tag 0x%x maj %s min %s error %d %s",
		aoeversion(n), n & ~AOEF_VER, flagstoa(n, aoeflags),
		fraoelong(p, AOE_TAG), b1, b2, nn, s);
	n = fraoechar(p, AOE_COMMAND);
	switch (n) {
	case AOECATA:
		s = fraoechar(p, AOE_VF) & AOEF_R ? "rsp" : "cmd";
		if (len < MINLEN_AOEA) {
			logmsg(type, "\t%s ATA: too short: %d bytes", s, len);
			loghex(type, p, len);
			return;
		}
		n = fraoechar(p, AOEA_AF);
		logmsg(type, "\t%s ATA: AFlags 0x%x [%s] err/feat 0x%02x nsec 0x%02x cmd/stat 0x%02x",
			s, n, flagstoa(n, ataflags),
			fraoechar(p, AOEA_EF), fraoechar(p, AOEA_NSEC),
			fraoechar(p, AOEA_CS));
		logmsg(type, "\tlba0 0x%02x lba1 0x%02x lba2 0x%02x lba3 0x%02x lba4 0x%02x lba5 0x%02x",
			fraoechar(p, AOEA_LBA0), fraoechar(p, AOEA_LBA1),
			fraoechar(p, AOEA_LBA2), fraoechar(p, AOEA_LBA3),
			fraoechar(p, AOEA_LBA4), fraoechar(p, AOEA_LBA5));
		if ((fraoechar(p, AOE_VF) & AOEF_R) != 0
		&&  (n = fraoechar(p, AOEA_EF)) != 0)
			logmsg(type, "\terr 0x%02x =%s",
				n, flagstoa(n, ataerrflags));
		p += AOEA_DATA;
		len -= AOEA_DATA;
		if (len > 0) {
			logmsg(type, "\tData: %d bytes", len);
			loghex(type, p, len);
			len = 0;
		}
		break;

	case AOECID:
		s = fraoechar(p, AOE_VF) & AOEF_R ? "rsp" : "cmd";
		if (len < MINLEN_AOEI) {
			logmsg(type, "\t%s Query-Config: too short: %d bytes", s, len);
			loghex(type, p, len);
			return;
		}
		n = fraoechar(p, AOEI_VCC);
		nn = fraoeshort(p, AOEI_FV);
		logmsg(type, "\t%s Query-Config: firmware %d.%d maxbuf %d maxsec %d AoEver %d",
			s, aoefvmaj(nn), aoefvmin(nn),
			fraoeshort(p, AOEI_MAXBUF),
			fraoechar(p, AOEI_MAXSEC),
			aoeiver(n));
		n = aoeicc(n);
		if (n < 0 || n > MAXCCMD)
			s = "(unknown)";
		else
			s = aoeccmd[n];
		nn = fraoeshort(p, AOEI_CLEN);
		logmsg(type, "\tconfig string: cmd %d = %s, strlen %d%s",
			n, s, nn, nn > len - AOEI_CSTR ? " (truncated)" : "");
		p += AOEI_CSTR;
		len -= AOEI_CSTR;
		if (nn > len)
			nn = len;
		if (nn > 0) {
			loghex(type, p, nn);
			p += nn;
			len -= nn;
		}
		break;

	default:
		p += AOE_ARG;
		len -= AOE_ARG;
		logmsg(type, "\tunknown command %d: %d bytes of data", n, len);
		loghex(type, p, len);
		len = 0;
		break;
	}
	if (len > 0) {
		logmsg(type, "\t%d bytes of trailing junk:", len);
		loghex(type, p, len);
	}
}

char *
ethertoa(ea, buf)
char *ea;
char *buf;
{
	int i;
	char *p;
	static char sbuf[LEN_EADDR*3];

	if (buf == NULL)
		buf = sbuf;
	p = buf;
	for (i = 0; i < LEN_EADDR; i++) {
		if (i)
			*p++ = ':';
		*p++ = digits[(ea[i]>>4)&0xf];
		*p++ = digits[ea[i]&0xf];
	}
	*p = 0;
	return (buf);
}

static char *
flagstoa(v, fp)
int v;
Flag *fp;
{
	static char buf[100];
	char *p;
	int n;

	for (p = buf; fp->v && p < buf + sizeof(buf) - 1; fp++)
		if (fp->v & v) {
			n = strlen(fp->s);
			if (p + n >= buf + sizeof(buf) - 1)
				break;
			memcpy(p, fp->s, n);
			p += n;
		}
	*p = 0;
	return (buf);
}

/*
 * dump some data in hex
 */

#define	BPL	16	/* raw bytes per line */

void
loghex(type, p, len)
int type;
char *p;
int len;
{
	char buf[(BPL*3)+1];	/* BPL pairs preceded by spacer + NUL */
	char *b;
	int i;

	for (i = 0, b = buf; len > 0; ++i, --len) {
		if (i >= BPL) {
			*b = 0;
			logmsg(type, "%s", buf);
			i = 0;
			b = buf;
		}
		if (i == 0)
			*b++ = '\t';
		else
			*b++ = ' ';
		*b++ = digits[(*p>>4)&0xf];
		*b++ = digits[*p & 0xf];
		p++;
	}
	if (b > buf) {
		*b = 0;
		logmsg(type, "%s", buf);
	}
}

/*
 * log a single line
 */
void
logmsg(int type, char *fmt, ...)
{
	va_list ap;
	int stype;
	char buf[MAXLINE];

	va_start(ap, fmt);
	vsnprintf(buf, sizeof(buf), fmt, ap);
	va_end(ap);
	switch (type & LTYPE) {
	case LD:
		stype = LOG_DEBUG;
		break;

	case LE:
		stype = LOG_ERR;
		break;

	case LN:
		stype = LOG_WARNING;
		break;

	default:
		stype = LOG_NOTICE;
		break;
	}
	syslog(stype, "%s", buf);
	if ((type & LS) || tostderr)
		fprintf(stderr, "%s: %s\n", progname, buf);
}
